#include "redef.h"
#include "intrinsic_sizes.h"
#include "switches.h"
! Fre energy versus time
! Free energy spectral analysis (kx,ky,kperp ...)
! Transfer function
Module diag_Gyro_LES
  
  Use par_mod 
  Use file_io, only: get_unit_nr
  Use vel_space, only: fm, mat_00
  use geometry
  use communications
  use coordinates
  use aux_fields
  Use gyro_average_ff_mod, only: gyro_average_ff
  Use gyro_average_df_mod, only: gyro_average_df
  use collisions, only: equ_collisions
  use diagnostics_energy, only: get_cfgamma_f,get_cfgamma_h,&
       &energy_integral
  use dfdzv_terms 
  use dgdxy_terms
  use dzv_terms
  use dfdxy_terms, only: add_dfdxy
  use dchidz_term, only: add_dchidz
  use dchidxy_terms, only: add_dchidxy, add_dchidxy_orig
  use sources_mod, only: add_f1_sources_block
  use spatial_averages, only: sum_int_3d, sum_int_z
  use blockindex
  use axpy
  use prefactors
  use numerical_damping
  use aux_func, only: gamma0
  USE calc_rhs, only: this_nonlinear_term
  
  Implicit None

  PUBLIC :: initialize_all_diags_GyroLES, exec_all_diags_GyroLES, finalize_all_diags_GyroLES
  PUBLIC :: mem_est_diag_GyroLES

  !quantities for free energy consevation vs time
  integer, public :: istep_fe_time = 0
  ! we will call this 2D diag every istep_fe_twoD and perform the average on windows of length avg_window
  integer, public :: istep_fe_twoD = 0, avg_window = 1  
  integer, dimension(8),public :: fnumber ! Will store the file id's for all quantity/species couples
  ! quantities for free energy transfer
  !moved to par_in.F90
  !integer, public :: istep_fe_transfer = 0  
  integer, public :: n_shells_ky = 25 ! For security in the allocatable variable
  integer, public :: n_shells_kx = 25 ! For security in the allocatable variable
  integer, public :: n_shells_kperp = 25 ! For security in the allocatable variable
  integer, public :: n_shells_log_kperp = 25 ! For security in the allocatable variable
  integer, public :: n_shells_log_ky = 25 ! For security in the allocatable variable
  integer, public :: n_shells_log_kperp_1 = 25 ! For security in the allocatable variable
  logical, public :: tky = .false.
  logical, public :: tkx = .false.
  logical, public :: tkperp = .false.
  logical, public :: t_log_kperp = .false.
  logical, public :: t_log_kperp_fb = .false.
  logical, public :: t_log_kperp_1 = .false.
  logical, public :: t_log_ky = .false.
  logical, public :: tk3_log_kperp = .false.
  logical, public :: tk3_log_kperp_1 = .false.
  logical, public :: tk3_log_kperp_h = .false.
  logical, public :: tk3_ky = .false.
  logical, public :: tk3_kx = .false.
  logical, public :: with_nvp = .false.
  real, public ::  delta_ky, delta_kx, delta_kperp, lambda1, lambda2

  PRIVATE
  !***** Common parameters to all diags *****
  Character(Len=8) :: filestat='replace', filepos='rewind'

  !***** Parameters for free energy consevation vs time *****
  Integer :: FE_TIME_FILE
  Integer, dimension(:,:),allocatable :: FE_TIME_FILE_SPEC
  Complex, dimension(:,:,:,:,:,:),allocatable :: g_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable :: gc_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable :: d_v_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable :: d_z_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: d_kperp_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: coll_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: c_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: p_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: fe_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: fe_g_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: fe_d_v_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: fe_d_z_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: fe_d_kperp_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: fe_coll_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: fe_c_rhs
  Complex, dimension(:,:,:,:,:,:),allocatable  :: fe_p_rhs
  Complex, dimension(:,:,:,:,:,:,:),allocatable :: fe_rhs_n
  Complex, dimension(:,:,:,:,:,:,:),allocatable  :: fe_g_rhs_n
  Complex, dimension(:,:,:,:,:,:,:),allocatable  :: fe_d_v_rhs_n
  Complex, dimension(:,:,:,:,:,:,:),allocatable  :: fe_d_z_rhs_n
  Complex, dimension(:,:,:,:,:,:,:),allocatable  :: fe_d_kperp_rhs_n
  Complex, dimension(:,:,:,:,:,:,:),allocatable  :: fe_coll_rhs_n
  Complex, dimension(:,:,:,:,:,:,:),allocatable  :: fe_c_rhs_n
  Complex, dimension(:,:,:,:,:,:,:),allocatable  :: fe_p_rhs_n
  Complex, dimension(:,:,:,:,:,:),allocatable  :: fe_lhs_1
  Complex, dimension(:,:,:,:,:,:),allocatable  :: fe_lhs_2
  Complex, dimension(:,:,:,:,:,:),allocatable  :: cfgamma
  Complex, dimension(:,:,:,:,:,:),allocatable  :: cfgamma_last
  ! Condiational terms if filter defined
  real, dimension(:), allocatable :: time_storage
  Complex, Dimension(:,:,:,:,:,:), Allocatable  :: g_last
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: f_last
  !Complex, Dimension(:,:,:,:,:,:), Allocatable :: h_last
  Complex, Dimension(:,:,:,:), Allocatable::  emfields_last
  Complex, Dimension(:,:,:,:,:,:), Allocatable  :: filter_g_1
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: filter_f_ 
  Complex, Dimension(:,:,:,:), Allocatable::  filter_emfields
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: nlt
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: filter_nlt    
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: nlt_filter_g_1 
  Complex, Dimension(:,:,:,:,:,:), Allocatable:: filter_nlt_filter_g_1  
  Complex, Dimension(:,:,:,:,:,:), Allocatable  :: t_fgs

  !****** Quantities for 2D spectral diags
  integer, dimension(:,:),allocatable, private :: fnumber_spec ! Will store the file id's for all quantity/species couples
  integer, dimension(:), allocatable, private :: twoD_counter ! Stores the number of windows for average
  integer, dimension(:,:), allocatable, private :: twoD_counter_spec ! Stores the number of windows for average
  real, dimension(:,:,:), allocatable, private :: cumul_kxky
  real, dimension(:,:,:,:), allocatable, private :: cumul_kxky_spec
  Complex, Dimension(:,:,:,:,:,:,:), Allocatable :: v7d
  Complex, Dimension(:,:,:,:,:,:,:), Allocatable :: v7d_spec
  !****** Parameteres for transfer function  
  real, private :: kperp_max, k0, k0_y
  Integer, dimension(:), allocatable, private :: transfer_counter_3d
  real, dimension(:,:,:), allocatable, private :: kperp
  integer, private :: n_transfer_models = 7
  Real, dimension(:), allocatable :: K_VALUE
  Integer,dimension(:,:,:),allocatable  :: FE_TF_FILE
  Integer,dimension(:,:,:,:),allocatable  :: FE_SP_FILE
  integer, dimension(:,:,:),allocatable :: FE_TF_SPEC_3D_FILE
  real, dimension(:,:,:,:), allocatable :: fe_tf_tot
  real, dimension(:,:,:,:), allocatable :: fe_tot
  integer, dimension(:), allocatable :: transfer_counter ! Stores the number of windows for average
  real, dimension(:,:,:,:), allocatable :: tf_cumul_1
  real, dimension(:,:,:,:), allocatable :: tf_cumul_2
  real, dimension(:,:,:,:), allocatable :: tf_cumul_3
  real, dimension(:,:,:,:), allocatable :: tf_cumul_4
  real, dimension(:,:,:,:), allocatable :: tf_cumul_5
  real, dimension(:,:,:,:), allocatable :: tf_cumul_6
  real, dimension(:,:,:,:), allocatable :: tf_cumul_7
  real, dimension(:,:,:,:), allocatable :: spec_cumul_1
  real, dimension(:,:,:,:), allocatable :: spec_cumul_2
  real, dimension(:,:,:,:), allocatable :: spec_cumul_3
  real, dimension(:,:,:,:), allocatable :: spec_cumul_4
  real, dimension(:,:,:,:), allocatable :: spec_cumul_5
  real, dimension(:,:,:,:), allocatable :: spec_cumul_6
  real, dimension(:,:,:,:), allocatable :: spec_cumul_7
  real, dimension(:,:,:,:,:), allocatable :: tf_cumul_3d_1
  real, dimension(:,:,:,:,:), allocatable :: tf_cumul_3d_2
  real, dimension(:,:,:,:,:), allocatable :: tf_cumul_3d_4
  real, dimension(:,:,:,:,:), allocatable :: tf_cumul_3d_6
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: nlt_ls2_g
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: nlt_ls2_f
 
  ! Allocatable variables for the 3d transfer
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: g_ls1 
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: g_ls2 
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: g_ls3
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: h_mod_ls1
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: h_mod_ls2
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: h_mod_ls3
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: nlt_ls2_ls3_g
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: nlt_ls2_ls3_f
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: nlt_ls2_ls3_h
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: cfgamma1
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: cfgamma2
  Complex, Dimension(:,:,:,:), Allocatable :: cfgamma3
  Complex, Dimension(:,:,:,:,:,:,:), Allocatable :: fe_tf_term
  Complex, Dimension(:,:,:,:,:,:,:), Allocatable :: fe_tf_term_p
  Complex, Dimension(:,:,:,:,:,:,:), Allocatable :: fe_tf_term_n
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: f_ls1
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: f_ls2
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: f_ls3
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: h_ls1
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: h_ls2
  Complex, Dimension(:,:,:,:,:,:), Allocatable :: h_ls3
  Complex, Dimension(:,:,:,:), Allocatable :: fields_ls1
  Complex, Dimension(:,:,:,:), Allocatable :: fields_ls2
  Complex, Dimension(:,:,:,:), Allocatable :: fields_ls3
    
Contains

!!!******************************************************************!!!
!!!******************************************************************!!!
  !>Give an estimate of the memory requirements of diag_GyroLES
  Real Function mem_est_diag_GyroLES(mem_req_in)
    real, parameter:: mem_cn=SIZE_OF_COMPLEX/(1024.)**2,&
         & mem_rn=SIZE_OF_REAL/(1024.)**2
    real:: mem_req_in
    real:: mem_loc=0.
    real:: mem_loc_max=0 !one diag a time, arrays are deallocated

    mem_loc = 0.
    if (istep_fe_time.gt.0) then
       !g_last,g_rhs,d_v_rhs,d_z_rhs, d_kperp_rhs,coll_rhs
       !c_rhs,fe_rhs,fe_g_rhs,fe_d_z_rhs,fe_d_v_rhs,fe_d_kperp_rhs
       !fe_coll_rhs,fe_c_rhs,fe_p_rhs,fe_rhs_n,fe_g_rhs_n,fe_d_v_rhs_n
       !fe_d_z_rhs_n,fe_d_kperp_rhs_n.....and cfgammas
       mem_loc = 37*lijklmn0*mem_cn
       !f_last, h_last
       mem_loc = mem_loc + lij0*lzvwn0*mem_cn
       !if (arakawa_zv) mem_loc = mem_loc + lij0*lzvwn0*mem_cn
       !emfields_last  
       mem_loc = mem_loc + lij0*lz0*n_fields*mem_rn
       if (fracy.gt.0.and.nonlinear) then
          mem_loc = mem_loc + 5*lijklmn0*mem_cn
          mem_loc = mem_loc + lx0*lj0*lzvwn0*mem_cn
          mem_loc = mem_loc + lij0*lz0*n_fields*mem_rn
       endif
    endif
    mem_loc_max=max(mem_loc,mem_loc_max)

    mem_loc = 0.
    if (istep_fe_twoD.gt.0) then
       !v7d,v7d_spec,g_rhs,gc_rhs,d_v_rhs,d_z_rhs,d_kperp_rhs,coll_rhs,p_rhs,c_rhs...
       !gfgamma012
       mem_loc = 43*lijklmn0*mem_cn
       !phi_term
       mem_loc = mem_loc + lij0*lzvwn0*mem_cn
       !cfgamma3
       mem_loc = mem_loc + lij0*lz0*ln0*mem_cn
    endif
    mem_loc_max=max(mem_loc,mem_loc_max)

    mem_loc = 0.
    if (istep_fe_transfer.gt.0) then
       !g_ls1,g_ls2,nlt_ls2_g,nlt_ls2_f, calc_transfer arrays
       mem_loc = 6*lijklmn0*mem_cn
       !f_ls2
       mem_loc = mem_loc + lij0*lzvwn0*mem_cn
       !fields_ls2
       mem_loc = mem_loc + lij0*lz0*ln0*mem_rn
       !to be considered?
       !allocate(fe_tf_tot(n_shells,n_shells,ln1:ln2,n_transfer_models))
       !allocate(fe_tot(n_shells,ln1:ln2,3,8))
       !allocate(k_value(1:n_shells))
    endif
    mem_loc_max=max(mem_loc,mem_loc_max)
    
    mem_est_diag_GyroLES = mem_loc_max+mem_req_in

  End Function mem_est_diag_GyroLES


  Subroutine initialize_all_diags_GyroLES

    IF (istep_fe_time.gt.0) call initialize_diag_fe_time
    IF (istep_fe_twoD.gt.0) call initialize_diag_spectral_twoD
    IF (istep_fe_transfer.gt.0) call initialize_diag_transfer

  End Subroutine initialize_all_diags_GyroLES

  Subroutine exec_all_diags_GyroLES(itime,time)

    Integer, intent(in) :: itime
    Real, intent(in) :: time
    logical :: no_aux_yet
    no_aux_yet = .true.

    if (istep_fe_time.gt.0) then
       if (itime.gt.0) then
          if (MODULO(itime,istep_fe_time) == 0) then
             if(arakawa_zv) then 
                call calc_aux_fields(g_1,emfields,f_,.true.,h_)  !Calculate f and h
             else
                call calc_aux_fields(g_1,emfields,f_,.false.)    !Calculate f
             endif
             no_aux_yet = .false.
             CALL diagnostic_fe_time(2)
          else if (MODULO(itime + 1,istep_fe_time) == 0) then
             if(arakawa_zv) then 
                call calc_aux_fields(g_1,emfields,f_,.true.,h_)  !Calculate f and h
             else
                call calc_aux_fields(g_1,emfields,f_,.false.)    !Calculate f
             endif
             no_aux_yet = .false.
             CALL diagnostic_fe_time(1)
          endif
       endif
    endif

    IF (istep_fe_twoD.gt.0.and.itime.gt.0) THEN
       if (modulo(itime,istep_fe_twoD) .eq. 0) then
          if(arakawa_zv) then 
             call calc_aux_fields(g_1,emfields,f_,.true.,h_)  !Calculate f and h
          else
             call calc_aux_fields(g_1,emfields,f_,.false.)    !Calculate f
          endif
          no_aux_yet = .false.
          call diag_spectral_twoD
       endif
    endif

    if (istep_fe_transfer.gt.0) then
         if (itime.gt.0) then    
            if (modulo(itime,istep_fe_transfer) .eq. 0) then
                !?call calc_aux_fields(g_1,emfields,f_,.true.,h_)  !Calculate h
                !?no_aux_yet = .false.
                call diagnostic_transfer
            endif
        endif
    endif

  End Subroutine exec_all_diags_GyroLES

  Subroutine finalize_all_diags_GyroLES

    IF (istep_fe_time.gt.0) call finalize_diag_fe_time
    IF (istep_fe_twoD.gt.0) call finalize_diag_spectral_twoD
    IF (istep_fe_transfer.gt.0) call finalize_diag_transfer

  End Subroutine finalize_all_diags_GyroLES



!!!**************************************************************************!!!
!!!******************** free energy balance (vs time)************************!!! 
!!!**************************************************************************!!!

  Subroutine initialize_diag_fe_time

    Implicit None

    Integer ::  n, o
    Character(len=20),dimension(3) :: time_file_spec_label

    time_file_spec_label = (/'/w_time_ ','/e_time_ ','/fe_time_'/)

    allocate(FE_TIME_FILE_SPEC(ln1:ln2,3))

    IF (mype.eq.0) then    
       call get_unit_nr(FE_TIME_FILE)
       OPEN(FE_TIME_FILE, file=trim(diagdir)//&
            &'/fe_time'//trim(file_extension),form = 'FORMATTED', &
            status=filestat, position=filepos)

       WRITE(FE_TIME_FILE,*) "#  1. time"
       WRITE(FE_TIME_FILE,*) "#  2. Free energy"
       WRITE(FE_TIME_FILE,*) "#  3. Drive"
       WRITE(FE_TIME_FILE,*) "#  4. Dissipation v"
       WRITE(FE_TIME_FILE,*) "#  5. Dissipation z"
       WRITE(FE_TIME_FILE,*) "#  6. Dissipation kperp"
       WRITE(FE_TIME_FILE,*) "#  7. Collision"
       WRITE(FE_TIME_FILE,*) "#  8. Parallel "        
       WRITE(FE_TIME_FILE,*) "#  9. Curvature "        
       WRITE(FE_TIME_FILE,*) "#  10. D FE/ dt (RHS)" 
       WRITE(FE_TIME_FILE,*) "#  11. D Fe /dt (LHS) "        
       WRITE(FE_TIME_FILE,*) "#  12. LHS - RHS"
       if (fracy.gt.0.and.nonlinear) then
100       format ('#',"        1",15I12)
          WRITE(FE_TIME_FILE,*) "#  13. subgrid"
          WRITE(FE_TIME_FILE,*) "#  14. filter fe drive"
          WRITE(FE_TIME_FILE,*) "#  15. filter fe dissipation"
          write(FE_TIME_FILE,100) 2,3,4,5,6,7,8,9,10,11,12,13,14,15
       else
110       format ('#',"        1",12I12)
          write(FE_TIME_FILE,110) 2,3,4,5,6,7,8,9,10,11,12
       endif
    Endif


    IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then    
       Do o=1,3
          DO n=ln1,ln2                
             call get_unit_nr(FE_TIME_FILE_SPEC(n,o))
             OPEN(FE_TIME_FILE_SPEC(n,o), file=trim(diagdir)//&
                  &trim(time_file_spec_label(o))//trim(spec(n)%name)//&
                  &trim(file_extension),form = 'FORMATTED', &
                  status=filestat, position=filepos)

             WRITE(FE_TIME_FILE_SPEC(n,o),*) "#  1. time"
             WRITE(FE_TIME_FILE_SPEC(n,o),*) "#  2. Free energy"
             WRITE(FE_TIME_FILE_SPEC(n,o),*) "#  3. Drive"
             WRITE(FE_TIME_FILE_SPEC(n,o),*) "#  4. Dissipation v"
             WRITE(FE_TIME_FILE_SPEC(n,o),*) "#  5. Dissipation z"
             WRITE(FE_TIME_FILE_SPEC(n,o),*) "#  6. Dissipation kperp"
             WRITE(FE_TIME_FILE_SPEC(n,o),*) "#  7. Collision"
             WRITE(FE_TIME_FILE_SPEC(n,o),*) "#  8. Parallel "        
             WRITE(FE_TIME_FILE_SPEC(n,o),*) "#  9. Curvature "        
190          format ('#',"        1",9I12)
             write(FE_TIME_FILE_SPEC(n,o),190) 2,3,4,5,6,7,8,9
          enddo
       enddo
    Endif

  End Subroutine initialize_diag_fe_time

!!!*****************************************************************************************!!!
  Subroutine diagnostic_fe_time(call_number)

    integer, intent(in) :: call_number
    complex,dimension(li1:li2,lj1:lj2,lk1:lk2):: e_temp
    complex,dimension(li1:li2,lj1:lj2,lk1:lk2,ln1:ln2):: e_n_temp
    complex, dimension(:,:,:),pointer :: ptr2
    complex, dimension(:,:,:,:),pointer :: ptr1, ptr3
    complex, dimension(:,:,:,:,:,:), pointer :: ptr4
    real :: rhs, lhs, fe_1, fe_2
    real,dimension(12) :: e_tot
    real,dimension(ln1:ln2,9,3) :: e_tot_n
    real,dimension(3) :: e_opt
    integer ::n,lbg1,lbg2,stage

    if(.not.(allocated(time_storage))) allocate(time_storage(2)) 
    if (.not.allocated(g_last))  allocate(g_last(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,ln1:ln2))
    if (.not.allocated(f_last))  allocate(f_last(li1:li2,lj1:lj2,lbz:ubz,lbv:ubv,lbw:ubw,ln1:ln2))
    !if (.not.allocated(h_last).and.arakawa_zv)  allocate(h_last(li1:li2,lj1:lj2,lbz:ubz,lbv:ubv,lbw:ubw,ln1:ln2))
    if (.not.allocated(emfields_last))  allocate(emfields_last(li1:li2,lj1:lj2,lbz:ubz,1:n_fields))

    if(call_number.eq.1) then
       g_last = g_1
       f_last = f_
       !if (arakawa_zv) h_last = h_
       emfields_last = emfields
       time_storage(1) = time
    endif

    if (call_number.eq.2) then

       Allocate(g_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(gc_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(d_v_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(d_z_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(d_kperp_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(coll_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(c_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(p_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(fe_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(fe_g_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(fe_d_v_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(fe_d_z_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(fe_d_kperp_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(fe_coll_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(fe_c_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(fe_p_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(fe_rhs_n(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2))
       Allocate(fe_g_rhs_n(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2))
       Allocate(fe_d_v_rhs_n(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2))
       Allocate(fe_d_z_rhs_n(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2))
       Allocate(fe_d_kperp_rhs_n(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2))
       Allocate(fe_coll_rhs_n(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2))
       Allocate(fe_c_rhs_n(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2))
       Allocate(fe_p_rhs_n(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2))
       Allocate(fe_lhs_1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(fe_lhs_2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(cfgamma(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(cfgamma_last(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(cfgamma1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(cfgamma2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       Allocate(cfgamma3(li1:li2, lj1:lj2, lk1:lk2, ln1:ln2))
          
       if (fracy.gt.0.and.nonlinear) then
          Allocate(filter_g_1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
          Allocate(filter_f_(lbi:ubi, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2))   
          Allocate(filter_emfields(li1:li2,lj1:lj2,lbz:ubz,1:n_fields))
          Allocate(nlt(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
          Allocate(filter_nlt (li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))     
          Allocate(nlt_filter_g_1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))  
          Allocate(filter_nlt_filter_g_1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))   
          Allocate(t_fgs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
       endif

       time_storage(2) = time
       ptr1 => null()
       ptr2 => null()
       ptr3 => null()
       ptr4 => null()
       lbg1 = 1 
       lbg2 = lklmn0  
       stage = 0

       e_tot = 0.0
       e_tot_n = 0.0

       call get_cfgamma_f(f_,emfields,cfgamma)  !compute f
       call get_cfgamma(f_,emfields,cfgamma1,cfgamma2)

       d_v_rhs = cmplx(0.0,0.0)
       fe_d_v_rhs = cmplx(0.0,0.0)
       fe_d_v_rhs_n = cmplx(0.0,0.0)
          
       if (arakawa_zv) then 
          if (hyp_on_h) then 
             call add_hypv_ak(h_,d_v_rhs,lbg1,lbg2)
          else
             call add_hypv_ak(f_,d_v_rhs,lbg1,lbg2)
          endif
       else
           call add_hypv_block(f_,d_v_rhs,lbg1,lbg2)
       endif


       d_z_rhs = cmplx(0.0,0.0)
       fe_d_z_rhs = cmplx(0.0,0.0)
       fe_d_z_rhs_n = cmplx(0.0,0.0)
       
       if (arakawa_zv) then
          if (hyp_on_h) then 
             call add_hypz_ak(h_,d_z_rhs,lbg1,lbg2)
             if (hypz_compensation) call equ_comp_hypz(emfields,ptr4,d_z_rhs,lbg1,lbg2)
          else  
             call add_hypz_ak(f_,d_z_rhs,lbg1,lbg2)
          endif  
       else
           call add_hypz_block(f_,d_z_rhs,lbg1,lbg2)
       endif

       d_kperp_rhs = cmplx(0.0,0.0)
       fe_d_kperp_rhs = cmplx(0.0,0.0)
       fe_d_kperp_rhs_n = cmplx(0.0,0.0)
       if ((hyp_x.gt.0.0).or.(hyp_y.gt.0.0).or.(hyp_perp.gt.0.0).or.(GyroLES)) then !same here
          !/todo:check if hyp_on_h should affect perpendicular hyperdiffusions!
          !if (hyp_on_h) then
          !   call add_dfdxy(h_,d_kperp_rhs,lbg1,lbg2) 
          !else
             call add_dfdxy(f_,d_kperp_rhs,lbg1,lbg2) 
          !endif
       endif

       gc_rhs = cmplx(0.0,0.0)
       call add_dgdxy(g_1, gc_rhs, ptr1, pdg1di, pdg1dj,lbg1,lbg2)
       call add_dchidxy_orig(emfields,ptr4, gc_rhs,ptr2,ptr3,lbg1,lbg2,.true.)

       g_rhs  = cmplx(0.0,0.0)
       fe_g_rhs  = cmplx(0.0,0.0)
       fe_g_rhs_n  = cmplx(0.0,0.0)
       call add_dchidxy_orig(emfields,ptr4, g_rhs,ptr2,ptr3,lbg1,lbg2,.false.)
       c_rhs = cmplx(0.0,0.0)
       fe_c_rhs = cmplx(0.0,0.0)
       fe_c_rhs_n = cmplx(0.0,0.0)
       c_rhs = gc_rhs - g_rhs   

       p_rhs = cmplx(0.0,0.0)
       fe_p_rhs = cmplx(0.0,0.0)
       fe_p_rhs_n = cmplx(0.0,0.0)

       !Parallel term goes always with h
       if (arakawa_zv) then
           call equ_dzv(h_,p_rhs,lbg1,lbg2)
           if (hyp_on_h) then
              if (hypz_compensation) call equ_comp_hypz(emfields,ptr4,p_rhs,lbg1,lbg2)
           endif
       else
           call equ_dfdzv(f_,p_rhs,lbg1,lbg2)
           call add_dchidz(emfields, ptr4, p_rhs, lbg1, lbg2)
       end if

       fe_coll_rhs = cmplx(0.0,0.0)                 
       fe_coll_rhs_n = cmplx(0.0,0.0)                 
       if (collision_op.ne.'none') then
          !collisions on f_
          call equ_collisions(f_,coll_rhs,replace_rhs=.true.)
       endif

       !Term 1 Time
       e_tot(1) = time
       Do n=ln1,ln2
          e_tot_n(n,1,1) = time
          e_tot_n(n,1,2) = time
          e_tot_n(n,1,3) = time
       Enddo


       !Term 2 Free energy  
       fe_rhs = cmplx(0.0,0.0)    
       e_temp=cmplx(0.0,0.0)
       fe_rhs = cfgamma*g_1/2.
       Call energy_integral(fe_rhs,e_temp)
       if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
       call sum_int_3d(e_temp,e_tot(2))

       ! Entropy
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_rhs_n(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*g_1(:,:,:,:,:,n)/2.
          Call integral_vpw(fe_rhs_n(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,2,1))
       Enddo


       if (n_spec.eq.1) then
          ! Electrostatic
          Do n=ln1,ln2
             e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
             fe_rhs_n(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*g_1(:,:,:,:,:,n)/2.
             Call integral_vpw(fe_rhs_n(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
             if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
             call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,2,2))
          Enddo

       else
          ! Electrostatic with modification
          call get_electrostatic(emfields,cfgamma3)
          Do n=ln1,ln2
             if(p_has_00_mode) cfgamma3(0,0,:,n)=cmplx(0.0,0.0)
             call sum_int_3d(cfgamma3(:,:,:,n),e_tot_n(n,2,2))
          Enddo
       endif 
          
       ! Total
       Do n =ln1,ln2
          e_tot_n(n,2,3) = e_tot_n(n,2,1) + e_tot_n(n,2,2) 
       Enddo

       !Term 3 Drive
       e_temp=cmplx(0.0,0.0)
       fe_g_rhs = cfgamma*g_rhs
       Call energy_integral(fe_g_rhs,e_temp)
       if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
       call sum_int_3d(e_temp,e_tot(3))

       ! Entropy
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_g_rhs_n(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*g_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_g_rhs_n(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,3,1))
       Enddo

       ! Electrostatic
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_g_rhs_n(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*g_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_g_rhs_n(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,3,2))
       Enddo

       ! Total
       Do n =ln1,ln2
          e_tot_n(n,3,3) = e_tot_n(n,3,1) + e_tot_n(n,3,2) 
       Enddo

       !Term 4 Dissipation  v  
       e_temp=cmplx(0.0,0.0)
       fe_d_v_rhs  = cfgamma*d_v_rhs
       Call energy_integral(fe_d_v_rhs,e_temp)
       if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
       call sum_int_3d(e_temp,e_tot(4))


       ! Entropy
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_d_v_rhs_n(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*d_v_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_d_v_rhs_n(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,4,1))
       Enddo

       ! Electrostatic
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_d_v_rhs_n(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*d_v_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_d_v_rhs_n(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,4,2))
       Enddo

       ! Total
       Do n =ln1,ln2
          e_tot_n(n,4,3) = e_tot_n(n,4,1) + e_tot_n(n,4,2) 
       Enddo

       !Term 5 Dissipation  z  
       e_temp=cmplx(0.0,0.0)
       fe_d_z_rhs  = cfgamma*d_z_rhs
       Call energy_integral(fe_d_z_rhs,e_temp)
       if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
       call sum_int_3d(e_temp,e_tot(5))

       ! Entropy
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_d_z_rhs_n(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*d_z_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_d_z_rhs_n(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,5,1))
       Enddo

       ! Electrostatic
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_d_z_rhs_n(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*d_z_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_d_z_rhs_n(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,5,2))
       Enddo

       ! Total
       Do n =ln1,ln2
          e_tot_n(n,5,3) = e_tot_n(n,5,1) + e_tot_n(n,5,2) 
       Enddo

       !Term 6 Dissipation  kperp  
       e_temp = cmplx(0.0,0.0)
       fe_d_kperp_rhs  = cfgamma*d_kperp_rhs
       Call energy_integral(fe_d_kperp_rhs,e_temp)
       if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
       call sum_int_3d(e_temp,e_tot(6))

       ! Entropy
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_d_kperp_rhs_n(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*d_kperp_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_d_kperp_rhs_n(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,6,1))
       Enddo

       ! Electrostatic
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_d_kperp_rhs_n(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*d_kperp_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_d_kperp_rhs_n(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,6,2))
       Enddo

       ! Total
       Do n =ln1,ln2
          e_tot_n(n,6,3) = e_tot_n(n,6,1) + e_tot_n(n,6,2) 
       Enddo

       !Term 7  collision     
       if (collision_op.ne.'none') then 
          e_temp=cmplx(0.0,0.0)
          fe_coll_rhs = cfgamma*coll_rhs
          Call energy_integral(fe_coll_rhs,e_temp)
          if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
          call sum_int_3d(e_temp,e_tot(7))
       else
          e_tot(7) = 0.0
       endif

       if (collision_op.ne.'none') then 
          ! Entropy
          Do n=ln1,ln2
             e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
             fe_coll_rhs_n(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*coll_rhs(:,:,:,:,:,n)
             Call integral_vpw(fe_coll_rhs_n(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
             if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
             call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,7,1))
          Enddo
          !    
          ! Electrostatic
          Do n=ln1,ln2
             e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
             fe_coll_rhs_n(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*coll_rhs(:,:,:,:,:,n)
             Call integral_vpw(fe_coll_rhs_n(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
             if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
             call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,7,2))
          Enddo

          ! Total
          Do n =ln1,ln2
             e_tot_n(n,7,3) = e_tot_n(n,7,1) + e_tot_n(n,7,2) 
          Enddo

       else
          Do n =ln1,ln2
             e_tot_n(n,7,1) = 0  
             e_tot_n(n,7,2) = 0  
             e_tot_n(n,7,3) = 0 
          enddo
       endif


       !Term 8 parallel  
       e_temp=cmplx(0.0,0.0)
       fe_p_rhs = cfgamma*p_rhs
       Call energy_integral(fe_p_rhs,e_temp)
       if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
       call sum_int_3d(e_temp,e_tot(8))
       !Substract hypper diffusions
       if (hyp_on_h) then
          e_tot(8) = e_tot(8) - e_tot(4) - e_tot(5)
       else
          if (.not.arakawa_zv) e_tot(8)= e_tot(8)-e_tot(4)-e_tot(5) !eliminated hyperdiffusions
       endif

       ! Entropy
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_p_rhs_n(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*p_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_p_rhs_n(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,8,1))
          !Substract hypper diffusions
          if (hyp_on_h) then
             e_tot_n(n,8,1) = e_tot_n(n,8,1) - e_tot_n(n,4,1) - e_tot_n(n,5,1)
          else
             if (.not.arakawa_zv) e_tot_n(n,8,1)= e_tot_n(n,8,1)-e_tot_n(n,4,1)-e_tot_n(n,5,1) !eliminated hyperdiffusions
          endif
       Enddo
       ! Electrostatic
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_p_rhs_n(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*p_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_p_rhs_n(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,8,2))
          !Substract hypper diffusions
          if (hyp_on_h) then
             e_tot_n(n,8,2) = e_tot_n(n,8,2) - e_tot_n(n,4,2) - e_tot_n(n,5,2)
          else
             if (.not.arakawa_zv) e_tot_n(n,8,2)= e_tot_n(n,8,2)-e_tot_n(n,4,2)-e_tot_n(n,5,2) !eliminated hyperdiffusions
          endif
       Enddo

       ! Total
       Do n =ln1,ln2
          e_tot_n(n,8,3) = e_tot_n(n,8,1) + e_tot_n(n,8,2) 
       Enddo


       !Term 9 curvature  
       e_temp =cmplx(0.0,0.0)
       fe_c_rhs  = cfgamma*c_rhs
       Call energy_integral(fe_c_rhs,e_temp)
       if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
       call sum_int_3d(e_temp,e_tot(9))
       if (magn_geometry.eq.'slab') e_tot(9) = 0.0

       ! Entropy
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_c_rhs_n(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*c_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_c_rhs_n(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,9,1))
       Enddo

       ! Electrostatic
       Do n=ln1,ln2
          e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
          fe_c_rhs_n(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*c_rhs(:,:,:,:,:,n)
          Call integral_vpw(fe_c_rhs_n(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
          if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
          call sum_int_3d(e_n_temp(:,:,:,n),e_tot_n(n,9,2))
       Enddo

       ! Total
       Do n =ln1,ln2
          e_tot_n(n,9,3) = e_tot_n(n,9,1) + e_tot_n(n,9,2) 
       Enddo


       ! Term 10: RHS, 11: LHS and 12: LHS - RHS 
       !if (arakawa_zv) then
       !   call get_cfgamma_h(h_last,cfgamma_last)
       !else
          call get_cfgamma_f(f_last,emfields_last,cfgamma_last)
       !endif


       fe_lhs_1 = cmplx (0.0,0.0)
       e_temp  = cmplx(0.0)
       fe_1 = 0.0
       fe_lhs_1 =cfgamma_last*g_last/2.
       Call energy_integral(fe_lhs_1,e_temp)
       if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
       call sum_int_3d(e_temp,fe_1)

       fe_lhs_2 = cmplx (0.0,0.0)
       e_temp = cmplx(0.0)
       fe_2 = 0.0
       fe_lhs_2 = cfgamma*g_1/2.
       Call energy_integral(fe_lhs_2,e_temp)
       if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
       call sum_int_3d(e_temp,fe_2)

       lhs = 0.0
       rhs = 0.0
       lhs  = (fe_2 - fe_1)/(time_storage(2) - time_storage(1))
       rhs  = e_tot(3) + e_tot(4) + e_tot(5) + e_tot(6) + e_tot(7) + e_tot(8) + e_tot(9)
       e_tot(10) = rhs
       e_tot(11) = lhs
       e_tot(12) = lhs - rhs


       !Terms 13,14,15 sub-grids contribution (optional)
       if  (fracy.gt.0.and.nonlinear)  then
          e_opt = 0.0  
          e_temp = cmplx(0.0,0.0)
          Call filter(g_1, filter_g_1)
          call calc_aux_fields(filter_g_1,filter_emfields,filter_f_,comp_h=.false.)

          nlt = cmplx(0.0,0.0)
          CALL this_nonlinear_term%add(g_1, ptr1, emfields,&
               &ptr2, ptr3, nlt, lbg1, lbg2, stage)

          Call filter(nlt, filter_nlt)

          nlt_filter_g_1 = cmplx(0.0,0.0)
          CALL this_nonlinear_term%add(filter_g_1, ptr1, filter_emfields,&
               &ptr2, ptr3, nlt_filter_g_1, lbg1, lbg2, stage)
          Call filter(nlt_filter_g_1, filter_nlt_filter_g_1)     

          t_fgs = filter_nlt - filter_nlt_filter_g_1
          call filter(t_fgs,t_fgs)
          call filter(cfgamma,cfgamma)

          fe_rhs = cmplx(0.0,0.0)
          fe_rhs  = cfgamma*t_fgs
          Call energy_integral(fe_rhs,e_temp)
          call sum_int_3d(e_temp,e_opt(1))

          fe_rhs  = cmplx(0.0,0.0)
          fe_rhs  = cfgamma*g_rhs
          Call energy_integral(fe_rhs,e_temp)
          if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
          call sum_int_3d(e_temp,e_opt(2))

          fe_rhs = cmplx(0.0,0.0)
          fe_rhs   = cfgamma*(d_z_rhs + d_v_rhs + d_kperp_rhs + fe_coll_rhs)
          Call energy_integral(fe_rhs,e_temp)
          if(p_has_00_mode) e_temp(0,0,:)=cmplx(0.0,0.0)
          call sum_int_3d(e_temp,e_opt(3))

       endif

       IF (mype.eq.0) then    
          if (fracy.gt.0.and.nonlinear) then
             write(FE_TIME_FILE,"(15ES12.4)") e_tot, e_opt
          else
             WRITE(FE_TIME_FILE,"(12ES12.4)") e_tot
          endif
         call flush(FE_TIME_FILE)
       Endif


       IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then    
          DO n=ln1,ln2
             write(FE_TIME_FILE_SPEC(n,1),"(9ES12.4)") e_tot_n(n,:,1)
             write(FE_TIME_FILE_SPEC(n,2),"(9ES12.4)") e_tot_n(n,:,2) 
             write(FE_TIME_FILE_SPEC(n,3),"(9ES12.4)") e_tot_n(n,:,3) 
            call flush(FE_TIME_FILE_SPEC(n,1))
            call flush(FE_TIME_FILE_SPEC(n,2))
            call flush(FE_TIME_FILE_SPEC(n,3))
          enddo
       Endif


       Deallocate(time_storage)    
       Deallocate(g_last)
       Deallocate(f_last)
       !Deallocate(h_last)
       Deallocate(emfields_last)
       Deallocate(g_rhs)
       Deallocate(gc_rhs)
       Deallocate(d_v_rhs)
       Deallocate(d_z_rhs)
       Deallocate(d_kperp_rhs)
       Deallocate(coll_rhs)
       Deallocate(c_rhs)
       Deallocate(p_rhs)
       Deallocate(fe_rhs)
       Deallocate(fe_rhs_n)
       Deallocate(fe_g_rhs)
       Deallocate(fe_g_rhs_n)
       Deallocate(fe_d_v_rhs)
       Deallocate(fe_d_v_rhs_n)
       Deallocate(fe_d_z_rhs)
       Deallocate(fe_d_z_rhs_n)
       Deallocate(fe_d_kperp_rhs)
       Deallocate(fe_d_kperp_rhs_n)
       Deallocate(fe_coll_rhs)
       Deallocate(fe_coll_rhs_n)
       Deallocate(fe_c_rhs)
       Deallocate(fe_c_rhs_n)
       Deallocate(fe_p_rhs)
       Deallocate(fe_p_rhs_n)
       Deallocate(fe_lhs_1)
       Deallocate(fe_lhs_2)
       Deallocate(cfgamma,cfgamma_last, cfgamma1, cfgamma2, cfgamma3)
       
       if (fracy.gt.0.and.nonlinear) then
          Deallocate(filter_g_1)
          Deallocate(filter_f_)
          Deallocate(filter_emfields)
          Deallocate(nlt)
          Deallocate(filter_nlt)
          Deallocate(nlt_filter_g_1)
          Deallocate(filter_nlt_filter_g_1)
          Deallocate(t_fgs)
       endif

    endif
  End Subroutine diagnostic_fe_time

!!!************************************************************************************!!!

  Subroutine finalize_diag_fe_time  

    Integer :: n

    IF (mype.eq.0) then    
       Close (FE_TIME_FILE)
    endif

    IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then    
       do n=ln1,ln2 
          Close (FE_TIME_FILE_SPEC(n,1))
          Close (FE_TIME_FILE_SPEC(n,2))
          Close (FE_TIME_FILE_SPEC(n,3))
       enddo
    endif

    Deallocate(FE_TIME_FILE_SPEC)
  End Subroutine finalize_diag_fe_time

  !******************************************************************************************
  !************************************ entropy/electrostatic balance ***********************
  !******************************************************************************************

  !******************************************************************************************************
  !**************************************  Diag 2D ******************************************************
  !*******************************************************************************************************

  Subroutine initialize_diag_spectral_twoD

    Character(len=20), dimension(8) :: label
    integer :: n,m,i,j,nx_index
    real, dimension(0:nx0-1) :: kxglob

    ! Arrays to cumulate for time average
    allocate(cumul_kxky(li1:li2,lj1:lj2,1:8))
    allocate(twoD_counter(8))
    allocate(cumul_kxky_spec(li1:li2,lj1:lj2,ln1:ln2,1:24))
    allocate(fnumber_spec(ln1:ln2,1:24))
    allocate(twoD_counter_spec(ln1:ln2,1:24))

    fnumber=0
    fnumber_spec=0
    twoD_counter=0
    twoD_counter_spec=0
    cumul_kxky=0
    cumul_kxky_spec=0


    ! We initialize the label to be appended to the name of file
    label(1)='fe'
    label(2)='drive'
    label(3)='hyp_v'
    label(4)='hyp_z'
    label(5)='hyp_kperp'
    label(6)='coll'
    label(7)='parallel'
    label(8)='curvature'

    ! We will write kx and ky in the header of the file 	    
    ! We create kxglob without hkx+1 mode and invert it	    

    if (evenx.eq.1) then
    kxglob(0:nx0/2-2)=kx(nx0/2+1:nx0-1)
    kxglob(nx0/2-1:nx0-1)=kx(0:nx0/2)
    nx_index =  nx0-2
    
    else
    kxglob(0:nx0/2-1)=kx(nx0/2+1:nx0-1)
    kxglob(nx0/2:nx0-1)=kx(0:nx0/2)
    nx_index =  nx0-1
    endif

    if (mype.eq.0) then
       do m=1,8
          call get_unit_nr(fnumber(m))
          OPEN(fnumber(m), file=trim(diagdir)//'/Spectral_2D_'&
               //trim(adjustl(label(m)))//trim(file_extension),form = 'FORMATTED', &
               status=filestat, position=filepos)         

          ! We can directly write preliminary numbers as headers	    
          
          Do i=0,nx_index
             write(fnumber(m),"(ES12.4)") kxglob(i)
          enddo
          Do  j=0,nky0-1     
             write(fnumber(m),"(ES12.4)") ky(j)
          enddo
       enddo

    endif

    If (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
        do n=ln1,ln2
            do m=1,8
                call get_unit_nr(fnumber_spec(n,m))
                OPEN(fnumber_spec(n,m), file=trim(diagdir)//'/Spectral_2D_W_'&
                    //trim(adjustl(label(m)))//'_'//trim(spec(n)%name)//trim(file_extension),form = 'FORMATTED', &
                    status=filestat, position=filepos)         

                ! We can directly write preliminary numbers as headers	    
                Do i=0,nx_index
                    write(fnumber_spec(n,m),"(ES12.4)") kxglob(i)
                enddo
                Do  j=0,nky0-1     
                    write(fnumber_spec(n,m),"(ES12.4)") ky(j)
                enddo
            enddo
        enddo
        do n=ln1,ln2
            do m=9,16
                call get_unit_nr(fnumber_spec(n,m))
                OPEN(fnumber_spec(n,m), file=trim(diagdir)//'/Spectral_2D_E_'&
                    //trim(adjustl(label(m-8)))//'_'//trim(spec(n)%name)//trim(file_extension),form = 'FORMATTED', &
                    status=filestat, position=filepos)         

                ! We can directly write preliminary numbers as headers	    
                Do i=0,nx_index
                    write(fnumber_spec(n,m),"(ES12.4)") kxglob(i)
                enddo
                Do  j=0,nky0-1     
                    write(fnumber_spec(n,m),"(ES12.4)") ky(j)
                enddo
            enddo
        enddo
        do n=ln1,ln2
            do m=17,24
                call get_unit_nr(fnumber_spec(n,m))
                OPEN(fnumber_spec(n,m), file=trim(diagdir)//'/Spectral_2D_FE_'&
                    //trim(adjustl(label(m-16)))//'_'//trim(spec(n)%name)//trim(file_extension),form = 'FORMATTED', &
                    status=filestat, position=filepos)         

                ! We can directly write preliminary numbers as headers	    
                Do i=0,nx_index
                    write(fnumber_spec(n,m),"(ES12.4)") kxglob(i)
                enddo
                Do  j=0,nky0-1     
                    write(fnumber_spec(n,m),"(ES12.4)") ky(j)
                enddo
            enddo
        enddo
    endif

  End Subroutine initialize_diag_spectral_twoD

  Subroutine finalize_diag_spectral_twoD

    Implicit None
    Integer:: n,m

    if (mype.eq.0) then
       do m=1,8
          close(fnumber(m))
       enddo
    endif

    If (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
        do n=ln1,ln2
            do m=1,24
                close(fnumber_spec(n,m))
            enddo
        enddo
    endif

    deallocate(fnumber_spec)
    deallocate(twoD_counter)
    deallocate(twoD_counter_spec)
    deallocate(cumul_kxky)
    deallocate(cumul_kxky_spec)

  End Subroutine finalize_diag_spectral_twoD


!!!******************************************************************************************
!!! Subroutine that goes from 6d to 2d     ***************************************************
!!!*******************************************************************************************

  Subroutine diag_spectral_twoD

    implicit none

    real, dimension(li1:li2,lj1:lj2,1:8) :: v3d
    real, dimension(li1:li2,lj1:lj2,ln1:ln2,1:24) :: v4d
    integer :: m,n,lbg1,lbg2
    complex, dimension(:,:,:),pointer :: ptr2
    complex, dimension(:,:,:,:),pointer :: ptr1, ptr3
    complex, dimension(:,:,:,:,:,:), pointer :: ptr4

    Allocate(v7d(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2,ln1:ln2,1:8)) 
    Allocate(v7d_spec(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2,ln1:ln2,1:24)) 
    Allocate(g_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(gc_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(d_v_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(d_z_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(d_kperp_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(coll_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(p_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(c_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(cfgamma(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(cfgamma1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(cfgamma2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(cfgamma3(li1:li2, lj1:lj2, lk1:lk2, ln1:ln2))
       

    ptr1 => null()
    ptr2 => null()
    ptr3 => null()
    ptr4 => null()
    lbg1 = 1 
    lbg2 = lklmn0  
    v7d=cmplx(0.0,0.0)
    v7d_spec=cmplx(0.0,0.0)
    v4d=0.0
    v3d=0.0

    if (arakawa_zv) then
       call get_cfgamma_h(h_,cfgamma)
       call get_cfgamma(f_,emfields,cfgamma1,cfgamma2)
    else
       call get_cfgamma_f(f_,emfields,cfgamma) !because we compute f
       call get_cfgamma(f_,emfields,cfgamma1,cfgamma2)
    endif

    d_v_rhs = cmplx(0.0,0.0)
    if (arakawa_zv) then 
       if (hyp_on_h) then 
          call add_hypv_ak(h_,d_v_rhs,lbg1,lbg2)
       else
          call add_hypv_ak(f_,d_v_rhs,lbg1,lbg2)
       endif
    else
        call add_hypv_block(f_,d_v_rhs,lbg1,lbg2)
    endif

    d_z_rhs = cmplx(0.0,0.0)
    if (arakawa_zv) then
       if (hyp_on_h) then 
          call add_hypz_ak(h_,d_z_rhs,lbg1,lbg2)
          if (hypz_compensation) call equ_comp_hypz(emfields,ptr4,d_z_rhs,lbg1,lbg2)
       else  
          call add_hypz_ak(f_,d_z_rhs,lbg1,lbg2)
       endif  
    else
        call add_hypz_block(f_,d_z_rhs,lbg1,lbg2)
    endif

    d_kperp_rhs = cmplx(0.0,0.0)
    if ((hyp_x.gt.0.0).or.(hyp_y.gt.0.0).or.(hyp_perp.gt.0.0).or.(GyroLES)) then
       !/todo:check if hyp_on_h should affect perpendicular hyperdiffusions!
       !if (hyp_on_h) then
       !   call add_dfdxy(h_,d_kperp_rhs,lbg1,lbg2) !f_in is f
       !else
          call add_dfdxy(f_,d_kperp_rhs,lbg1,lbg2) !f_in is f
       !endif
    endif

    !Collisions works with f
    if (collision_op.ne.'none') then
       call equ_collisions(f_,coll_rhs,replace_rhs=.true.)
    endif


    !it contains curvature and gradient term  
    gc_rhs = cmplx(0.0,0.0)
    call add_dgdxy(g_1, gc_rhs, ptr1, pdg1di, pdg1dj,lbg1,lbg2)
    call add_dchidxy_orig(emfields,ptr4, gc_rhs,ptr2,ptr3,lbg1,lbg2,.true.)

    !only includes the gradient term, without the  curvature (= false)   
    g_rhs  = cmplx(0.0,0.0)
    call add_dchidxy_orig(emfields,ptr4, g_rhs,ptr2,ptr3,lbg1,lbg2,.false.)

    !Calculation of the term 8th term, the curvature term
    !substract the gradient contribution  
    c_rhs = cmplx(0.0,0.0)
    c_rhs = gc_rhs - g_rhs   

    !Calculation of the term 7th term, the parallel term
    p_rhs = cmplx(0.0,0.0)

    !Parallel term goes always with h
    if (arakawa_zv) then
       call equ_dzv(h_,p_rhs,lbg1,lbg2)
       if (hyp_on_h) then
          if (hypz_compensation) call equ_comp_hypz(emfields,ptr4,p_rhs,lbg1,lbg2)
       endif
    else
       call equ_dfdzv(f_,p_rhs,lbg1,lbg2)
       call add_dchidz(emfields, ptr4, p_rhs, lbg1, lbg2)
    end if
    ! Then we can go on in usual conditions

    !Free energy / 1
    v7d(:,:,:,:,:,:,1) = cfgamma*g_1/2.
    v7d_spec(:,:,:,:,:,:,1) = cfgamma1*g_1/2.
    v7d_spec(:,:,:,:,:,:,9) = cfgamma2*g_1/2.
    v7d_spec(:,:,:,:,:,:,17) = v7d(:,:,:,:,:,:,1)

    !Drive  / 2  
    v7d(:,:,:,:,:,:,2) = cfgamma*g_rhs
    v7d_spec(:,:,:,:,:,:,2) = cfgamma1*g_rhs
    v7d_spec(:,:,:,:,:,:,10) = cfgamma2*g_rhs
    v7d_spec(:,:,:,:,:,:,18) = v7d(:,:,:,:,:,:,2)

    !Dissipation  v / 3
    v7d(:,:,:,:,:,:,3) = cfgamma*d_v_rhs
    v7d_spec(:,:,:,:,:,:,3) = cfgamma1*d_v_rhs
    v7d_spec(:,:,:,:,:,:,11) = cfgamma2*d_v_rhs
    v7d_spec(:,:,:,:,:,:,19) = v7d(:,:,:,:,:,:,3)

    !Dissipation  z / 4
    v7d(:,:,:,:,:,:,4) = cfgamma*d_z_rhs
    v7d_spec(:,:,:,:,:,:,4) = cfgamma1*d_z_rhs
    v7d_spec(:,:,:,:,:,:,12) = cfgamma2*d_z_rhs
    v7d_spec(:,:,:,:,:,:,20) = v7d(:,:,:,:,:,:,4)

    !Dissipation  kperp / 5 
    v7d(:,:,:,:,:,:,5) = cfgamma*d_kperp_rhs
    v7d_spec(:,:,:,:,:,:,5) = cfgamma1*d_kperp_rhs
    v7d_spec(:,:,:,:,:,:,13) = cfgamma2*d_kperp_rhs
    v7d_spec(:,:,:,:,:,:,21) = v7d(:,:,:,:,:,:,5)

    !Collision / 6      
    if (collision_op.ne.'none') then
       v7d(:,:,:,:,:,:,6) = cfgamma*coll_rhs
       v7d_spec(:,:,:,:,:,:,6) = cfgamma1*coll_rhs
       v7d_spec(:,:,:,:,:,:,14) = cfgamma2*coll_rhs
       v7d_spec(:,:,:,:,:,:,22) = v7d(:,:,:,:,:,:,6)
    endif

    !parallel / 7      
    v7d(:,:,:,:,:,:,7) = cfgamma*p_rhs
    v7d_spec(:,:,:,:,:,:,7) = cfgamma1*p_rhs
    v7d_spec(:,:,:,:,:,:,15) = cfgamma2*p_rhs
    v7d_spec(:,:,:,:,:,:,23) = v7d(:,:,:,:,:,:,7)

    !curvature / 8      
    v7d(:,:,:,:,:,:,8) = cfgamma*p_rhs
    v7d_spec(:,:,:,:,:,:,8) = cfgamma1*c_rhs
    v7d_spec(:,:,:,:,:,:,16) = cfgamma2*c_rhs
    v7d_spec(:,:,:,:,:,:,24) = v7d(:,:,:,:,:,:,8)


    ! At this point, we have in v7d all relevant values that we will have to write : quantities
    ! We have now to issue the 2D array and write the files
    deallocate(g_rhs)
    deallocate(gc_rhs)
    deallocate(d_v_rhs)
    deallocate(d_z_rhs)
    deallocate(d_kperp_rhs)
    deallocate(coll_rhs)
    deallocate(p_rhs)
    deallocate(c_rhs)
    deallocate(cfgamma)
    deallocate(cfgamma1)
    deallocate(cfgamma2)
 
    do m=1,8
       call twoD_array(v7d(:,:,:,:,:,:,m),v3d(:,:,m))
       ! We do proceed with cumulating if condition is verified
       if (itime.gt.0 .and. .not.( mod(itime,istep_fe_twoD*avg_window) .eq. 0) ) then
          cumul_kxky(:,:,m)=cumul_kxky(:,:,m)+v3d(:,:,m)     
          twoD_counter(m) = twoD_counter(m) +1   
       endif
       ! Then we do output files
       if  (itime.gt.0 .and. mod(itime,istep_fe_twoD*avg_window) .eq. 0 ) then
          cumul_kxky(:,:,m)=cumul_kxky(:,:,m)+v3d(:,:,m)     
          twoD_counter(m) = twoD_counter(m) +1   
          v3d(:,:,m)=cumul_kxky(:,:,m)/real(twoD_counter(m))
          call write2d(v3d(:,:,m),m)
          cumul_kxky(:,:,m) = 0
          twoD_counter(m)=0
       endif ! itime > 0 / twoD_counter...
    enddo ! loop over quant

    do n=ln1,ln2
        do m=1,24
            if ((m.eq.9).and.(n_spec.gt.1)) then
                call get_electrostatic(emfields,cfgamma3)
                call sum_int_z(real(cfgamma3(:,:,:,n)),v4d(:,:,n,m))
            else if ((m.eq.17).and.(n_spec.gt.1)) then
                v4d(:,:,n,m) = v4d(:,:,n,1) + v4d(:,:,n,9)
            else
                call twoD_array_spec(v7d_spec(:,:,:,:,:,n,m),v4d(:,:,n,m))
            endif
            ! We do proceed with cumulating if condition is verified
            if (itime.gt.0 .and. .not.( mod(itime,istep_fe_twoD*avg_window) .eq. 0) ) then
            cumul_kxky_spec(:,:,n,m)=cumul_kxky_spec(:,:,n,m)+v4d(:,:,n,m)     
            twoD_counter_spec(n,m) = twoD_counter_spec(n,m) +1   
            endif
            ! Then we do output files
            if  (itime.gt.0 .and. mod(itime,istep_fe_twoD*avg_window) .eq. 0 ) then
                cumul_kxky_spec(:,:,n,m)=cumul_kxky_spec(:,:,n,m)+v4d(:,:,n,m)     
                twoD_counter_spec(n,m) = twoD_counter_spec(n,m) +1   
                v4d(:,:,n,m)=cumul_kxky_spec(:,:,n,m)/real(twoD_counter_spec(n,m))
                call write2d_spec(v4d(:,:,n,m),n,m)
                cumul_kxky_spec(:,:,n,m) = 0
                twoD_counter_spec(n,m)=0
            endif ! itime > 0 / twoD_counter...
        enddo ! loop over quant
    enddo ! loop over species

    deallocate(v7d)
    deallocate(v7d_spec)
    deallocate(cfgamma3)
 
  End subroutine diag_spectral_twoD


!!!**************************************************************************!!!
!!!********************* transfer functions analysis ************************!!! 
!!!**************************************************************************!!!

  Subroutine initialize_diag_transfer

    real :: kperp_maxloc
    integer :: i, j, k, ierr

    allocate(FE_TF_SPEC_3D_FILE(ln1:ln2,3,n_transfer_models))
    Allocate(FE_TF_FILE(ln1:ln2,n_transfer_models,n_transfer_models)) 
    Allocate(FE_SP_FILE(ln1:ln2,3,8,n_transfer_models)) 
    allocate(transfer_counter(n_transfer_models))
    allocate(transfer_counter_3d(1:6))
    allocate(kperp(li1:li2,lj1:lj2,lk1:lk2))

    Do k=lk1,lk2
       Do j=lj1,lj2
          Do i=li1,li2
             kperp(i,j,k) = sqrt(geom%gii(pi1,pj1,k)*kx(i)**2+2.0*geom%gij(pi1,pj1,k)*&
                  &kx(i)*ky(j)+geom%gjj(pi1,pj1,k)*ky(j)**2)
          Enddo
       Enddo
    Enddo

    kperp_maxloc = maxval(kperp)
    call mpi_allreduce(kperp_maxloc,kperp_max,1,MPI_REAL_TYPE,MPI_MAX,mpi_comm_xyz,ierr)

    ! Initialization parameters for the shells and transfer quantities
    n_shells_ky = nky0
    delta_ky = kymin
    n_shells_kx = real(nx0/2)
    delta_kx = real(2.*pi/lx)
    delta_kperp = real(kperp_max)/real(n_shells_kperp)
    k0 = kperp_max/(2**(real(n_shells_log_kperp-1)/5.))
    k0_y = kymax/(2**(real(n_shells_log_ky-1)/6.))

    IF (tky) call calc_initialize_diag_transfer(n_shells_ky,1)
    IF (tkx) call calc_initialize_diag_transfer(n_shells_kx,2)
    IF (tkperp) call calc_initialize_diag_transfer(n_shells_kperp,3)
    IF (t_log_kperp) call calc_initialize_diag_transfer(n_shells_log_kperp,4)
    IF (t_log_ky) call calc_initialize_diag_transfer(n_shells_log_ky,5)
    IF (t_log_kperp_1) call calc_initialize_diag_transfer(n_shells_log_kperp_1,6)
    IF (t_log_kperp_fb) call calc_initialize_diag_transfer(n_shells_log_kperp,7)
    
    IF (tk3_ky) call calc_initialize_diag_transfer_3d(n_shells_ky,1)
    IF (tk3_kx) call calc_initialize_diag_transfer_3d(n_shells_kx,2)
    IF (tk3_log_kperp) call calc_initialize_diag_transfer_3d(n_shells_log_kperp,4)
    IF (tk3_log_kperp_1) call calc_initialize_diag_transfer_3d(n_shells_log_kperp_1,6)
    IF (tk3_log_kperp_h) call calc_initialize_diag_transfer_3d_h(n_shells_log_kperp,4)

  End Subroutine initialize_diag_transfer


  Subroutine diagnostic_transfer

    IF (.not.with_nvp) THEN
        IF (tky) call calc_diagnostic_transfer(n_shells_ky,1)
        IF (tkx) call calc_diagnostic_transfer(n_shells_kx,2)
        IF (tkperp) call calc_diagnostic_transfer(n_shells_kperp,3)
        IF (t_log_kperp) call calc_diagnostic_transfer(n_shells_log_kperp,4)
        IF (t_log_ky) call calc_diagnostic_transfer(n_shells_log_ky,5)
        IF (t_log_kperp_1) call calc_diagnostic_transfer(n_shells_log_kperp_1,6)
        IF (t_log_kperp_fb) call calc_diagnostic_transfer(n_shells_log_kperp,7)
    ELSE 
        IF (tky) call calc_diagnostic_transfer_mod(n_shells_ky,1)
        IF (tkx) call calc_diagnostic_transfer_mod(n_shells_kx,2)
        IF (tkperp) call calc_diagnostic_transfer_mod(n_shells_kperp,3)
        IF (t_log_kperp) call calc_diagnostic_transfer_mod(n_shells_log_kperp,4)
        IF (t_log_ky) call calc_diagnostic_transfer_mod(n_shells_log_ky,5)
        IF (t_log_kperp_1) call calc_diagnostic_transfer_mod(n_shells_log_kperp_1,6)
        IF (t_log_kperp_fb) call calc_diagnostic_transfer_mod(n_shells_log_kperp,7)
    ENDIF

    IF (tk3_ky) call calc_diagnostic_transfer_3d(n_shells_ky,1)
    IF (tk3_kx) call calc_diagnostic_transfer_3d(n_shells_kx,2)
    IF (tk3_log_kperp) call calc_diagnostic_transfer_3d(n_shells_log_kperp,4)
    IF (tk3_log_kperp_1) call calc_diagnostic_transfer_3d(n_shells_log_kperp_1,6)
    IF (tk3_log_kperp_h) call calc_diagnostic_transfer_3d_h(n_shells_log_kperp,4)

  End Subroutine diagnostic_transfer


  Subroutine finalize_diag_transfer

    IF (tky) call calc_finalize_diag_transfer(1)
    IF (tkx) call calc_finalize_diag_transfer(2)
    IF (tkperp) call calc_finalize_diag_transfer(3)
    IF (t_log_kperp) call calc_finalize_diag_transfer(4)
    IF (t_log_ky) call calc_finalize_diag_transfer(5)
    IF (t_log_kperp_1) call calc_finalize_diag_transfer(6)
    IF (t_log_kperp_fb) call calc_finalize_diag_transfer(7)
    
    IF (tk3_ky) call calc_finalize_diag_transfer_3d(1)
    IF (tk3_kx) call calc_finalize_diag_transfer_3d(2)
    IF (tk3_log_kperp) call calc_finalize_diag_transfer_3d(4)
    IF (tk3_log_kperp_1) call calc_finalize_diag_transfer_3d(6)
    IF (tk3_log_kperp_h) call calc_finalize_diag_transfer_3d_h(4)


    deallocate(kperp)
    Deallocate(FE_TF_FILE)
    Deallocate(FE_SP_FILE)
    deallocate(FE_TF_SPEC_3D_FILE)
    Deallocate(transfer_counter)
    Deallocate(transfer_counter_3d)

  End Subroutine finalize_diag_transfer

!!!! ************************************************************************ !!!
  Subroutine calc_initialize_diag_transfer(n_shells,which_model)

    Implicit None
    Integer, intent(in) :: n_shells, which_model
    Integer :: n,o,t
    Character(len=4) :: xstring
    Character(len=20) :: label
    Character(len=20), dimension(8) :: label1
    Character(len=20), dimension(3) :: tf_file_label, spectral_file_label
    Write(xstring, '(i4)') n_shells
    
    label1 = (/'fe       ','drive    ','hyp_z    ',&
              &'hyp_v    ','hyp_kperp','coll     ',&
              &'curvature','parallel '/)

    tf_file_label = (/'/w_transfer_ ','/e_transfer_ ','/fe_transfer_'/)
    spectral_file_label = (/'/Spectral_W_ ','/Spectral_E_ ','/Spectral_FE_'/)

    select case(which_model)
    case(1)
       allocate(tf_cumul_1(n_shells,n_shells,ln1:ln2,n_transfer_models))
       allocate(spec_cumul_1(n_shells,ln1:ln2,3,8))
       tf_cumul_1 = 0
       spec_cumul_1 = 0 
       label ='ky_'
    case(2)
       allocate(tf_cumul_2(n_shells,n_shells,ln1:ln2,n_transfer_models))
       allocate(spec_cumul_2(n_shells,ln1:ln2,3,8))
       tf_cumul_2 = 0
       spec_cumul_2 = 0 
       label ='kx_'
    case(3)
       allocate(tf_cumul_3(n_shells,n_shells,ln1:ln2,n_transfer_models))
       allocate(spec_cumul_3(n_shells,ln1:ln2,3,8))
       tf_cumul_3 = 0 
       spec_cumul_3 = 0 
       label ='kperp_'
    case(4)
       allocate(tf_cumul_4(n_shells,n_shells,ln1:ln2,n_transfer_models))
       allocate(spec_cumul_4(n_shells,ln1:ln2,3,8))
       tf_cumul_4 = 0 
       spec_cumul_4 = 0 
       label ='log_kperp_'
    case(5)
       allocate(tf_cumul_5(n_shells,n_shells,ln1:ln2,n_transfer_models))
       allocate(spec_cumul_5(n_shells,ln1:ln2,3,8))
       tf_cumul_5 = 0 
       spec_cumul_5 = 0 
       label ='log_ky_'
    case(6)
       allocate(tf_cumul_6(n_shells,n_shells,ln1:ln2,n_transfer_models))
       allocate(spec_cumul_6(n_shells,ln1:ln2,3,8))
       tf_cumul_6 = 0 
       spec_cumul_6 = 0 
       label ='log_kperp1_'
     case(7)
       allocate(tf_cumul_7(n_shells,n_shells,ln1:ln2,n_transfer_models))
       allocate(spec_cumul_7(n_shells,ln1:ln2,3,8))
       tf_cumul_7 = 0 
       spec_cumul_7 = 0 
       label ='log_kperp_fb_'
   case default
   end select
    transfer_counter(which_model)=0

    IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
    IF (.not.with_nvp) THEN
       Do o=1,3
          Do n=ln1,ln2
             call get_unit_nr(FE_TF_FILE(n,o,which_model))
             OPEN(FE_TF_FILE(n,o,which_model), file=trim(diagdir)//trim(tf_file_label(o))&
                  //trim(adjustl(label))//trim(spec(n)%name)//trim(file_extension),form= 'FORMATTED', &
                  status=filestat, position=filepos)
          enddo
       enddo
    ENDIF
       do t=1,8
          do o=1,3
             do n=ln1,ln2
                call get_unit_nr(FE_SP_FILE(n,o,t,which_model))
                OPEN(FE_SP_FILE(n,o,t,which_model), file=trim(diagdir)//trim(spectral_file_label(o))&
                     //trim(adjustl(label1(t)))//'_'//trim(adjustl(label))//&
                     trim(spec(n)%name)//trim(file_extension),form = 'FORMATTED', &
                     status=filestat, position=filepos)         
             enddo
          enddo
       enddo
    endif

  End Subroutine calc_initialize_diag_transfer

!!!*****************************************************************************************!!!

  Subroutine calc_diagnostic_transfer(n_shells,which_model)


    integer, intent(in) :: which_model,n_shells
    !g,f,h and nonlinearities and fields arrays

    !Other factors used
    complex, dimension(:,:,:),pointer :: ptr2
    integer :: lbg1,lbg2,stage, ls1,ls2    
    complex, dimension(:,:,:,:),pointer :: ptr1, ptr3
    complex, dimension(:,:,:,:,:,:), pointer :: ptr4
    !transfer array and flux

    allocate(g_ls1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(g_ls2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(nlt_ls2_g(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(nlt_ls2_f(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(f_ls2(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    allocate(fields_ls2(li1:li2, lj1:lj2, lbz:ubz,1:n_fields)) 
    allocate(fe_tf_tot(n_shells,n_shells,ln1:ln2,n_transfer_models))
    allocate(fe_tot(n_shells,ln1:ln2,3,8))
    allocate(k_value(1:n_shells))

    !abn 8mar2010  some initialization values
    ptr1 => null()
    ptr2 => null()
    ptr3 => null()
    ptr4 => null()
    lbg1 = 1
    lbg2 = lklmn0
    stage = 0

    ! This initialization that should be here and cant be inside the ls1,ls2 loop
    fe_tf_tot = 0.0
    fe_tot = 0.0

    do ls2=1,n_shells
       select case(which_model)
       case(1) ! ky shells           
          call calc_n_shells_ky(ls2,g_1,g_ls2)
          k_value(ls2) = (ls2-1)*delta_ky 
          call calc_spectral(g_ls2,fe_tot(ls2,:,:,:))
       case(2) ! kx shells
          call calc_n_shells_kx(ls2,g_1,g_ls2)
          k_value(ls2) = (ls2-1)*delta_kx
          call calc_spectral(g_ls2,fe_tot(ls2,:,:,:)) 
       case(3) ! kperp shells
          call calc_n_shells_kperp(ls2,g_1,g_ls2)
          k_value(ls2) = ls2*delta_kperp
          call calc_spectral(g_ls2,fe_tot(ls2,:,:,:)) 
       case(4) ! log kperp shells
          call calc_n_shells_log_kperp(k0,ls2,g_1,g_ls2)
          k_value(ls2) = k0*2**(real((ls2-1)/5.))
          call calc_spectral(g_ls2,fe_tot(ls2,:,:,:)) 
      case(5) ! log ky shells
          call calc_n_shells_log_ky(k0_y,ls2,g_1,g_ls2)
          k_value(ls2) = k0_y*2**(real((ls2-1)/6.))
          call calc_spectral(g_ls2,fe_tot(ls2,:,:,:)) 
      case(6) ! perpendicular shells modification 
          call calc_n_shells_log_kperp_1(ls2,g_1,g_ls2)
          If (ls2.eq.1) k_value(ls2) = lambda1
          If (ls2.eq.2) k_value(ls2) = lambda1 +lambda2
          IF (ls2.ge.3 .and. ls2.lt.n_shells) k_value(ls2) = &
            (lambda1+lambda2)*2**(real((ls2-2)/5.))
          If (ls2.eq.n_shells) k_value(ls2) = kperp_max
          call calc_spectral(g_ls2,fe_tot(ls2,:,:,:)) 
       case(7) ! log kperp shells
          call calc_n_shells_log_kperp(k0,ls2,g_1,g_ls2)
          k_value(ls2) = k0*2**(real((ls2-1)/5.))
          call calc_spectral(g_ls2,fe_tot(ls2,:,:,:)) 
       end select

       call calc_aux_fields(g_ls2, fields_ls2, f_ls2,.false.)    
       nlt_ls2_g = cmplx(0.0,0.0) 
       nlt_ls2_f = cmplx(0.0,0.0)
       CALL this_nonlinear_term%add(g_ls2, ptr1, emfields,&
            &ptr2, ptr3, nlt_ls2_g, lbg1, lbg2, stage)

       CALL this_nonlinear_term%add(g_1, ptr1, fields_ls2,&
            &ptr2, ptr3, nlt_ls2_f, lbg1, lbg2, stage)

       do ls1=1,n_shells
          select case(which_model)
          case(1) ! ky shells
             call calc_n_shells_ky(ls1,g_1,g_ls1)
             call calc_transfer(ls2,ls1,g_ls1,nlt_ls2_f,nlt_ls2_g,fe_tf_tot(ls2,ls1,:,:))
          case(2) ! kx shells
             call calc_n_shells_kx(ls1,g_1,g_ls1)
             call calc_transfer(ls2,ls1,g_ls1,nlt_ls2_f,nlt_ls2_g,fe_tf_tot(ls2,ls1,:,:))
          case(3) ! kperp1 shells
             call calc_n_shells_kperp(ls1,g_1,g_ls1)
             call calc_transfer(ls2,ls1,g_ls1,nlt_ls2_f,nlt_ls2_g,fe_tf_tot(ls2,ls1,:,:))
          case(4) ! log kperp shells
             call calc_n_shells_log_kperp(k0,ls1,g_1,g_ls1)
             call calc_transfer(ls2,ls1,g_ls1,nlt_ls2_f,nlt_ls2_g,fe_tf_tot(ls2,ls1,:,:))
          case(5) ! log ky shells
             call calc_n_shells_log_ky(k0_y,ls1,g_1,g_ls1)
             call calc_transfer(ls2,ls1,g_ls1,nlt_ls2_f,nlt_ls2_g,fe_tf_tot(ls2,ls1,:,:))
          case(6) ! log ky shells
             call calc_n_shells_log_kperp_1(ls1,g_1,g_ls1)
             call calc_transfer(ls2,ls1,g_ls1,nlt_ls2_f,nlt_ls2_g,fe_tf_tot(ls2,ls1,:,:))
          case(7) ! log kperp shells for back scattering
             call calc_n_shells_log_kperp(k0,ls1,g_1,g_ls1)
             call calc_transfer_fb(ls2,ls1,g_ls1,nlt_ls2_f,nlt_ls2_g,fe_tf_tot(ls2,ls1,:,:))
         end select
       enddo
    enddo
    
    deallocate(g_ls1) 
    deallocate(g_ls2) 
    deallocate(nlt_ls2_g) 
    deallocate(nlt_ls2_f) 
    deallocate(f_ls2) 
    deallocate(fields_ls2) 

    if ( .not.( mod(itime,istep_fe_transfer*avg_window) .eq. 0) ) then
       ! We cumulate in the first case
       select case(which_model)
       case(1)
          tf_cumul_1 = fe_tf_tot + tf_cumul_1
          spec_cumul_1 = fe_tot + spec_cumul_1
       case(2)
          tf_cumul_2 = fe_tf_tot + tf_cumul_2
          spec_cumul_2 = fe_tot + spec_cumul_2
       case(3)
          tf_cumul_3 = fe_tf_tot + tf_cumul_3
          spec_cumul_3 = fe_tot + spec_cumul_3
       case(4)
          tf_cumul_4 = fe_tf_tot + tf_cumul_4
          spec_cumul_4 = fe_tot + spec_cumul_4
       case(5)
          tf_cumul_5 = fe_tf_tot + tf_cumul_5
          spec_cumul_5 = fe_tot + spec_cumul_5
       case(6)
          tf_cumul_6 = fe_tf_tot + tf_cumul_6
          spec_cumul_6 = fe_tot + spec_cumul_6
       case(7)
          tf_cumul_7 = fe_tf_tot + tf_cumul_7
          spec_cumul_7 = fe_tot + spec_cumul_7
       case default
       end select

       transfer_counter(which_model) = transfer_counter(which_model) + 1 
    endif

    ! Then we do output files
    if ( itime .gt. 0 .and. mod(itime,istep_fe_transfer*avg_window) .eq. 0 ) then
       ! Here we use v2d as a buffer for write routine and average it

       select case(which_model)
       case(1)
          tf_cumul_1 = fe_tf_tot + tf_cumul_1
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          fe_tf_tot =  tf_cumul_1 /real(transfer_counter(which_model))

          spec_cumul_1 = fe_tot + spec_cumul_1
          fe_tot =  spec_cumul_1/real(transfer_counter(which_model))
       case(2)
          tf_cumul_2 = fe_tf_tot + tf_cumul_2
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          fe_tf_tot =  tf_cumul_2/real(transfer_counter(which_model))
       
          spec_cumul_2 = fe_tot + spec_cumul_2
          fe_tot =  spec_cumul_2/real(transfer_counter(which_model))
       case(3)
          tf_cumul_3 = fe_tf_tot + tf_cumul_3
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          fe_tf_tot =  tf_cumul_3/real(transfer_counter(which_model))
       
          spec_cumul_3 = fe_tot + spec_cumul_3
          fe_tot =  spec_cumul_3/real(transfer_counter(which_model))
       case(4)
          tf_cumul_4 = fe_tf_tot + tf_cumul_4
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          fe_tf_tot =  tf_cumul_4/real(transfer_counter(which_model))
          
          spec_cumul_4 = fe_tot + spec_cumul_4
          fe_tot =  spec_cumul_4/real(transfer_counter(which_model))
       case(5)
          tf_cumul_5 = fe_tf_tot + tf_cumul_5
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          fe_tf_tot =  tf_cumul_5/real(transfer_counter(which_model))
      
          spec_cumul_5 = fe_tot + spec_cumul_5
          fe_tot =  spec_cumul_5/real(transfer_counter(which_model))
       case(6)
          tf_cumul_6 = fe_tf_tot + tf_cumul_6
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          fe_tf_tot =  tf_cumul_6/real(transfer_counter(which_model))
      
          spec_cumul_6 = fe_tot + spec_cumul_6
          fe_tot =  spec_cumul_6/real(transfer_counter(which_model))
        case(7)
          tf_cumul_7 = fe_tf_tot + tf_cumul_7
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          fe_tf_tot =  tf_cumul_7/real(transfer_counter(which_model))
          
          spec_cumul_7 = fe_tot + spec_cumul_7
          fe_tot =  spec_cumul_7/real(transfer_counter(which_model))
       case default
       end select

       Call write_transfer(n_shells,which_model)    

       select case(which_model)
       case(1)
           tf_cumul_1 = 0
           Call write_spectral(n_shells,which_model)
           spec_cumul_1 = 0 
        case(2)
           tf_cumul_2 = 0
           Call write_spectral(n_shells,which_model)
           spec_cumul_2 = 0 
        case(3)
           tf_cumul_3 = 0 
           Call write_spectral(n_shells,which_model)
           spec_cumul_3 = 0 
        case(4)
           tf_cumul_4 = 0 
           Call write_spectral(n_shells,which_model)
           spec_cumul_4 = 0 
        case(5)
           tf_cumul_5 = 0 
           Call write_spectral(n_shells,which_model)
           spec_cumul_5 = 0 
        case(6)
           tf_cumul_6 = 0 
           Call write_spectral(n_shells,which_model)
           spec_cumul_6 = 0 
        case(7)
           tf_cumul_7 = 0 
           Call write_spectral(n_shells,which_model)
           spec_cumul_7 = 0 
        case default
        end select
 
       transfer_counter(which_model)=0
    endif
    deallocate(fe_tf_tot,fe_tot,k_value) 

  End subroutine calc_diagnostic_transfer


!!!*****************************************************************************************!!!

  Subroutine calc_diagnostic_transfer_mod(n_shells,which_model)


    integer, intent(in) :: which_model,n_shells
    integer :: ls2    
    allocate(g_ls2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(fe_tot(n_shells,ln1:ln2,3,8))
    allocate(k_value(1:n_shells))

    ! This initialization that should be here and cant be inside the ls1,ls2 loop
    fe_tot = 0.0

    do ls2=1,n_shells
       select case(which_model)
       case(1) ! ky shells           
          call calc_n_shells_ky(ls2,g_1,g_ls2)
          k_value(ls2) = (ls2-1)*delta_ky 
          call calc_spectral_mod(g_ls2,fe_tot(ls2,:,:,:)) 
       case(2) ! kx shells
          call calc_n_shells_kx(ls2,g_1,g_ls2)
          k_value(ls2) = (ls2-1)*delta_kx
          call calc_spectral_mod(g_ls2,fe_tot(ls2,:,:,:)) 
       case(3) ! kperp shells
          call calc_n_shells_kperp(ls2,g_1,g_ls2)
          k_value(ls2) = ls2*delta_kperp
          call calc_spectral_mod(g_ls2,fe_tot(ls2,:,:,:)) 
       case(4) ! log kperp shells
          call calc_n_shells_log_kperp(k0,ls2,g_1,g_ls2)
          k_value(ls2) = k0*2**(real((ls2-1)/5.))
          call calc_spectral_mod(g_ls2,fe_tot(ls2,:,:,:)) 
      case(5) ! log ky shells
          call calc_n_shells_log_ky(k0_y,ls2,g_1,g_ls2)
          k_value(ls2) = k0_y*2**(real((ls2-1)/6.))
          call calc_spectral_mod(g_ls2,fe_tot(ls2,:,:,:)) 
      case(6) ! perpendicular shells modification 
          call calc_n_shells_log_kperp_1(ls2,g_1,g_ls2)
          If (ls2.eq.1) k_value(ls2) = lambda1
          If (ls2.eq.2) k_value(ls2) = lambda1 +lambda2
          IF (ls2.ge.3 .and. ls2.lt.n_shells) k_value(ls2) = &
            (lambda1+lambda2)*2**(real((ls2-2)/5.))
          If (ls2.eq.n_shells) k_value(ls2) = kperp_max
          call calc_spectral_mod(g_ls2,fe_tot(ls2,:,:,:)) 
       case(7) ! log kperp shells
          call calc_n_shells_log_kperp(k0,ls2,g_1,g_ls2)
          k_value(ls2) = k0*2**(real((ls2-1)/5.))
          call calc_spectral_mod(g_ls2,fe_tot(ls2,:,:,:)) 
       end select
    enddo
    
    deallocate(g_ls2) 

    if ( .not.( mod(itime,istep_fe_transfer*avg_window) .eq. 0) ) then
       ! We cumulate in the first case
       select case(which_model)
       case(1)
          spec_cumul_1 = fe_tot + spec_cumul_1
       case(2)
          spec_cumul_2 = fe_tot + spec_cumul_2
       case(3)
          spec_cumul_3 = fe_tot + spec_cumul_3
       case(4)
          spec_cumul_4 = fe_tot + spec_cumul_4
       case(5)
          spec_cumul_5 = fe_tot + spec_cumul_5
       case(6)
          spec_cumul_6 = fe_tot + spec_cumul_6
       case(7)
          spec_cumul_7 = fe_tot + spec_cumul_7
       case default
       end select

       transfer_counter(which_model) = transfer_counter(which_model) + 1 
    endif

    ! Then we do output files
    if ( itime .gt. 0 .and. mod(itime,istep_fe_transfer*avg_window) .eq. 0 ) then
       ! Here we use v2d as a buffer for write routine and average it

       select case(which_model)
       case(1)
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          spec_cumul_1 = fe_tot + spec_cumul_1
          fe_tot =  spec_cumul_1/real(transfer_counter(which_model))
       case(2)
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          spec_cumul_2 = fe_tot + spec_cumul_2
          fe_tot =  spec_cumul_2/real(transfer_counter(which_model))
       case(3)
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          spec_cumul_3 = fe_tot + spec_cumul_3
          fe_tot =  spec_cumul_3/real(transfer_counter(which_model))
       case(4)
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          spec_cumul_4 = fe_tot + spec_cumul_4
          fe_tot =  spec_cumul_4/real(transfer_counter(which_model))
       case(5)
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          spec_cumul_5 = fe_tot + spec_cumul_5
          fe_tot =  spec_cumul_5/real(transfer_counter(which_model))
       case(6)
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          spec_cumul_6 = fe_tot + spec_cumul_6
          fe_tot =  spec_cumul_6/real(transfer_counter(which_model))
        case(7)
          transfer_counter(which_model) = transfer_counter(which_model) + 1 
          spec_cumul_7 = fe_tot + spec_cumul_7
          fe_tot =  spec_cumul_7/real(transfer_counter(which_model))
       case default
       end select

       select case(which_model)
       case(1)
           Call write_spectral(n_shells,which_model)
           spec_cumul_1 = 0 
        case(2)
           Call write_spectral(n_shells,which_model)
           spec_cumul_2 = 0 
        case(3)
           Call write_spectral(n_shells,which_model)
           spec_cumul_3 = 0 
        case(4)
           Call write_spectral(n_shells,which_model)
           spec_cumul_4 = 0 
        case(5)
           Call write_spectral(n_shells,which_model)
           spec_cumul_5 = 0 
        case(6)
           Call write_spectral(n_shells,which_model)
           spec_cumul_6 = 0 
        case(7)
           Call write_spectral(n_shells,which_model)
           spec_cumul_7 = 0 
        case default
        end select
 
       transfer_counter(which_model)=0
    endif
    deallocate(fe_tot,k_value) 

  End subroutine calc_diagnostic_transfer_mod

!!!*****************************************************************************************!!!

  Subroutine calc_transfer(ls2,ls1,g_ls1,nlt_ls2_f,nlt_ls2_g,fe_tf_tot)

    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) ::g_ls1, nlt_ls2_f, nlt_ls2_g
    integer, intent(in) :: ls1, ls2
    real, dimension(ln1:ln2,3),intent(inout) :: fe_tf_tot
    complex, dimension(li1:li2,lj1:lj2,lbz:ubz,1: n_fields) :: fields_ls1
    complex, dimension(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2) :: f_ls1
    !temporary array to integrate over vp , mu and species
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ln1:ln2)::e_n_temp
    !Arrays of terms to be analyzed for transfer analysis
    integer :: n
    
    allocate(cfgamma1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(cfgamma2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(fe_tf_term(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2)) 
    
    call calc_aux_fields(g_ls1, fields_ls1, f_ls1,.false.)    
    call get_cfgamma(f_ls1,fields_ls1,cfgamma1,cfgamma2)

    ! Calculte the tranfers of E, W and FE different for ions than for electrons

    do n =ln1,ln2
       e_n_temp(:,:,:,n) = cmplx(0.0,0.0)
       fe_tf_term(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*nlt_ls2_g(:,:,:,:,:,n)
       call integral_vpw(fe_tf_term(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
       if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
       call sum_int_3d(e_n_temp(:,:,:,n),fe_tf_tot(n,1))
    enddo

    do n=ln1,ln2
       e_n_temp(:,:,:,n) = cmplx(0.0,0.0)
       fe_tf_term(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*nlt_ls2_f(:,:,:,:,:,n)
       call integral_vpw(fe_tf_term(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
       if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
       call sum_int_3d(e_n_temp(:,:,:,n),fe_tf_tot(n,2))
    enddo


    do n=ln1,ln2
       fe_tf_tot(n,3) = fe_tf_tot(n,1) + fe_tf_tot(n,2)
    enddo

    deallocate(cfgamma1) 
    deallocate(cfgamma2) 
    deallocate(fe_tf_term) 
  

  End subroutine calc_transfer

  Subroutine calc_transfer_fb(ls2,ls1,g_ls1,nlt_ls2_f,nlt_ls2_g,fe_tf_tot)

    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) :: g_ls1, nlt_ls2_f, nlt_ls2_g
    integer, intent(in) :: ls1, ls2
    real, dimension(ln1:ln2,6),intent(inout) :: fe_tf_tot
    complex, dimension(li1:li2,lj1:lj2,lbz:ubz,1: n_fields) :: fields_ls1
    complex, dimension(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2) :: f_ls1
    !temporary array to integrate over vp , mu and species
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ln1:ln2,2)::e_n_temp
    !Arrays of terms to be analyzed for transfer analysis
    integer :: n
    
    allocate(cfgamma1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(cfgamma2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(fe_tf_term(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2)) 
    allocate(fe_tf_term_p(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2)) 
    allocate(fe_tf_term_n(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2)) 
    
    call calc_aux_fields(g_ls1, fields_ls1, f_ls1,.false.)    
    call get_cfgamma(f_,emfields,cfgamma1,cfgamma2)

    ! Calculte the tranfers of E, W and FE different for ions than for electrons

    do n =ln1,ln2
       e_n_temp(:,:,:,n,:) = cmplx(0.0,0.0)
       fe_tf_term(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*nlt_ls2_g(:,:,:,:,:,n)
       call calc_n_shells_log_kperp_pn(k0,ls1,fe_tf_term(:,:,:,:,:,n,1),fe_tf_term_p(:,:,:,:,:,n,1),&
                                        fe_tf_term_n(:,:,:,:,:,n,1))
       call integral_vpw(fe_tf_term_p(:,:,:,:,:,n,1),e_n_temp(:,:,:,n,1))
       call integral_vpw(fe_tf_term_n(:,:,:,:,:,n,1),e_n_temp(:,:,:,n,2))
       if(p_has_00_mode) e_n_temp(0,0,:,n,1) =cmplx(0.0,0.0)
       if(p_has_00_mode) e_n_temp(0,0,:,n,2) =cmplx(0.0,0.0)
       call sum_int_3d(e_n_temp(:,:,:,n,1),fe_tf_tot(n,1))
       call sum_int_3d(e_n_temp(:,:,:,n,2),fe_tf_tot(n,4))
    enddo

    do n =ln1,ln2
       e_n_temp(:,:,:,n,:) = cmplx(0.0,0.0)
       fe_tf_term(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*nlt_ls2_f(:,:,:,:,:,n)
       call calc_n_shells_log_kperp_pn(k0,ls1,fe_tf_term(:,:,:,:,:,n,2),fe_tf_term_p(:,:,:,:,:,n,2),&
                                        fe_tf_term_n(:,:,:,:,:,n,2))
       call integral_vpw(fe_tf_term_p(:,:,:,:,:,n,2),e_n_temp(:,:,:,n,1))
       call integral_vpw(fe_tf_term_n(:,:,:,:,:,n,2),e_n_temp(:,:,:,n,2))
       if(p_has_00_mode) e_n_temp(0,0,:,n,1) =cmplx(0.0,0.0)
       if(p_has_00_mode) e_n_temp(0,0,:,n,2) =cmplx(0.0,0.0)
       call sum_int_3d(e_n_temp(:,:,:,n,1),fe_tf_tot(n,2))
       call sum_int_3d(e_n_temp(:,:,:,n,2),fe_tf_tot(n,5))
    enddo

    do n=ln1,ln2
       fe_tf_tot(n,3) = fe_tf_tot(n,1) + fe_tf_tot(n,2)
       fe_tf_tot(n,6) = fe_tf_tot(n,4) + fe_tf_tot(n,5)
    enddo

    deallocate(cfgamma1) 
    deallocate(cfgamma2) 
    deallocate(fe_tf_term) 
    deallocate(fe_tf_term_p) 
    deallocate(fe_tf_term_n) 
  
  End subroutine calc_transfer_fb
!!*****************************************************************************************!!!
Subroutine calc_spectral(g_ls2,fe_tot)

    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) :: g_ls2
    real, dimension(ln1:ln2,3,8),intent(inout) :: fe_tot
    complex, dimension(li1:li2,lj1:lj2,lbz:ubz,1: n_fields) :: fields_ls2
    complex, dimension(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2) :: f_ls2
    !temporary array to integrate over vp , mu and species
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ln1:ln2)::e_n_temp
    !Other factors used
    complex, dimension(:,:,:),pointer :: ptr2
    complex, dimension(:,:,:,:),pointer :: ptr1, ptr3
    complex, dimension(:,:,:,:,:,:), pointer :: ptr4
    integer ::n,lbg1,lbg2

    Allocate(v7d_spec(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2,ln1:ln2,1:14)) 
    Allocate(g_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(gc_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(d_v_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(d_z_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(d_kperp_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(coll_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(p_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(c_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(cfgamma1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(cfgamma2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(cfgamma3(li1:li2, lj1:lj2, lk1:lk2, ln1:ln2))
    
    ptr1 => null()
    ptr2 => null()
    ptr3 => null()
    ptr4 => null()
    lbg1 = 1 
    lbg2 = lklmn0
    v7d_spec=cmplx(0.0,0.0)

    call calc_aux_fields(g_ls2, fields_ls2, f_ls2,.false.) !compute f and h 
    call get_cfgamma(f_ls2,fields_ls2,cfgamma1,cfgamma2)

    d_v_rhs = cmplx(0.0,0.0)
    if (arakawa_zv) then 
       if (hyp_on_h) then 
          call add_hypv_ak(h_,d_v_rhs,lbg1,lbg2)
       else
          call add_hypv_ak(f_,d_v_rhs,lbg1,lbg2)
       endif
    else
        call add_hypv_block(f_,d_v_rhs,lbg1,lbg2)
    endif

    d_z_rhs = cmplx(0.0,0.0)
    if (arakawa_zv) then
       if (hyp_on_h) then 
          call add_hypz_ak(h_,d_z_rhs,lbg1,lbg2)
          if (hypz_compensation) call equ_comp_hypz(emfields,ptr4,d_z_rhs,lbg1,lbg2)
       else  
          call add_hypz_ak(f_,d_z_rhs,lbg1,lbg2)
       endif  
    else
        call add_hypz_block(f_,d_z_rhs,lbg1,lbg2)
    endif

    d_kperp_rhs = cmplx(0.0,0.0)
    if ((hyp_x.gt.0.0).or.(hyp_y.gt.0.0).or.(hyp_perp.gt.0.0).or.(GyroLES)) then
       !/todo:check if hyp_on_h should affect perpendicular hyperdiffusions!
       call add_dfdxy(f_,d_kperp_rhs,lbg1,lbg2)
    endif

    !Collisions works with f
    coll_rhs = cmplx(0.0,0.0)                 
    if (collision_op.ne.'none') then
       call equ_collisions(f_,coll_rhs,.true.)
    endif

    !Gradient and curvature
    gc_rhs = cmplx(0.0,0.0)
    call add_dgdxy(g_1, gc_rhs, ptr1, pdg1di, pdg1dj,lbg1,lbg2)
    call add_dchidxy_orig(emfields,ptr4, gc_rhs,ptr2,ptr3,lbg1,lbg2,.true.)
    !call add_dchidxy_1(emfields,gc_rhs,ptr2,lbg1,lbg2,.true.)

    g_rhs  = cmplx(0.0,0.0)
    call add_dchidxy_orig(emfields,ptr4, g_rhs,ptr2,ptr3,lbg1,lbg2,.false.)
    
    c_rhs = cmplx(0.0,0.0)
    c_rhs = gc_rhs - g_rhs   

    !Parallel term
    p_rhs = cmplx(0.0,0.0)
    if (arakawa_zv) then
       call equ_dzv(h_,p_rhs,lbg1,lbg2) !f_ is h
       if (hyp_on_h) then
          if (hypz_compensation) call equ_comp_hypz(emfields,ptr4,p_rhs,lbg1,lbg2)
       endif
    else
        call equ_dfdzv(f_,p_rhs,lbg1,lbg2)
        call add_dchidz(emfields, ptr4, p_rhs, lbg1, lbg2)
    end if

    !TERM 1  FREE ENERGY      
    !W
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*g_1(:,:,:,:,:,n)/2.
        call integral_vpw(v7d_spec(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,1))
    enddo                     
    
    
    if (n_spec.eq.1) then
    ! Electrostatic
        Do n=ln1,ln2
            e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
            v7d_spec(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*g_1(:,:,:,:,:,n)/2.
            Call integral_vpw(v7d_spec(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
            if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
            call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,1))
        Enddo
    else
        ! Electrostatic with modification
        call get_electrostatic(fields_ls2,cfgamma3)
        Do n=ln1,ln2
            if(p_has_00_mode) cfgamma3(0,0,:,n)=cmplx(0.0,0.0)
            call sum_int_3d(cfgamma3(:,:,:,n),fe_tot(n,2,1))
        Enddo
    endif 
    
    
    ! FE
    do n =ln1,ln2
        fe_tot(n,3,1) = fe_tot(n,1,1) + fe_tot(n,2,1)
    enddo
 
    ! TERM 2  DRIVE
    ! W
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,3)= cfgamma1(:,:,:,:,:,n)*g_rhs(:,:,:,:,:,n)    
        call integral_vpw(v7d_spec(:,:,:,:,:,n,3),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,2))
    enddo    
    ! E
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,4) = cfgamma2(:,:,:,:,:,n)*g_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,4),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,2))
    enddo       
    ! FE
    do n =ln1,ln2
        fe_tot(n,3,2) = fe_tot(n,1,2) + fe_tot(n,2,2)
    enddo
    
        
    !TERM 3 Hyp Z       
    !W 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,5) = cfgamma1(:,:,:,:,:,n)*d_z_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,5),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,3))
    enddo
    !E 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,6) = cfgamma2(:,:,:,:,:,n)*d_z_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,6),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,3))
    enddo    
    !FE
    do n =ln1,ln2
        fe_tot(n,3,3) = fe_tot(n,1,3) + fe_tot(n,2,3)
    enddo 

    !TERM 4 Hyp V 
    !W
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,7) = cfgamma1(:,:,:,:,:,n)*d_v_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,7),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,4))
    enddo
    !E
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,8) = cfgamma2(:,:,:,:,:,n)*d_v_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,8),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,4))
    enddo    
    !FE
    do n =ln1,ln2
        fe_tot(n,3,4) = fe_tot(n,1,4) + fe_tot(n,2,4)
    enddo    

    !TERM 5 hyp perp 
    !W 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,9) = cfgamma1(:,:,:,:,:,n)*d_kperp_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,9),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,5))
    enddo
    !E
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,10) = cfgamma2(:,:,:,:,:,n)*d_kperp_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,10),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,5))
    enddo    
    !FE
    do n =ln1,ln2
        fe_tot(n,3,5) = fe_tot(n,1,5) + fe_tot(n,2,5)
    enddo    
 
    !TERM 6 Collisions  
    if (collision_op.ne.'none') then
    ! W collisions
        do n =ln1,ln2
             e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
             v7d_spec(:,:,:,:,:,n,11) = cfgamma1(:,:,:,:,:,n)*coll_rhs(:,:,:,:,:,n)
             call integral_vpw(v7d_spec(:,:,:,:,:,n,11),e_n_temp(:,:,:,n))
             if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
             call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,6))
        enddo            
    ! E collisions
        do n =ln1,ln2
            e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
            v7d_spec(:,:,:,:,:,n,12) = cfgamma2(:,:,:,:,:,n)*coll_rhs(:,:,:,:,:,n)
            call integral_vpw(v7d_spec(:,:,:,:,:,n,12),e_n_temp(:,:,:,n))
            if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
            call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,6))
        enddo
       ! FE collisions
        do n =ln1,ln2
            fe_tot(n,3,6) = fe_tot(n,1,6) + fe_tot(n,2,6)
        enddo
     else
        do n =ln1,ln2
             fe_tot(n,1,6) = 0.0
             fe_tot(n,2,6) = 0.0
             fe_tot(n,3,6) = 0.0
        enddo
     endif

    !TERM 7 curvature 
    !W 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,13) = cfgamma1(:,:,:,:,:,n)*c_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,13),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,7))
    enddo
    !E
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,13) = cfgamma2(:,:,:,:,:,n)*c_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,13),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,7))
    enddo    
    !FE
    do n =ln1,ln2
        fe_tot(n,3,7) = fe_tot(n,1,7) + fe_tot(n,2,7)
    enddo    
 
    !TERM 8 Parallel 
    !W 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,14) = cfgamma1(:,:,:,:,:,n)*p_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,14),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,8))
    enddo
    !E
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,14) = cfgamma2(:,:,:,:,:,n)*p_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,14),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,8))
    enddo    
  
    if (hyp_on_h) then
       do n =ln1,ln2
          fe_tot(n,1,8) = fe_tot(n,1,8) - fe_tot(n,1,4) - fe_tot(n,1,3)
          fe_tot(n,2,8) = fe_tot(n,2,8) - fe_tot(n,2,4) - fe_tot(n,2,3)
       enddo
    else
       do n =ln1,ln2
          if (.not.arakawa_zv) fe_tot(n,1,8) = fe_tot(n,1,8)-fe_tot(n,1,4)-fe_tot(n,1,3) !eliminated hyperdiffusions
          if (.not.arakawa_zv) fe_tot(n,2,8) = fe_tot(n,2,8)-fe_tot(n,2,4)-fe_tot(n,2,3) !eliminated hyperdiffusions
       enddo
    endif
    !FE
    do n =ln1,ln2
        fe_tot(n,3,8) = fe_tot(n,1,8) + fe_tot(n,2,8)
    enddo    
 
    deallocate(v7d_spec)
    deallocate(g_rhs)
    deallocate(gc_rhs)
    deallocate(p_rhs)
    deallocate(c_rhs)
    deallocate(d_v_rhs)
    deallocate(d_z_rhs)
    deallocate(d_kperp_rhs)
    deallocate(coll_rhs)
    deallocate(cfgamma1)
    deallocate(cfgamma2)
    deallocate(cfgamma3)
 
End Subroutine calc_spectral

Subroutine calc_spectral_mod(g_ls2,fe_tot)

    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) :: g_ls2
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2) :: cfgamma1m
    real, dimension(ln1:ln2,3,8),intent(inout) :: fe_tot
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, lm1:lm2,ln1:ln2,1:16) :: v6d_spec
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, lm1:lm2, ln1:ln2) :: cfgamma1_nvp
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, lm1:lm2, ln1:ln2) ::  g_1_nvp, g_rhs_nvp,&
                                                     &d_v_rhs_nvp, d_z_rhs_nvp, d_kperp_rhs_nvp, h_nvp
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, lm1:lm2, ln1:ln2) ::  coll_rhs_nvp, p_rhs_nvp, c_rhs_nvp
    complex, dimension(li1:li2,lj1:lj2,lbz:ubz,1: n_fields) :: fields_ls2
    complex, dimension(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2) :: f_ls2
    !temporary array to integrate over vp , mu and species
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ln1:ln2)::e_n_temp
    !Other factors used
    complex, dimension(:,:,:),pointer :: ptr2
    complex, dimension(:,:,:,:),pointer :: ptr1, ptr3
    complex, dimension(:,:,:,:,:,:), pointer :: ptr4
    integer ::n,lbg1,lbg2

    Allocate(v7d_spec(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2,ln1:ln2,8)) 
    allocate(h_ls2(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    Allocate(g_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(gc_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(d_v_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(d_z_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(d_kperp_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(coll_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(p_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(c_rhs(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(cfgamma1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(cfgamma2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2))
    Allocate(cfgamma3(li1:li2, lj1:lj2, lk1:lk2, ln1:ln2))
    
    ptr1 => null()
    ptr2 => null()
    ptr3 => null()
    ptr4 => null()
    lbg1 = 1 
    lbg2 = lklmn0
    
    v7d_spec=cmplx(0.0,0.0)
    v6d_spec=cmplx(0.0,0.0)

    call calc_aux_fields(g_ls2, fields_ls2, f_ls2,comp_h=.true.,h_out=h_ls2)    
    call get_cfgamma_mod(f_ls2,fields_ls2,cfgamma1m)
    call get_cfgamma(f_ls2,fields_ls2,cfgamma1,cfgamma2)
    
    d_v_rhs = cmplx(0.0,0.0)
    if (arakawa_zv) then 
       if (hyp_on_h) then 
          call add_hypv_ak(h_,d_v_rhs,lbg1,lbg2)
       else
          call add_hypv_ak(f_,d_v_rhs,lbg1,lbg2)
       endif
    else
        call add_hypv_block(f_,d_v_rhs,lbg1,lbg2)
    endif

    d_z_rhs = cmplx(0.0,0.0)
    if (arakawa_zv) then
       if (hyp_on_h) then 
          call add_hypz_ak(h_,d_z_rhs,lbg1,lbg2)
       else  
          call add_hypz_ak(f_,d_z_rhs,lbg1,lbg2)
       endif  
    else
        call add_hypz_block(f_,d_z_rhs,lbg1,lbg2)
    endif

    d_kperp_rhs = cmplx(0.0,0.0)
    if ((hyp_x.gt.0.0).or.(hyp_y.gt.0.0).or.(hyp_perp.gt.0.0).or.(GyroLES)) then
       !/todo:check if hyp_on_h should affect perpendicular hyperdiffusions!
       call add_dfdxy(f_,d_kperp_rhs,lbg1,lbg2)
    endif

    !Collisions works with f
    coll_rhs = cmplx(0.0,0.0)                 
    if (collision_op.ne.'none') then
       call equ_collisions(f_,coll_rhs,.true.)
    endif

    !Gradient and curvature
    gc_rhs = cmplx(0.0,0.0)
    call add_dgdxy(g_1, gc_rhs, ptr1, pdg1di, pdg1dj,lbg1,lbg2)
    call add_dchidxy_orig(emfields,ptr4, gc_rhs,ptr2,ptr3,lbg1,lbg2,.true.)

    g_rhs  = cmplx(0.0,0.0)
    call add_dchidxy_orig(emfields,ptr4, g_rhs,ptr2,ptr3,lbg1,lbg2,.false.)
    
    c_rhs = cmplx(0.0,0.0)
    c_rhs = gc_rhs - g_rhs   

    !Parallel term
    p_rhs = cmplx(0.0,0.0)
    if (arakawa_zv) then
       call equ_dzv(h_,p_rhs,lbg1,lbg2) !f_ is h
    else
        call equ_dfdzv(f_,p_rhs,lbg1,lbg2)
        call add_dchidz(emfields, ptr4, p_rhs, lbg1, lbg2)
    end if

    do n=ln1,ln2
        call integral_vp(cfgamma1m(:,:,:,:,:,n),cfgamma1_nvp(:,:,:,:,n))
        call integral_vp(g_ls2(:,:,:,:,:,n),g_1_nvp(:,:,:,:,n))
        call integral_vp(h_ls2(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,n),h_nvp(:,:,:,:,n))
        call integral_vp(g_rhs(:,:,:,:,:,n),g_rhs_nvp(:,:,:,:,n))
        call integral_vp(d_v_rhs(:,:,:,:,:,n),d_v_rhs_nvp(:,:,:,:,n))
        call integral_vp(d_z_rhs(:,:,:,:,:,n),d_z_rhs_nvp(:,:,:,:,n))
        call integral_vp(d_kperp_rhs(:,:,:,:,:,n),d_kperp_rhs_nvp(:,:,:,:,n))
        call integral_vp(coll_rhs(:,:,:,:,:,n),coll_rhs_nvp(:,:,:,:,n))
        call integral_vp(p_rhs(:,:,:,:,:,n),p_rhs_nvp(:,:,:,:,n))
        call integral_vp(c_rhs(:,:,:,:,:,n),c_rhs_nvp(:,:,:,:,n))
    enddo

    !TERM 1  FREE ENERGY      
    
    !W 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v6d_spec(:,:,:,:,n,1) = (conjg(g_1_nvp(:,:,:,:,n)))*g_1_nvp(:,:,:,:,n)/(2.*cfgamma1_nvp(:,:,:,:,n))
        call integral_w(v6d_spec(:,:,:,:,n,1),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,1))
    enddo                     
    
    
    if (n_spec.eq.1) then
    ! Electrostatic
        Do n=ln1,ln2
            e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
            v7d_spec(:,:,:,:,:,n,1) = cfgamma2(:,:,:,:,:,n)*g_1(:,:,:,:,:,n)/2.
            Call integral_vpw(v7d_spec(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
            if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
            call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,1))
        Enddo
    else
        ! Electrostatic with modification
        call get_electrostatic(fields_ls2,cfgamma3)
        Do n=ln1,ln2
            if(p_has_00_mode) cfgamma3(0,0,:,n)=cmplx(0.0,0.0)
            call sum_int_3d(cfgamma3(:,:,:,n),fe_tot(n,2,1))
        Enddo
    endif 
    
   !FE
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v6d_spec(:,:,:,:,n,2) = (conjg(h_nvp(:,:,:,:,n)))*g_1_nvp(:,:,:,:,n)/(2.*(cfgamma1_nvp(:,:,:,:,n)))
        call integral_w(v6d_spec(:,:,:,:,n,2),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,3,1))
    enddo                     
    
    ! TERM 2  DRIVE
    ! W
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v6d_spec(:,:,:,:,n,3) = (conjg(g_1_nvp(:,:,:,:,n)))*g_rhs_nvp(:,:,:,:,n)/(cfgamma1_nvp(:,:,:,:,n))
        call integral_w(v6d_spec(:,:,:,:,n,3),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,2))
    enddo    
    ! E
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*g_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,2))
    enddo       
    ! FE
    do n =ln1,ln2
        fe_tot(n,3,2) = fe_tot(n,1,2) + fe_tot(n,2,2)
    enddo
    
        
    !TERM 3 Hyp Z       
    !W 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v6d_spec(:,:,:,:,n,5) = (conjg(g_1_nvp(:,:,:,:,n)))*d_z_rhs_nvp(:,:,:,:,n)/(cfgamma1_nvp(:,:,:,:,n))
        call integral_w(v6d_spec(:,:,:,:,n,5),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,3))
    enddo
    !E 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,3) = cfgamma2(:,:,:,:,:,n)*d_z_rhs(:,:,:,:,:,n)
        call integral_w(v7d_spec(:,:,:,:,:,n,3),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,3))
    enddo    
    !FE
    do n =ln1,ln2
        fe_tot(n,3,3) = fe_tot(n,1,3) + fe_tot(n,2,3)
    enddo 

    !TERM 4 Hyp V 
    !W
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v6d_spec(:,:,:,:,n,7) = (conjg(g_1_nvp(:,:,:,:,n)))*d_v_rhs_nvp(:,:,:,:,n)/(cfgamma1_nvp(:,:,:,:,n))
        call integral_w(v6d_spec(:,:,:,:,n,7),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,4))
    enddo
    !E
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,4) = cfgamma2(:,:,:,:,:,n)*d_v_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,4),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,4))
    enddo    
    !FE
    do n =ln1,ln2
        fe_tot(n,3,4) = fe_tot(n,1,4) + fe_tot(n,2,4)
    enddo    

    !TERM 5 hyp perp 
    !W 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v6d_spec(:,:,:,:,n,9) = (conjg(g_1_nvp(:,:,:,:,n)))*d_kperp_rhs_nvp(:,:,:,:,n)/(cfgamma1_nvp(:,:,:,:,n))
        call integral_w(v6d_spec(:,:,:,:,n,9),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,5))
    enddo
    !E
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,5) = cfgamma2(:,:,:,:,:,n)*d_kperp_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,5),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,5))
    enddo    
    !FE
    do n =ln1,ln2
        fe_tot(n,3,5) = fe_tot(n,1,5) + fe_tot(n,2,5)
    enddo    
 
    !TERM 6 Collisions  
    if (collision_op.ne.'none') then
    ! W collisions
        do n =ln1,ln2
             e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
             v6d_spec(:,:,:,:,n,11) = (conjg(g_1_nvp(:,:,:,:,n)))*coll_rhs_nvp(:,:,:,:,n)/(cfgamma1_nvp(:,:,:,:,n))
             call integral_w(v6d_spec(:,:,:,:,n,11),e_n_temp(:,:,:,n))
             if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
             call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,6))
        enddo            
    ! E collisions
        do n =ln1,ln2
            e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
            v7d_spec(:,:,:,:,:,n,6) = cfgamma2(:,:,:,:,:,n)*coll_rhs(:,:,:,:,:,n)
            call integral_vpw(v7d_spec(:,:,:,:,:,n,6),e_n_temp(:,:,:,n))
            if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
            call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,6))
        enddo
       ! FE collisions
        do n =ln1,ln2
            fe_tot(n,3,6) = fe_tot(n,1,6) + fe_tot(n,2,6)
        enddo
     else
        do n =ln1,ln2
             fe_tot(n,1,6) = 0.0
             fe_tot(n,2,6) = 0.0
             fe_tot(n,3,6) = 0.0
        enddo
     endif

    !TERM 7 curvature 
    !W 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v6d_spec(:,:,:,:,n,13) = (conjg(g_1_nvp(:,:,:,:,n)))*c_rhs_nvp(:,:,:,:,n)/(cfgamma1_nvp(:,:,:,:,n))
        call integral_w(v6d_spec(:,:,:,:,n,13),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,7))
    enddo
    !E
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,7) = cfgamma2(:,:,:,:,:,n)*c_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,7),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,7))
    enddo    
    !FE
    do n =ln1,ln2
        fe_tot(n,3,7) = fe_tot(n,1,7) + fe_tot(n,2,7)
    enddo    
 
    !TERM 8 Parallel 
    !W 
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v6d_spec(:,:,:,:,n,15) = (conjg(g_1_nvp(:,:,:,:,n)))*p_rhs_nvp(:,:,:,:,n)/(cfgamma1_nvp(:,:,:,:,n))
        call integral_w(v6d_spec(:,:,:,:,n,15),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,1,8))
    enddo
    !E
    do n =ln1,ln2
        e_n_temp(:,:,:,n)=cmplx(0.0,0.0)
        v7d_spec(:,:,:,:,:,n,8) = cfgamma2(:,:,:,:,:,n)*p_rhs(:,:,:,:,:,n)
        call integral_vpw(v7d_spec(:,:,:,:,:,n,8),e_n_temp(:,:,:,n))
        if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
        call sum_int_3d(e_n_temp(:,:,:,n),fe_tot(n,2,8))
    enddo    
  
    if (hyp_on_h) then
       do n =ln1,ln2
          fe_tot(n,1,8) = fe_tot(n,1,8) - fe_tot(n,1,4) - fe_tot(n,1,3)
          fe_tot(n,2,8) = fe_tot(n,2,8) - fe_tot(n,2,4) - fe_tot(n,2,3)
       enddo
    else
       do n =ln1,ln2
          if (.not.arakawa_zv) fe_tot(n,1,8) = fe_tot(n,1,8)-fe_tot(n,1,4)-fe_tot(n,1,3) !eliminated hyperdiffusions
          if (.not.arakawa_zv) fe_tot(n,2,8) = fe_tot(n,2,8)-fe_tot(n,2,4)-fe_tot(n,2,3) !eliminated hyperdiffusions
       enddo
    endif
    !FE
    do n =ln1,ln2
        fe_tot(n,3,8) = fe_tot(n,1,8) + fe_tot(n,2,8)
    enddo    
 
    deallocate(v7d_spec)
    deallocate(g_rhs)
    deallocate(h_ls2)
    deallocate(gc_rhs)
    deallocate(p_rhs)
    deallocate(c_rhs)
    deallocate(d_v_rhs)
    deallocate(d_z_rhs)
    deallocate(d_kperp_rhs)
    deallocate(coll_rhs)
    deallocate(cfgamma1)
    deallocate(cfgamma2)
    deallocate(cfgamma3)
 
End Subroutine calc_spectral_mod

!!!************************************************************************************!!!

  Subroutine calc_finalize_diag_transfer(which_model)

    Integer, intent(in) :: which_model
    Integer :: n,o,t

    If (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
       IF (.not.with_nvp) THEN
            Do o=1,3
                 Do n=ln1,ln2
                    close(FE_TF_FILE(n,o,which_model))
                Enddo
            Enddo
       ENDIF

       Do t=1,8
          Do o=1,3
             Do n=ln1,ln2
                close(FE_SP_FILE(n,o,t,which_model))
             Enddo
          Enddo
       enddo
    Endif

    select case(which_model)
    case(1)
       IF (.not.with_nvp) deallocate(tf_cumul_1)
       deallocate(spec_cumul_1)
    case(2)
       IF (.not.with_nvp) deallocate(tf_cumul_2)
       deallocate(spec_cumul_2)
    case(3)
       IF (.not.with_nvp) deallocate(tf_cumul_3)
       deallocate(spec_cumul_3)
    case(4)
       IF (.not.with_nvp) deallocate(tf_cumul_4)
       deallocate(spec_cumul_4)
    case(5)
       IF (.not.with_nvp) deallocate(tf_cumul_5)
       deallocate(spec_cumul_5)
    case(6)
       IF (.not.with_nvp) deallocate(tf_cumul_6)
       deallocate(spec_cumul_6)
    case(7)
       IF (.not.with_nvp) deallocate(tf_cumul_7)
       deallocate(spec_cumul_7)
    case default
    end select

  End subroutine calc_finalize_diag_transfer


!!!*****************************************************************************************!!!

  Subroutine write_transfer(n_shells,which_model)

    integer, intent(in) :: which_model, n_shells
    integer :: ls1,ls2,n,o
  

    select case(which_model)
    case(1,2,3,4,5,6) 
    IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
       Do o=1,3
          Do n=ln1,ln2
             Write(FE_TF_FILE(n,o,which_model),*) '#', time
             Do ls2=1,n_shells
                Do ls1=1,n_shells
                   Write(FE_TF_FILE(n,o,which_model),"(2I4,2ES12.4,ES12.4)") ls2,ls1,k_value(ls2),&
                        k_value(ls1),fe_tf_tot(ls2,ls1,n,o)
                Enddo
                Write(FE_TF_FILE(n,o,which_model),*)
             Enddo
             Write(FE_TF_FILE(n,o,which_model),*)
             call flush(FE_TF_FILE(n,o,which_model))
          enddo
       Enddo
    Endif 

   case(7) 
    IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
        do o=1,3    
            Do n=ln1,ln2
                Write(FE_TF_FILE(n,o,which_model),*) '#', time
                Do ls2=1,n_shells
                    Do ls1=1,n_shells
                        Write(FE_TF_FILE(n,o,which_model),"(2I4,2ES12.4,3ES12.4)") ls2,ls1,k_value(ls2),&
                            k_value(ls1),fe_tf_tot(ls2,ls1,n,o),fe_tf_tot(ls2,ls1,n,o+3),&
                                     fe_tf_tot(ls2,ls1,n,o) + fe_tf_tot(ls2,ls1,n,o+3)
                    Enddo
                    Write(FE_TF_FILE(n,o,which_model),*)
                Enddo
                Write(FE_TF_FILE(n,o,which_model),*)
                call flush(FE_TF_FILE(n,o,which_model))
            Enddo
        Enddo
   Endif
   case default
   End Select

  End subroutine write_transfer
!!!*****************************************************************************************!!!

  Subroutine write_spectral(n_shells,which_model)

    integer, intent(in) :: n_shells, which_model
    integer :: ls1,n,o,t
    real, dimension(n_shells) :: w_k_value,delta_k_value
   
    select case(which_model)
    
    case(1,2)
        w_k_value(:) = k_value(:)
        delta_k_value(:) = k_value(2)
    
    case(3,4,5,6)
        do ls1=1,n_shells
            if (ls1.eq.1) then
                w_k_value(ls1) = k_value(ls1)/2.
                delta_k_value = k_value(ls1)
            else 
                w_k_value(ls1) = (k_value(ls1) + k_value(ls1-1))/2.
                delta_k_value(ls1) = (k_value(ls1) - k_value(ls1-1))
            endif
        enddo
    end select

    IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
       do t = 1,8
          Do o=1,3
             Do n=ln1,ln2
                Do ls1=1,n_shells
              !  Write(FE_SP_FILE(n,o,t,which_model),"(I4,4ES12.4)") ls1, k_value(ls1), w_k_value(ls1), delta_k_value(ls1), fe_tot(ls1,n,o,t)
                    Write(FE_SP_FILE(n,o,t,which_model),"(I4,2ES12.4)") ls1, k_value(ls1), fe_tot(ls1,n,o,t)
                Enddo
                Write(FE_SP_FILE(n,o,t,which_model),*)
                call flush(FE_SP_FILE(n,o,t,which_model))
             enddo
          enddo
!         write(*,*) "curvature and parallel w", sum(fe_tot(:,n,1,7)),sum(fe_tot(:,n,1,8))
!         write(*,*) "curvature and parallel e", sum(fe_tot(:,n,2,7)),sum(fe_tot(:,n,2,8))
!         write(*,*) "hyp_z and hyp_v", sum(fe_tot(:,n,3,3)),sum(fe_tot(:,n,3,4))
!         write(*,*) "collisoins", sum(fe_tot(:,n,3,6))
!         write(*,*) "gradient", sum(fe_tot(:,n,3,2))
!         write(*,*) "free energy", sum(fe_tot(:,n,3,1))
       enddo
    Endif

  End subroutine write_spectral


!!!******************************************************************************************

  Subroutine calc_n_shells_ky(ls1,g_1,g_ls1)
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) :: g_1
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(out) :: g_ls1
    integer, intent(in) :: ls1
    integer :: j
    g_ls1=cmplx(0.0,0.0)

    do j=lj1,lj2
       if (ls1.eq.1) then
          if ((ky(j).ge.(0)).and.(ky(j).lt.(delta_ky*ls1))) then
             g_ls1(:,j,:,:,:,:) = g_1(:,j,:,:,:,:)
          endif
       else 
          if ((ky(j).ge.(delta_ky*(ls1-1))).and.(ky(j).lt.(delta_ky*ls1))) then
             g_ls1(:,j,:,:,:,:) = g_1(:,j,:,:,:,:)
          endif
       endif
    enddo

  End Subroutine calc_n_shells_ky

!!!******************************************************************************************

  Subroutine calc_n_shells_kx(ls1,g_1,g_ls1)
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) :: g_1
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(out) :: g_ls1
    integer, intent(in) :: ls1
    integer :: i

    !Calculate the shell splitting
    !delta_kx of the shells
    g_ls1=cmplx(0.0,0.0)
    do i=li1,li2
       if (ls1.eq.1) then
          if ((abs(kx(i)).eq.(0))  ) then
             g_ls1(i,:,:,:,:,:) = g_1(i,:,:,:,:,:)
          endif
       else if ((ls1.gt.1).and.(ls1.le.(n_shells_kx)) ) then
          if ((abs(kx(i)).ge.(delta_kx*(ls1-1))).and.(abs(kx(i)).lt.(delta_kx*ls1))  ) then
             g_ls1(i,:,:,:,:,:) = g_1(i,:,:,:,:,:)
          endif
       endif
    enddo
  End Subroutine calc_n_shells_kx


!!!******************************************************************************************

  Subroutine calc_n_shells_kperp(ls1,g_1,g_ls1)

    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) :: g_1
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(out) :: g_ls1
    integer, intent(in) :: ls1
    integer :: i,j,k

    !Calculate the shell splitting
    !delta_k of the shells       
    g_ls1=cmplx(0.0,0.0)
    do k=lk1,lk2
       do j=lj1,lj2
          do i=li1,li2
             if (ls1.eq.1) then
                if ((kperp(i,j,k).ge.(0)).and.(kperp(i,j,k).le.(delta_kperp*real(ls1)))) then
                   g_ls1(i,j,k,:,:,:) = g_1(i,j,k,:,:,:)
                endif
             else if ((ls1.gt.1).and.(ls1.le.(n_shells_kperp)) ) then
                if ((kperp(i,j,k).gt.(real(ls1-1)*delta_kperp)).and.(kperp(i,j,k).le.(delta_kperp*real(ls1)))) then
                   g_ls1(i,j,k,:,:,:) = g_1(i,j,k,:,:,:)
                endif
             endif
          enddo
       enddo
    enddo

  End Subroutine calc_n_shells_kperp

!!!******************************************************************************************

  Subroutine calc_n_shells_log_kperp(k0,ls1,g_1,g_ls1)

    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) :: g_1
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(out) :: g_ls1
    integer, intent(in) :: ls1
    real, intent(in) :: k0
    integer :: i,j,k

    !Calculate the shell splitting
    !delta_k of the shells       
    g_ls1=cmplx(0.0,0.0)
    do k=lk1,lk2
       do j=lj1,lj2
          do i=li1,li2
            if (ls1.eq.1) then
                if ((kperp(i,j,k).ge.(0)).and.(kperp(i,j,k).le.(k0*2**real((ls1-1)/5.)))) then
                    g_ls1(i,j,k,:,:,:) = g_1(i,j,k,:,:,:)
                endif
            else if ((ls1.gt.1).and.(ls1.lt.n_shells_log_kperp)) then
                if ((kperp(i,j,k).gt.(k0*2**real((ls1-2)/5.))).and.(kperp(i,j,k).le.(k0*2**real((ls1-1)/5.)))) then
                    g_ls1(i,j,k,:,:,:) = g_1(i,j,k,:,:,:)
                endif
            else
                if ((kperp(i,j,k).gt.(k0*2**real((ls1-2)/5.)))) then
                    g_ls1(i,j,k,:,:,:) = g_1(i,j,k,:,:,:)
                endif
            endif
         enddo
       enddo
    enddo

  End Subroutine calc_n_shells_log_kperp

!**********************************************************************************************

  Subroutine calc_n_shells_log_kperp_pn(k0,ls1,g_1,g_ls1_p,g_ls1_n)

    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2),intent(in) :: g_1
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2),intent(out) :: g_ls1_p, g_ls1_n
    integer, intent(in) :: ls1
    real, intent(in) :: k0
    integer :: i,j,k,l,m

    !Calculate the shell splitting
    !delta_k of the shells       
    g_ls1_p=cmplx(0.0,0.0)
    g_ls1_n=cmplx(0.0,0.0)
    do k=lk1,lk2
       do j=lj1,lj2
          do i=li1,li2
            if (ls1.eq.1) then
                if ((kperp(i,j,k).ge.(0)).and.(kperp(i,j,k).le.(k0*2**real((ls1-1)/5.)))) then
                    do m=lm1,lm2
                        do l=ll1,ll2
                            if (real(g_1(i,j,k,l,m)).ge.0) then
                                g_ls1_p(i,j,k,l,m) = cmplx(real(g_1(i,j,k,l,m)),0.0)
                            else 
                                g_ls1_n(i,j,k,l,m) = cmplx(real(g_1(i,j,k,l,m)),0.0)
                            endif
                        enddo
                    enddo
                endif
            else if ((ls1.gt.1).and.(ls1.lt.n_shells_log_kperp)) then
                if ((kperp(i,j,k).gt.(k0*2**real((ls1-2)/5.))).and.(kperp(i,j,k).le.(k0*2**real((ls1-1)/5.)))) then
                    do m=lm1,lm2
                        do l=ll1,ll2
                            if (real(g_1(i,j,k,l,m)).ge.0) then    
                                g_ls1_p(i,j,k,l,m) = cmplx(real(g_1(i,j,k,l,m)),0.0)
                            else 
                                g_ls1_n(i,j,k,l,m) = cmplx(real(g_1(i,j,k,l,m)),0.0)
                            endif
                        enddo
                    enddo  
                endif    
            else
                if ((kperp(i,j,k).gt.(k0*2**real((ls1-2)/5.)))) then
                    do m=lm1,lm2
                        do l=ll1,ll2
                            if (real(g_1(i,j,k,l,m)).ge.0) then
                                g_ls1_p(i,j,k,l,m) = cmplx(real(g_1(i,j,k,l,m)),0.0)
                            else 
                                g_ls1_n(i,j,k,l,m) = cmplx(real(g_1(i,j,k,l,m)),0.0)
                            endif
                        enddo
                    enddo
                endif
            endif
         enddo
       enddo
    enddo

  End Subroutine calc_n_shells_log_kperp_pn


!!!******************************************************************************************

Subroutine calc_n_shells_log_kperp_1(ls1,g_1,g_ls1)
    
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) :: g_1
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(out) :: g_ls1
    integer, intent(in) :: ls1
    integer :: i,j,k
    
    !Calculate the shell splitting
    !Lambda 1 and Lambda 2, define the size of the first 2 shells.
    !delta_k of the shells (excluding the first two shells)        
    g_ls1=cmplx(0.0,0.0)
        do k=lk1,lk2
            do j=lj1,lj2
                do i=li1,li2
                    if (ls1.eq.1) then
                        if ((kperp(i,j,k).ge.(0)).and.(kperp(i,j,k).le.(lambda1))) then
                            g_ls1(i,j,k,:,:,:) = g_1(i,j,k,:,:,:)
                        endif
                    else if (ls1.eq.2) then
                        if ((kperp(i,j,k).gt.(lambda1)).and.(kperp(i,j,k).le.(lambda1+lambda2))) then
                             g_ls1(i,j,k,:,:,:) = g_1(i,j,k,:,:,:)
                        endif
                    else if (ls1.eq.3) then
                        if ((kperp(i,j,k).gt.(lambda1+lambda2)).and.(kperp(i,j,k).le.((lambda1+lambda2)*2**real((ls1-2)/5.)))) then
                            g_ls1(i,j,k,:,:,:) = g_1(i,j,k,:,:,:)
                        endif
                    else if ((ls1.gt.3).and.(ls1.lt.n_shells_log_kperp_1)) then
                        if ((kperp(i,j,k).gt.((lambda1+lambda2)*2**real((ls1-3)/5.))).and.&
                            (kperp(i,j,k).le.((lambda1+lambda2)*2**real((ls1-2)/5.)))) then
                            g_ls1(i,j,k,:,:,:) = g_1(i,j,k,:,:,:)
                        endif
                    else
                        if ((kperp(i,j,k).gt.((lambda1+lambda2)*2**real((ls1-3)/5.)))) then
                            g_ls1(i,j,k,:,:,:) = g_1(i,j,k,:,:,:)
                        endif 
                    endif
                enddo
            enddo
        enddo   
            
End Subroutine calc_n_shells_log_kperp_1


!!!******************************************************************************************
  Subroutine calc_n_shells_log_ky(k0_y,ls1,g_1,g_ls1)

    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) :: g_1
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(out) :: g_ls1
    integer, intent(in) :: ls1
    real, intent(in) :: k0_y
    integer :: j

    !Calculate the shell splitting
    !delta_k of the shells       
    g_ls1=cmplx(0.0,0.0)
    do j=lj1,lj2
        if (ls1.eq.1) then
            if ((ky(j).ge.(0)).and.(ky(j).le.(k0_y*2**real((ls1-1)/6.)))) then
                g_ls1(:,j,:,:,:,:) = g_1(:,j,:,:,:,:)
            endif
        else if ((ls1.gt.1).and.(ls1.lt.n_shells_log_ky)) then
            if ((ky(j).gt.(k0_y*2**real((ls1-2)/6.))).and.(ky(j).le.(k0_y*2**real((ls1-1)/6.)))) then
                g_ls1(:,j,:,:,:,:) = g_1(:,j,:,:,:,:)
            endif
        else
            if ((ky(j).gt.(k0_y*2**real((ls1-2)/6.)))) then
                g_ls1(:,j,:,:,:,:) = g_1(:,j,:,:,:,:)
            endif
        endif
    enddo

  End Subroutine calc_n_shells_log_ky


!!!******************************************************************************************

  Subroutine calc_n_shells_log_kperp_zonal(k0,ls1,g_1,g_ls1)
    !check why the 2 procedures dont give the same result
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(in) :: g_1
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2),intent(out) :: g_ls1
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)  :: g_1_phi
    complex, dimension(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)  :: g_1_del
    complex, dimension(li1:li2,ll1:ll2, lm1:lm2, ln1:ln2)  :: g_phi
    complex, dimension(li1:li2,lj1:lj2,lbz:ubz,lbv:ubv,lbw:ubw,ln1:ln2) :: f_del_1
    complex, dimension(li1:li2)  :: fields_phi
    complex, dimension(li1:li2,lj1:lj2,lbz:ubz)  :: fields_del
    complex, dimension(li1:li2,lj1:lj2,lbz:ubz,1:n_fields)  :: fields_del_1
    integer, intent(in) :: ls1
    real, intent(in) :: k0
    integer :: i,j,k,l,m,n

    !Calculate the shell splitting
    !delta_k of the shells       
    g_ls1=cmplx(0.0,0.0)
    g_1_phi = cmplx(0.0,0.0)
    g_phi = cmplx(0.0,0.0)
    fields_phi = cmplx(0.0,0.0)
    fields_del(:,:,:) = emfields(:,:,:,1)
    f_del_1 = cmplx(0.0,0.0)
    
    g_1_del = g_1
    if (p_has_0_mode) then
        do  n=ln1,ln2
            do m = lm1,lm2
                do l =ll1,ll2
                   do i=li1,li2
                      g_phi(i,l,m,n) = Sum(g_1(i,lj1,lk1:lk2,l,m,n)*geom%jacobian(pi1,pj1,lk1:lk2)) / (nz0*geom%avg_jaco)
                   Enddo
                Enddo
            Enddo
        enddo
       do n=ln1,ln2
            do m =lm1,lm2
                do l =ll1,ll2
                    Call sum_to_all_complex(g_phi(:,l,m,n), Size(g_phi(:,l,m,n)), mpi_comm_z)
                enddo
            enddo
        enddo
        Do n=ln1,ln2
            do m=lm1,lm2
                do l=ll1,ll2         
                    Do k=lk1,lk2
                        g_1_del(:,lj1,k,l,m,n) = g_1_del(:,lj1,k,l,m,n) - g_phi(:,l,m,n)
                        g_1_phi(:,lj1,k,l,m,n) = g_phi(:,l,m,n)
                    End Do
                enddo
            enddo
        enddo
  
        do i = li1,li2
            fields_phi(i) = Sum(emfields(i,lj1,lk1:lk2,1)*geom%jacobian(pi1,pj1,lk1:lk2)) / (nz0*geom%avg_jaco)
        enddo
        Call sum_to_all_complex(fields_phi, Size(fields_phi), mpi_comm_z)
       do k=lk1,lk2
            fields_del(:,lj1,k) = fields_del(:,lj1,k) - fields_phi
        End Do
    endif

    call calc_aux_fields(g_1_del, fields_del_1, f_del_1,.false.)
    ! Compare fields_del, with fields_del_1
!    counter = 0
!    if (counter.eq.0) then
!    Do k=lk1,lk2
!        do j=lj1,lj2
!            do i=li1,li2
!        if (mype.eq.0.and.j.eq.0) write(*,*) "checking zonal flow components",&
!             & i,j,k,fields_del_1(i,j,k,1) + fields_phi(i) - emfields(i,j,k,1),&
!             & fields_del(i,j,k) - emfields(i,j,k,1) + fields_phi(i)
!if (mype.eq.0.and.j.ne.0) write(*,*) "checking non zonal flow components",&
!             & i,j,k,fields_del_1(i,j,k,1) - emfields(i,j,k,1), &
!             & fields_del(i,j,k) -emfields(i,j,k,1)
!            enddo
!        enddo
!    enddo
!    counter = 1
!    endif


    do k=lk1,lk2
       do j=lj1,lj2
          do i=li1,li2
            if (ls1.eq.1) then
                g_ls1(i,j,k,:,:,:) = g_1_phi(i,j,k,:,:,:)
            else if (ls1.eq.2) then
                if ((kperp(i,j,k).ge.(0)).and.(kperp(i,j,k).le.(k0*2**real((ls1-2)/5.)))) then
                    g_ls1(i,j,k,:,:,:) = g_1_del(i,j,k,:,:,:)
                endif
            else if ((ls1.gt.2).and.(ls1.lt.n_shells_log_kperp)) then
                if ((kperp(i,j,k).gt.(k0*2**real((ls1-3)/5.))).and.(kperp(i,j,k).le.(k0*2**real((ls1-2)/5.)))) then
                    g_ls1(i,j,k,:,:,:) = g_1_del(i,j,k,:,:,:)
                endif
            else
                if ((kperp(i,j,k).gt.(k0*2**real((ls1-3)/5.)))) then
                    g_ls1(i,j,k,:,:,:) = g_1_del(i,j,k,:,:,:)
                endif
            endif
         enddo
       enddo
    enddo

  End Subroutine calc_n_shells_log_kperp_zonal

!!!******************************************************************!!!
!!!******************************** Locality transfer functions *****
!********************************************************************
Subroutine calc_initialize_diag_transfer_3d(n_shells,which_model)

    Implicit None
    Integer, intent(in) :: n_shells, which_model
    Integer :: n,o 
    Character(len=4) :: xstring
    Character(len=80) :: label, header
    Character(len=40), dimension(3) :: tf_spec_3d_file_label

    tf_spec_3d_file_label = (/'/w_transfer_3d_ ',&
         &'/e_transfer_3d_ ','/fe_transfer_3d_'/)

    select case(which_model)
    case(1)
       allocate(tf_cumul_3d_1(n_shells,n_shells,n_shells,ln1:ln2,3))
       tf_cumul_3d_1 = 0
       transfer_counter_3d(1) = 0
       label ='ky_'
       Write(header,*) '#', kymax
    case(2)
       allocate(tf_cumul_3d_2(n_shells,n_shells,n_shells,ln1:ln2,3))
       tf_cumul_3d_2 = 0
       transfer_counter_3d(2) = 0
       label ='kx_'
       Write(header,*) '#'
    case(4)
       allocate(tf_cumul_3d_4(n_shells,n_shells,n_shells,ln1:ln2,3))
       tf_cumul_3d_4 = 0
       transfer_counter_3d(4) = 0
       label ='log_kperp_'
       Write(header,*) '#', k0, kperp_max
    case(6)
       allocate(tf_cumul_3d_6(n_shells,n_shells,n_shells,ln1:ln2,3))
       tf_cumul_3d_6 = 0
       transfer_counter_3d(6) = 0
       label ='log_kperp1_'
       Write(header,*) '#', lambda1, kperp_max
    case default
       return
    end select

    Write(xstring, '(i4)') n_shells

    IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
       Do o=1,3
          Do n=ln1,ln2
             call get_unit_nr(FE_TF_SPEC_3D_FILE(n,o,which_model))
             OPEN(FE_TF_SPEC_3D_FILE(n,o,which_model), file=trim(diagdir)//trim(tf_spec_3d_file_label(o))&
                  //trim(adjustl(label))//trim(spec(n)%name)//trim(file_extension),form= 'FORMATTED', &
                  status=filestat, position=filepos)
             Write(FE_TF_SPEC_3D_FILE(n,o,which_model),*) trim(header)
          enddo
       enddo
    endif

End subroutine calc_initialize_diag_transfer_3d

Subroutine calc_initialize_diag_transfer_3d_h(n_shells,which_model)

    Implicit None
    Integer, intent(in) :: n_shells, which_model
    Integer :: n
    Character(len=4) :: xstring
    Character(len=80) :: label, header

    select case(which_model)
    case(1)
       allocate(tf_cumul_3d_1(n_shells,n_shells,n_shells,ln1:ln2,1))
       tf_cumul_3d_1 = 0
       transfer_counter_3d(1) = 0
       label ='ky_'
       Write(header,*) '#', kymax
    case(2)
       allocate(tf_cumul_3d_2(n_shells,n_shells,n_shells,ln1:ln2,1))
       tf_cumul_3d_2 = 0
       transfer_counter_3d(2) = 0
       label ='kx_'
       Write(header,*) '#'
    case(4)
        allocate(tf_cumul_3d_4(n_shells,n_shells,n_shells,ln1:ln2,1))
        tf_cumul_3d_4 = 0
        transfer_counter_3d(4) = 0
        label ='log_kperp_'
        Write(header,*) '#', k0, kperp_max
     case(6)
        allocate(tf_cumul_3d_6(n_shells,n_shells,n_shells,ln1:ln2,1))
        tf_cumul_3d_6 = 0
        transfer_counter_3d(6) = 0
        label ='log_kperp1_'
        Write(header,*) '#', lambda1, kperp_max
     case default
        return
     end select
     
     Write(xstring, '(i4)') n_shells
     
     IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) THEN
        Do n=ln1,ln2
           call get_unit_nr(FE_TF_SPEC_3D_FILE(n,1,which_model))
           OPEN(FE_TF_SPEC_3D_FILE(n,1,which_model), file=trim(diagdir)//'/fe_transfer_3d_'&
                &//trim(adjustl(label))//trim(spec(n)%name)//trim(file_extension),&
                &form= 'FORMATTED', status=filestat, position=filepos)
           Write(FE_TF_SPEC_3D_FILE(n,1,which_model),*) trim(header)
        enddo
    Endif
End Subroutine calc_initialize_diag_transfer_3D_h

Subroutine calc_diagnostic_transfer_3d(n_shells,which_model)

    integer, intent(in) :: n_shells, which_model
    integer  :: ls1, ls2, ls3, o
    complex, dimension(:,:,:),pointer :: ptr2
    integer :: lbg1,lbg2,stage,n    
    complex, dimension(:,:,:,:),pointer :: ptr1, ptr3
    complex, dimension(:,:,:,:,:,:), pointer :: ptr4
    real, dimension(n_shells,n_shells,n_shells,ln1:ln2,3):: fe_tf_tot_3d
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ln1:ln2):: e_n_temp
    
    allocate(g_ls1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(g_ls2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(g_ls3(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(nlt_ls2_ls3_g(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(nlt_ls2_ls3_f(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(cfgamma1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(cfgamma2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(fe_tf_term(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2)) 
    allocate(f_ls1(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    allocate(f_ls2(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    allocate(f_ls3(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    allocate(fields_ls1(li1:li2, lj1:lj2, lbz:ubz,n_fields)) 
    allocate(fields_ls2(li1:li2, lj1:lj2, lbz:ubz,n_fields)) 
    allocate(fields_ls3(li1:li2, lj1:lj2, lbz:ubz,n_fields)) 
    
    ptr1 => null()
    ptr2 => null()
    ptr3 => null()
    ptr4 => null()
    lbg1 = 1
    lbg2 = lklmn0
    stage = 0

    ! This initialization that should be here and cant be inside the ls1,ls2 loop
    fe_tf_tot_3d = 0.0

    do ls3=1,n_shells
       select case(which_model)
          case(1) 
             call calc_n_shells_ky(ls3,g_1,g_ls3)
          case(2) 
             call calc_n_shells_kx(ls3,g_1,g_ls3)
          case(4) 
             call calc_n_shells_log_kperp(k0,ls3,g_1,g_ls3)
          case(6) 
             call calc_n_shells_log_kperp_1(ls3,g_1,g_ls3)
          case default
          end select
        call calc_aux_fields(g_ls3, fields_ls3, f_ls3,.false.)    
        do ls2=1,n_shells
            if (which_model.eq.1) call calc_n_shells_ky(ls2,g_1,g_ls2)
            if (which_model.eq.2) call calc_n_shells_kx(ls2,g_1,g_ls2)
            if (which_model.eq.4) call calc_n_shells_log_kperp(k0,ls2,g_1,g_ls2)
            if (which_model.eq.6) call calc_n_shells_log_kperp_1(ls2,g_1,g_ls2)
            call calc_aux_fields(g_ls2, fields_ls2, f_ls2,.false.)    
            nlt_ls2_ls3_g = cmplx(0.0,0.0) 
            nlt_ls2_ls3_f = cmplx(0.0,0.0)

            CALL this_nonlinear_term%add(g_ls2, ptr1, fields_ls3,&
                 &ptr2, ptr3, nlt_ls2_ls3_g, lbg1, lbg2, stage)
            CALL this_nonlinear_term%add(g_ls3, ptr1, fields_ls2,&
                 &ptr2, ptr3, nlt_ls2_ls3_f, lbg1, lbg2, stage)
            do ls1=1,n_shells
                if (which_model.eq.1) call calc_n_shells_ky(ls1,g_1,g_ls1)
                if (which_model.eq.2) call calc_n_shells_kx(ls1,g_1,g_ls1)
                if (which_model.eq.4) call calc_n_shells_log_kperp(k0,ls1,g_1,g_ls1)
                if (which_model.eq.6) call calc_n_shells_log_kperp_1(ls1,g_1,g_ls1)
                call calc_aux_fields(g_ls1, fields_ls1, f_ls1,.false.)    
                call get_cfgamma(f_ls1,fields_ls1,cfgamma1,cfgamma2)
                fe_tf_term = cmplx(0.0,0.0)
                ! Calculte the tranfers of E, W and FE different for ions than for electrons
                do n =ln1,ln2
                    e_n_temp(:,:,:,n) = cmplx(0.0,0.0)
                    fe_tf_term(:,:,:,:,:,n,1) = cfgamma1(:,:,:,:,:,n)*nlt_ls2_ls3_g(:,:,:,:,:,n)
                    call integral_vpw(fe_tf_term(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
                    if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
                    call sum_int_3d(e_n_temp(:,:,:,n),fe_tf_tot_3d(ls3,ls2,ls1,n,1))
                enddo
                do n=ln1,ln2
                    e_n_temp(:,:,:,n) = cmplx(0.0,0.0)
                    fe_tf_term(:,:,:,:,:,n,2) = cfgamma2(:,:,:,:,:,n)*nlt_ls2_ls3_f(:,:,:,:,:,n)
                    call integral_vpw(fe_tf_term(:,:,:,:,:,n,2),e_n_temp(:,:,:,n))
                    if(p_has_00_mode) e_n_temp(0,0,:,n)=cmplx(0.0,0.0)
                    call sum_int_3d(e_n_temp(:,:,:,n),fe_tf_tot_3d(ls3,ls2,ls1,n,2))
                enddo
                do n=ln1,ln2
                    fe_tf_tot_3d(ls3,ls2,ls1,n,3) = fe_tf_tot_3d(ls3,ls2,ls1,n,1) + fe_tf_tot_3d(ls3,ls2,ls1,n,2)
                enddo
            enddo
        enddo
    enddo        
    
    deallocate(g_ls1) 
    deallocate(g_ls2) 
    deallocate(g_ls3) 
    deallocate(nlt_ls2_ls3_g) 
    deallocate(nlt_ls2_ls3_f) 
    deallocate(cfgamma1) 
    deallocate(cfgamma2) 
    deallocate(fe_tf_term) 
    deallocate(f_ls1) 
    deallocate(f_ls2) 
    deallocate(f_ls3) 
    deallocate(fields_ls1) 
    deallocate(fields_ls2) 
    deallocate(fields_ls3) 
    
    if ( .not.( mod(itime,istep_fe_transfer*avg_window) .eq. 0) ) then
       ! We cumulate in the first case
        if (which_model.eq.1) then
            tf_cumul_3d_1 = fe_tf_tot_3d + tf_cumul_3d_1
            transfer_counter_3d(1) = transfer_counter_3d(1) + 1 
       endif
        if (which_model.eq.2) then
            tf_cumul_3d_2 = fe_tf_tot_3d + tf_cumul_3d_2
            transfer_counter_3d(2) = transfer_counter_3d(2) + 1 
       endif
       if (which_model.eq.4) then
            tf_cumul_3d_4 = fe_tf_tot_3d + tf_cumul_3d_4
            transfer_counter_3d(4) = transfer_counter_3d(4) + 1 
       endif
       if (which_model.eq.6) then
            tf_cumul_3d_6 = fe_tf_tot_3d + tf_cumul_3d_6
            transfer_counter_3d(6) = transfer_counter_3d(6) + 1 
       endif
   endif

    ! Then we do output files
    if ( itime .gt. 0 .and. mod(itime,istep_fe_transfer*avg_window) .eq. 0 ) then
       ! Here we use v2d as a buffer for write routine and average it

       select case(which_model)
       case(1)
          tf_cumul_3d_1 = fe_tf_tot_3d + tf_cumul_3d_1
          transfer_counter_3d(1) = transfer_counter_3d(1) + 1 
          fe_tf_tot_3d =  tf_cumul_3d_1 /real(transfer_counter_3d(1))
       case(2)
          tf_cumul_3d_2 = fe_tf_tot_3d + tf_cumul_3d_2
          transfer_counter_3d(2) = transfer_counter_3d(2) + 1 
          fe_tf_tot_3d =  tf_cumul_3d_2 /real(transfer_counter_3d(2))
       case(4)
          tf_cumul_3d_4 = fe_tf_tot_3d + tf_cumul_3d_4
          transfer_counter_3d(4) = transfer_counter_3d(4) + 1 
          fe_tf_tot_3d =  tf_cumul_3d_4 /real(transfer_counter_3d(4))
       case(6)
          tf_cumul_3d_6 = fe_tf_tot_3d + tf_cumul_3d_6
          transfer_counter_3d(6) = transfer_counter_3d(6) + 1 
          fe_tf_tot_3d =  tf_cumul_3d_6 /real(transfer_counter_3d(6))
       case default
       end select
       
       IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
          Do o=1,3
             Do n=ln1,ln2
                Write(FE_TF_SPEC_3D_FILE(n,o,which_model),*) '#', time
                Do ls3=1,n_shells
                   Do ls2=1,n_shells
                      Write(FE_TF_SPEC_3D_FILE(n,o,which_model),*) 'Transfer spectra T(L1) for shells L2, L3' 
                      Write(FE_TF_SPEC_3D_FILE(n,o,which_model),"(2I4)") ls2, ls3
                      Do ls1=1,n_shells
                         Write(FE_TF_SPEC_3D_FILE(n,o,which_model),"(I4,ES12.4)") ls1, fe_tf_tot_3d(ls3,ls2,ls1,n,o)
                      Enddo
                   Enddo
                Enddo
                call flush(FE_TF_SPEC_3D_FILE(n,o,which_model))
             enddo
          enddo
        Endif
 
        select case(which_model)
        case(1) 
           tf_cumul_3d_1 = 0
           transfer_counter_3d(1) = 0
        case(2) 
           tf_cumul_3d_2 = 0
           transfer_counter_3d(2) = 0
        case(4)
           tf_cumul_3d_4 = 0
           transfer_counter_3d(4) = 0
        case(6)
           tf_cumul_3d_6 = 0
           transfer_counter_3d(6) = 0
        case default
        end select
    endif

  End subroutine calc_diagnostic_transfer_3d



Subroutine calc_diagnostic_transfer_3d_h(n_shells,which_model)

    integer, intent(in) :: n_shells, which_model
    integer  :: ls1, ls2,ls3
    complex, dimension(:,:,:),pointer :: ptr2
    integer :: lbg1,lbg2,stage,n    
    complex, dimension(:,:,:,:),pointer :: ptr1, ptr3
    complex, dimension(:,:,:,:,:,:), pointer :: ptr4
    real, dimension(n_shells,n_shells,n_shells,ln1:ln2,1):: fe_tf_tot_3d
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ln1:ln2):: e_n_temp
    
    allocate(g_ls1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(g_ls2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(g_ls3(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(h_mod_ls1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(h_mod_ls2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(h_mod_ls3(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(nlt_ls2_ls3_h(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(cfgamma(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(cfgamma1(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(cfgamma2(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2)) 
    allocate(fe_tf_term(li1:li2, lj1:lj2, lk1:lk2, ll1:ll2, lm1:lm2, ln1:ln2,2)) 
    allocate(f_ls1(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    allocate(f_ls2(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    allocate(f_ls3(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    allocate(h_ls1(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    allocate(h_ls2(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    allocate(h_ls3(li1:li2, lj1:lj2, lbz:ubz, lbv:ubv, lbw:ubw, ln1:ln2)) 
    allocate(fields_ls1(li1:li2, lj1:lj2, lbz:ubz,n_fields)) 
    allocate(fields_ls2(li1:li2, lj1:lj2, lbz:ubz,n_fields)) 
    allocate(fields_ls3(li1:li2, lj1:lj2, lbz:ubz,n_fields)) 
    
    ptr1 => null()
    ptr2 => null()
    ptr3 => null()
    ptr4 => null()
    lbg1 = 1
    lbg2 = lklmn0
    stage = 0

    ! This initialization that should be here and cant be inside the ls1,ls2 loop
    fe_tf_tot_3d = 0.0

    do ls3=1,n_shells
        if (which_model.eq.1) call calc_n_shells_ky(ls3,g_1,g_ls3)
        if (which_model.eq.2) call calc_n_shells_kx(ls3,g_1,g_ls3)
        if (which_model.eq.4) call calc_n_shells_log_kperp(k0,ls3,g_1,g_ls3)
        if (which_model.eq.6) call calc_n_shells_log_kperp_1(ls3,g_1,g_ls3)
        call calc_aux_fields(g_ls3, fields_ls3, f_ls3,comp_h=.true.,h_out=h_ls3)    
        h_mod_ls3 = h_ls3(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,ln1:ln2) 
        do ls2=1,n_shells
            if (which_model.eq.1) call calc_n_shells_ky(ls2,g_1,g_ls2)
            if (which_model.eq.2) call calc_n_shells_kx(ls2,g_1,g_ls2)
            if (which_model.eq.4) call calc_n_shells_log_kperp(k0,ls2,g_1,g_ls2)
            if (which_model.eq.6) call calc_n_shells_log_kperp_1(ls2,g_1,g_ls2)
            call calc_aux_fields(g_ls2, fields_ls2, f_ls2,comp_h=.true.,h_out=h_ls2)    
            h_mod_ls2 = h_ls2(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,ln1:ln2) 
            nlt_ls2_ls3_h = cmplx(0.0,0.0) 
            call this_nonlinear_term%add(h_mod_ls2, ptr1, fields_ls3,&
                &ptr2, ptr3, nlt_ls2_ls3_h, lbg1, lbg2, stage)
            do ls1=1,n_shells
                if (which_model.eq.1) call calc_n_shells_ky(ls1,g_1,g_ls1)
                if (which_model.eq.2) call calc_n_shells_kx(ls1,g_1,g_ls1)
                if (which_model.eq.4) call calc_n_shells_log_kperp(k0,ls1,g_1,g_ls1)
                if (which_model.eq.6) call calc_n_shells_log_kperp_1(ls1,g_1,g_ls1)
                call calc_aux_fields(g_ls1, fields_ls1, f_ls1,comp_h=.true.,h_out=h_ls1)    
                call get_cfgamma(f_ls1,fields_ls1,cfgamma1,cfgamma2)
                cfgamma = cmplx(0.0,0.0)
                cfgamma = cfgamma1 + cfgamma2
                h_mod_ls1 = h_ls1(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,ln1:ln2) 
            
                fe_tf_term = cmplx(0.0,0.0)
                ! Calculate the tranfers FE 
                do n =ln1,ln2
                    e_n_temp(:,:,:,n) = cmplx(0.0,0.0)
                    fe_tf_term(:,:,:,:,:,n,1) = cfgamma(:,:,:,:,:,n)*nlt_ls2_ls3_h(:,:,:,:,:,n)
                    call integral_vpw(fe_tf_term(:,:,:,:,:,n,1),e_n_temp(:,:,:,n))
                    if(p_has_00_mode) e_n_temp(0,0,:,n) =cmplx(0.0,0.0)
                    call sum_int_3d(e_n_temp(:,:,:,n),fe_tf_tot_3d(ls3,ls2,ls1,n,1))
                enddo
            enddo
        enddo
    enddo        
    
    deallocate(g_ls1) 
    deallocate(g_ls2) 
    deallocate(g_ls3) 
    deallocate(h_mod_ls1) 
    deallocate(h_mod_ls2) 
    deallocate(h_mod_ls3) 
    deallocate(nlt_ls2_ls3_h) 
    deallocate(cfgamma1) 
    deallocate(cfgamma2) 
    deallocate(cfgamma) 
    deallocate(fe_tf_term) 
    deallocate(f_ls1) 
    deallocate(f_ls2) 
    deallocate(f_ls3) 
    deallocate(fields_ls1) 
    deallocate(fields_ls2) 
    deallocate(fields_ls3) 
    deallocate(h_ls1) 
    deallocate(h_ls2) 
    deallocate(h_ls3) 
    
    if ( .not.( mod(itime,istep_fe_transfer*avg_window) .eq. 0) ) then
       ! We cumulate in the first case
       select case(which_model)
       case(1)
          tf_cumul_3d_1 = fe_tf_tot_3d + tf_cumul_3d_1
          transfer_counter_3d(1) = transfer_counter_3d(1) + 1 
       case(2)
          tf_cumul_3d_2 = fe_tf_tot_3d + tf_cumul_3d_2
          transfer_counter_3d(2) = transfer_counter_3d(2) + 1 
       case(4)
          tf_cumul_3d_4 = fe_tf_tot_3d + tf_cumul_3d_4
          transfer_counter_3d(4) = transfer_counter_3d(4) + 1 
       case(6)
          tf_cumul_3d_6 = fe_tf_tot_3d + tf_cumul_3d_6
          transfer_counter_3d(6) = transfer_counter_3d(6) + 1 
       case default
       end select
   endif

    ! Then we do output files
    if ( itime .gt. 0 .and. mod(itime,istep_fe_transfer*avg_window) .eq. 0 ) then
       ! Here we use v2d as a buffer for write routine and average it

       select case(which_model)
       case(1)
          tf_cumul_3d_1 = fe_tf_tot_3d + tf_cumul_3d_1
          transfer_counter_3d(1) = transfer_counter_3d(1) + 1 
          fe_tf_tot_3d =  tf_cumul_3d_1 /real(transfer_counter_3d(1))
       case(2)
          tf_cumul_3d_2 = fe_tf_tot_3d + tf_cumul_3d_2
          transfer_counter_3d(2) = transfer_counter_3d(2) + 1 
          fe_tf_tot_3d =  tf_cumul_3d_2 /real(transfer_counter_3d(2))
       case(4)
          tf_cumul_3d_4 = fe_tf_tot_3d + tf_cumul_3d_4
          transfer_counter_3d(4) = transfer_counter_3d(4) + 1 
          fe_tf_tot_3d =  tf_cumul_3d_4 /real(transfer_counter_3d(4))
       case(6)
          tf_cumul_3d_6 = fe_tf_tot_3d + tf_cumul_3d_6
          transfer_counter_3d(6) = transfer_counter_3d(6) + 1 
          fe_tf_tot_3d =  tf_cumul_3d_6 /real(transfer_counter_3d(6))
       case default
       end select
       
        IF (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
            Do n=ln1,ln2
                Write(FE_TF_SPEC_3D_FILE(n,1,which_model),*) '#', time
                Do ls3=1,n_shells
                    Do ls2=1,n_shells
                        Write(FE_TF_SPEC_3D_FILE(n,1,which_model),*) 'Transfer spectra T(L1) for shells L2, L3' 
                        Write(FE_TF_SPEC_3D_FILE(n,1,which_model),"(2I4)") ls2, ls3
                        Do ls1=1,n_shells
                            Write(FE_TF_SPEC_3D_FILE(n,1,which_model),"(I4,ES12.4)") ls1, fe_tf_tot_3d(ls3,ls2,ls1,n,1)
                        Enddo
                   Enddo
                Enddo
               call flush(FE_TF_SPEC_3D_FILE(n,1,which_model))
           enddo
        Endif
 
       select case(which_model)
       case(1)
          tf_cumul_3d_1 = 0 
          transfer_counter_3d(1) = 0
       case(2)
          tf_cumul_3d_2 = 0 
          transfer_counter_3d(2) = 0
       case(4)
          tf_cumul_3d_4 = 0 
          transfer_counter_3d(4) = 0
       case(6)
          tf_cumul_3d_6 = 0 
          transfer_counter_3d(6) = 0
       case default
       end select
    endif

  End subroutine calc_diagnostic_transfer_3d_h

  Subroutine calc_finalize_diag_transfer_3d(which_model)

    integer,intent(in) :: which_model
    Integer :: n 

    If (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
       Do n=ln1,ln2
          close(FE_TF_SPEC_3D_FILE(n,1,which_model))
          close(FE_TF_SPEC_3D_FILE(n,2,which_model))
          close(FE_TF_SPEC_3D_FILE(n,3,which_model))
       Enddo
    Endif

    if (which_model.eq.1) deallocate(tf_cumul_3d_1)
    if (which_model.eq.2) deallocate(tf_cumul_3d_2)
    if (which_model.eq.4) deallocate(tf_cumul_3d_4)
    if (which_model.eq.6) deallocate(tf_cumul_3d_6)

  End subroutine calc_finalize_diag_transfer_3d

  Subroutine calc_finalize_diag_transfer_3d_h(which_model)

    integer,intent(in) :: which_model
    Integer :: n 

    If (mype.eq.pexyzvwspec(0,0,0,0,0,my_pespec)) then
       Do n=ln1,ln2
          close(FE_TF_SPEC_3D_FILE(n,1,which_model))
       Enddo
    Endif

    if (which_model.eq.1) deallocate(tf_cumul_3d_1)
    if (which_model.eq.2) deallocate(tf_cumul_3d_2)
    if (which_model.eq.4) deallocate(tf_cumul_3d_4)
    if (which_model.eq.6) deallocate(tf_cumul_3d_6)

  End subroutine calc_finalize_diag_transfer_3d_h


!!!**************************************************************************!!!
!!!********************* Subroutine that the module needs ************************!!! 
!!!**************************************************************************!!!

  subroutine integral_vpw(v5d,v3d)
    !This subroutine performs the mat_00 integral and doesnt have explicit species
    !dependence

    implicit none
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2), intent(in) :: v5d
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2),intent(out) :: v3d
    integer :: k,m,l

    v3d=cmplx(0.0,0.0)
    do k=lk1,lk2
       do m=lm1,lm2
          do l=ll1,ll2
             call saxpy(2*li0*lj0,mat_00(pi1,pj1,k,l,m),v5d(li1,lj1,k,l,m),1,v3d(li1,lj1,k),1)
          enddo
       enddo
    enddo

    call my_complex_sum_vw(v3d,size(v3d)) 


  end subroutine integral_vpw

  subroutine integral_vp(v5d,v4d)
    !This subroutine performs the v parallel integral and doesnt have explicit species
    !dependence

    implicit none
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2), intent(in) :: v5d
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,lm1:lm2),intent(out) :: v4d
    integer :: m,l

 
    v4d=cmplx(0.0,0.0)
    
    do m=lm1,lm2
        do l=ll1,ll2
              call saxpy(2*li0*lj0*lk0,vp_weight(l),v5d(li1,lj1,lk1,l,m),1,v4d(li1,lj1,lk1,m),1)
        enddo
    enddo

    call my_complex_sum_v(v4d,size(v4d)) 

  end subroutine integral_vp

  subroutine integral_w(v4d,v3d)
    !This subroutine perform the mu integral, which it is
    !2*pi v_perp d_vperp = pi \hat{B} dmu,  and doesnt have explicit species
    !dependence

    implicit none
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,lm1:lm2), intent(in) :: v4d
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2),intent(out) :: v3d
    real, dimension(pi1:pi2,pj1:pj2,lk1:lk2,lm1:lm2) :: mat_00_mu
    integer :: k,m,j

    v3d=cmplx(0.0,0.0)
    
    Do m=lm1,lm2
        Do k=lk1,lk2
            Do j=pj1,pj2
                mat_00_mu(pi1:pi2,j,k,m) = pi*mu_weight(m)*&
                    &geom%Bfield(pi1:pi2,j,k)
            Enddo
        Enddo
    Enddo

    do m=lm1,lm2
       do k=lk1,lk2
             call saxpy(2*li0*lj0,mat_00_mu(pi1,pj1,k,m),v4d(li1,lj1,k,m),1,v3d(li1,lj1,k),1)
       enddo
    enddo

    call my_complex_sum_w(v3d,size(v3d)) 


  end subroutine integral_w

!!!******************************************************************************************
!!!*********************** Write 2d  ************************

  subroutine write2d(arr2d,m)

    integer, intent(in) :: m
    Real, Dimension(li1:li2,lj1:lj2), intent(in):: arr2d
    Real, Dimension(0:nx0-1, 0:nky0-1)  ::  dest2d, dest2d_linear
    Real, Dimension(0:nx0-1,lj1:lj2):: arr2dx
    Real :: dest1d_linear
    Integer :: i,j,ierr


    if ((my_pev+my_pew+my_pespec).eq.0) then
       ! Gather arr2d over x-Distribution on pex==0.
       do j = lj1, lj2
          Call mpi_gather(arr2d(li1,j), li0, MPI_REAL_TYPE,&
               arr2dx(0,j), li0, MPI_REAL_TYPE,&
               0, mpi_comm_x, ierr)
       enddo


       ! Gather arr2d over y-Distribution on pey==0.
       Call mpi_gather(arr2dx(0,lj1),size(arr2dx),MPI_REAL_TYPE,&
            dest2d(0,0),size(arr2dx),MPI_REAL_TYPE,0,mpi_comm_y,ierr)

       if (mype.eq.0) then

          write(fnumber(m),"(ES12.4)") time
          
          if (nonlinear) then
            write(fnumber(m),"(ES12.4)") 2*sum(sum(dest2d(0:nx0-1,1:nky0-1),1),1) + sum(dest2d(:,0),1)
          else
            dest2d_linear = dest2d  
            if ((x_local).and.(evenx.eq.1)) dest2d_linear(hkx+1,:) = 0.0
            if (p_has_0_mode) dest2d_linear(:,lj1) = 0.5*dest2d_linear(:,lj1)
            dest1d_linear = 2.0*sum(sum(dest2d_linear,1),1)
            write(fnumber(m),"(ES12.4)") dest1d_linear
          endif

          if (evenx.eq.1) then

            do i=nx0/2+1,nx0-1
                do j = 0,nky0-1
                    write(fnumber(m),"(ES12.4)") dest2d(i,j)
                enddo
            enddo

            do i=0,nx0/2-1
                do j=0,nky0-1
                    write(fnumber(m),"(ES12.4)") dest2d(i,j)
                enddo
            enddo
          else

            do i=nx0/2+1,nx0-1
                do j = 0,nky0-1
                    write(fnumber(m),"(ES12.4)") dest2d(i,j)
                enddo
            enddo

            do i=0,nx0/2
                do j=0,nky0-1
                    write(fnumber(m),"(ES12.4)") dest2d(i,j)
                enddo
            enddo
          endif 
         call flush(fnumber(m))
       endif
    endif

    call my_barrier()

  End Subroutine write2d

!!!*********************** Write 2d for species dependent  ************************

  subroutine write2d_spec(arr2d,n,m)

    integer, intent(in) :: n,m
    Real, Dimension(li1:li2,lj1:lj2), intent(in):: arr2d
    Real, Dimension(0:nx0-1, 0:nky0-1)  ::  dest2d, dest2d_linear
    real :: dest1d_linear
    Real, Dimension(0:nx0-1,lj1:lj2):: arr2dx
    Integer :: i,j,ierr


    if ((my_pev+my_pew).eq.0) then
       ! Gather arr2d over x-Distribution on pex==0.
       do j = lj1, lj2
          Call mpi_gather(arr2d(li1,j), li0, MPI_REAL_TYPE,&
               arr2dx(0,j), li0, MPI_REAL_TYPE,&
               0, mpi_comm_x, ierr)
       enddo


       ! Gather arr2d over y-Distribution on pey==0.
       Call mpi_gather(arr2dx(0,lj1),size(arr2dx),MPI_REAL_TYPE,&
            dest2d(0,0),size(arr2dx),MPI_REAL_TYPE,0,mpi_comm_y,ierr)

       if (mype.eq. pexyzvwspec(0,0,0,0,0,my_pespec) ) then
          write(fnumber_spec(n,m),"(ES12.4)") time

          if (nonlinear) then
            write(fnumber_spec(n,m),"(ES12.4)") 2*sum(sum(dest2d(0:nx0-1,1:nky0-1),1),1) + sum(dest2d(:,0),1)
          else
            dest2d_linear = dest2d  
            if ((x_local).and.(evenx.eq.1)) dest2d_linear(hkx+1,:) = 0.0
            if (p_has_0_mode) dest2d_linear(:,lj1) = 0.5*dest2d_linear(:,lj1)
            dest1d_linear = 2.0*sum(sum(dest2d_linear,1),1)
            write(fnumber_spec(n,m),"(ES12.4)") dest1d_linear
          endif

          if (evenx.eq.1) then

            do i=nx0/2+1,nx0-1
                do j = 0,nky0-1
                    write(fnumber_spec(n,m),"(ES12.4)") dest2d(i,j)
                enddo
            enddo
            do i=0,nx0/2-1
                do j=0,nky0-1
                    write(fnumber_spec(n,m),"(ES12.4)") dest2d(i,j)
                enddo
            enddo

        else
            do i=nx0/2,nx0-1
                do j = 0,nky0-1
                    write(fnumber_spec(n,m),"(ES12.4)") dest2d(i,j)
                enddo
            enddo
            do i=0,nx0/2-1
                do j=0,nky0-1
                    write(fnumber_spec(n,m),"(ES12.4)") dest2d(i,j)
                enddo
            enddo
        endif
        call flush(fnumber_spec(n,m))
       endif
    endif

    call my_barrier()

  End Subroutine write2d_spec

!!!******************************************************************************************

  subroutine twoD_array(v6d,v2d_out)

    implicit none
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,ln1:ln2), intent(in) :: v6d
    real, dimension(li1:li2,lj1:lj2), intent(out) :: v2d_out
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2) :: v3d
    real, dimension(li1:li2,lj1:lj2,lk1:lk2) :: v3d_bis

    v3d = cmplx(0.0)
    v3d_bis = 0.0    
    v2d_out = 0.0

    call energy_integral(v6d,v3d)

    v3d_bis=real(v3d)
    if ((x_local).and.(evenx.eq.1)) v3d_bis(hkx+1,:,:) = 0.0
    if (p_has_0_mode) v3d_bis(li1,lj1,:) = 0.0

    call sum_int_z(v3d_bis,v2d_out)

  end subroutine twoD_array

!!!******************************************************************************************

  subroutine twoD_array_spec(v5d,v2d_out)

    implicit none
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2), intent(in) :: v5d
    real, dimension(li1:li2,lj1:lj2), intent(out) :: v2d_out
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2) :: v3d
    real, dimension(li1:li2,lj1:lj2,lk1:lk2) :: v3d_bis

    v3d = cmplx(0.0)
    v3d_bis = 0.0    
    v2d_out = 0.0

    call integral_vpw(v5d,v3d)

    v3d_bis=real(v3d)
    if ((x_local).and.(evenx.eq.1)) v3d_bis(hkx+1,:,:) = 0.0
    if (p_has_0_mode) v3d_bis(li1,lj1,:) = 0.0

    call sum_int_z(v3d_bis,v2d_out)

  end subroutine twoD_array_spec

!!!**************************************************************************!!!
!!!*********************** Calculate the filter  ************************
  subroutine filter(f_in,f_out)

    implicit none

    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,ln1:ln2), intent(in) :: f_in
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,ln1:ln2), intent(out) :: f_out
    Integer :: fkx, fky

    ! Filter in the kx direction, it gives you the indice of the maximum value of the kx mode you keep
    fkx = int((fracx*hkx))
    ! Filter in the ky direction, it gives you the indice of the maximum value of the ky mode you keep 
    !fky = int((fracy*lj2))
    fky = int(fracy*(nky0-1))

    !Copy the input function to the output function FGS
    f_out = f_in
    !Filter FGS in the kx direction for modes higher than fkx and lower than -fkx 

    f_out( li1 + fkx + 1 : li2 - fkx,:,:,:,:,:) = CMPLX(0.,0.)

    IF (fky.ge.lj1 .and. fky.ge.lj2) then
    ELSE IF (fky.ge.lj1.and.fky.lt.lj2) then
       f_out(:,fky + 1 : lj2,:,:,:,:) = CMPLX(0.,0.)
    ELSE IF (fky.ge.lj1.and.fky.eq.lj2) then
    ELSE
       f_out(:,:,:,:,:,:)= CMPLX(0.,0.)
    ENDIF

  End Subroutine filter

!!!*********************************************************************************

  SUBROUTINE get_cfgamma(f_in,fields_in,cfgamma1,cfgamma2)
    Implicit None
    complex,dimension(li1:li2,lj1:lj2,lbz:ubz,1:n_fields),intent(in) :: fields_in
    complex,dimension(li1:li2,lj1:lj2,lbz:ubz,lbv:ubv,lbw:ubw,ln1:ln2),intent(in) :: f_in
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,ln1:ln2), intent(out):: cfgamma2,cfgamma1
    complex, dimension(li1:li2, lj1:lj2):: phi_bar
    integer :: k,l,m,n

    DO n=ln1,ln2
       DO m=lm1,lm2
          DO l=ll1,ll2
             DO k=lk1,lk2
                CALL gyro_average_ff(fields_in(:,:,k,1),phi_bar,k,m,n)
                cfgamma2(:,:,k,l,m,n)=spec(n)%temp*spec(n)%dens*&
                     (spec(n)%charge*CONJG(phi_bar(:,:))/spec(n)%temp)
                cfgamma1(:,:,k,l,m,n)=spec(n)%temp*spec(n)%dens/fm(pi1,pj1,k,l,m,pn1)*&
                     (CONJG(f_in(:,:,k,l,m,n)))
             END DO
          END DO
       END DO
    END DO

  END SUBROUTINE get_cfgamma

  !WARNING: f_in and fields_in are not used in this subroutine
  !does this make sense?????
  SUBROUTINE get_cfgamma_mod(f_in,fields_in,cfgamma1m)
    Implicit None
    complex,dimension(li1:li2,lj1:lj2,lbz:ubz,1:n_fields),intent(in) :: fields_in
    complex,dimension(li1:li2,lj1:lj2,lbz:ubz,lbv:ubv,lbw:ubw,ln1:ln2),intent(in) :: f_in
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,ln1:ln2), intent(out):: cfgamma1m
!    complex, dimension(li1:li2, lj1:lj2):: phi_bar
    integer :: k,l,m,n

    DO n=ln1,ln2
       DO m=lm1,lm2
          DO l=ll1,ll2
             DO k=lk1,lk2
                !not used...
                !CALL gyro_average_ff(fields_in(:,:,k,1),phi_bar,k,m,n)
                cfgamma1m(:,:,k,l,m,n)=fm(pi1,pj2,k,l,m,pn1)/(spec(n)%temp*spec(n)%dens) 
             END DO
          END DO
       END DO
    END DO

  END SUBROUTINE get_cfgamma_mod


  SUBROUTINE get_electrostatic(fields_in,cfgamma3)
    Implicit None
    complex,dimension(li1:li2,lj1:lj2,lbz:ubz,1:n_fields),intent(in) :: fields_in
    complex, dimension(li1:li2,lj1:lj2,lk1:lk2,ln1:ln2), intent(out):: cfgamma3
    Real :: b2spec, k2, b2
    integer :: i,j,k,n

    Do k=lk1,lk2
       Do j = lj1,lj2
          Do i = li1,li2
!\todo could be precomputed during initialization
             k2 = geom%gii(pi1,pj1,k)*ki(i)**2+2.0*geom%gij(pi1,pj1,k)*&
                  & ki(i)*kj(j)+geom%gjj(pi1,pj1,k)*kj(j)**2
             b2 = k2/(geom%Bfield(pi1,pj1,k)**2)

!\todo not performant: innermost loop over last index
             do n=ln1,ln2
                b2spec = b2*spec(n)%mass*spec(n)%temp/(spec(n)%charge**2)
                cfgamma3(i,j,k,n)=(spec(n)%dens*spec(n)%charge**2&
                     /(2.0*spec(n)%temp))*(1.0 - gamma0(b2spec))*&
                     CONJG(fields_in(i,j,k,1))*fields_in(i,j,k,1)
             enddo
          enddo
       enddo
    enddo

  END SUBROUTINE get_electrostatic

End Module diag_Gyro_LES

