FUNCTION energyspect_info

  RETURN, {$
    type      : 'energy',$
    title     : 'Energy spectra plots',$
    help_text : ['Plots spectra for the free energy, energy drive, dissipation, etc. '],$
    ext_vars  : [ $
    	    	 ['kx_range','0', '  Specifies plot range in kx e.g., [-0.5,0.5] (Default: max,min)' ],$
    	    	 ['ky_range','0', '  Specifies plot range in ky e.g., [-1.5,1.5] (Default: max,min)' ]$
                  ] }

END

;######################################################################

PRO energyspect_init, diag

  COMMON global_vars

  e = (*diag).external_vars
  IF NOT KEYWORD_SET(*e.kx_range) THEN (*e.kx_range)=[999,999]
  IF NOT KEYWORD_SET(*e.ky_range) THEN (*e.ky_range)=[999,999]

  ;IF N_ELEMENTS(*e.vars) LT 1 THEN (*e.vars) = [0,1]
  IF par.lx EQ 0 THEN BEGIN
PRINT, (*diag).name + ' warning: ky dependent lx, results may be unphysical'
lxorig = series.lx
series.lx = 100.0
dxorig = series.dx
series.dx = series.lx / par.nx0
  ENDIF

  ; x direction 0/+1 in real space (local/global BC),
  ; 0/-1 in Fourier (odd/even nkx0)
  add_x_pt = 0
  IF NOT (par.nkx0 MOD 2) THEN add_x_pt = -1
  ;IF *e.kx2x AND par.x_local AND NOT (par.nkx0 MOD 2) THEN add_x_pt = 1

  ; y direction: +1 in real space (BC), -1 in Fourier (zero mode)
  ;IF *e.ky2y THEN add_y_pt = 1 ELSE add_y_pt = -1
  add_y_pt = -1

  i = set_internal_vars(diag,{$
    add_x_pt  : add_x_pt,$
    add_y_pt  : add_y_pt,$
    kx_range  : *e.kx_range,$
    ky_range  : *e.ky_range,$
    xind      : FINDGEN(par.nkx0+add_x_pt),$
    yind      : FINDGEN(2*par.nky0+add_y_pt),$
    time_id   : PTR_NEW(),$
    data_kx   : PTR_NEW(),$
    data_ky   : PTR_NEW(),$
    data_id   : PTR_NEW()})

    ;fft_format_en, kxky=*e.vars
    (*i).xind = ((*i).xind - par.nkx0 + par.nkx0 / 2 + 1) * $
      (2.0 * !PI / series.lx)
    (*i).yind = ((*i).yind - par.nky0 + 1) * (2.0 * !PI / series.ly)

IF par.lx EQ 0 THEN BEGIN
 series.lx = lxorig
 series.dx = dxorig
ENDIF

END

;######################################################################

PRO energyspect_loop, diag

  COMMON global_vars

  ;jac_fac = (*series.geom).jacobian

  i = (*diag).internal_vars

  nkx0 = par.nkx0
  nky0 = par.nky0

  data_kx = FLTARR(nkx0+(*i).add_x_pt,6,/NOZERO) ;num_vars
  data_ky = FLTARR(2*nky0+(*i).add_y_pt,6,/NOZERO) ;num_vars

  data = FLTARR(nkx0+(*i).add_x_pt,2*nky0+(*i).add_y_pt,$
    6,/NOZERO) ;num_vars

  data_zavg = FLTARR(nkx0,nky0,$
    6,/NOZERO) ;num_vars

  FOR var=0, 6 - 1 DO BEGIN ;num_vars
   FOR kxind=0, nkx0-1 DO BEGIN
    FOR kyind=0, nky0-1 DO BEGIN
     ;data_zavg(kxind,kyind,var)=TOTAL((*energy[(*i).vars[var]].kxky)[kxind,kyind,*]*$
;		(*series.geom).jacobian[0,*])      
     data_zavg(kxind,kyind,var)=TOTAL((*energy[var].kxky)[kxind,kyind,*]$
		*(*series.geom).jacobian[*])/TOTAL((*series.geom).jacobian)
     ;data_zavg(kxind,kyind,var)=(*energy[(*i).vars[var]].kxky)[kxind,kyind,8]
    ENDFOR 
   ENDFOR
  ENDFOR

;;;;;;;;CONVERSION FROM ENERGYCONT;;;;;;;;;;;;;;;;;;;  
;;;;;;;;CONVERSION FROM ENERGYCONT;;;;;;;;;;;;;;;;;;;  
;;;;;;;;CONVERSION FROM ENERGYCONT;;;;;;;;;;;;;;;;;;;  
  FOR var = 0, 6 - 1 DO BEGIN ;num_vars
        data[nkx0-nkx0/2-1,nky0-1,var] = $
          data_zavg[0:nkx0-nkx0/2-1,*,var]
          ;ABS((*energy[(*i).vars[var]].kxky)[0:nkx0-nkx0/2-1,*,(*i).zind])
        ; negative kx, positive ky
        data[0,nky0-1,var] =$
          data_zavg[nkx0/2+1:nkx0-1,*,var]
           ;ABS((*energy[(*i).vars[var]].kxky)[nkx0/2+1:nkx0-1,*,(*i).zind])
        ; negative ky
        data[0,0,var] = REVERSE(REVERSE(data[*,nky0-1:*,var],2),1)
  ENDFOR
;;;;;;;;CONVERSION FROM ENERGYCONT;;;;;;;;;;;;;;;;;;;  
;;;;;;;;CONVERSION FROM ENERGYCONT;;;;;;;;;;;;;;;;;;;  
;;;;;;;;CONVERSION FROM ENERGYCONT;;;;;;;;;;;;;;;;;;;  


  FOR var = 0, 6 - 1 DO BEGIN ;num_vars
   data_kx[*,var]=TOTAL(data[*,*,var],2)
   data_ky[*,var]=TOTAL(data[*,*,var],1)
  ENDFOR

  ;(*i).data_id = store_step((*i).data_id,data)
  (*i).time_id = store_step((*i).time_id,energy_time)
  (*i).data_id = time_avg((*i).data_id,data,energy_time,/fwd)
  (*i).data_kx = time_avg((*i).data_kx,data_kx,energy_time,fwd=0)
  (*i).data_ky = time_avg((*i).data_ky,data_ky,energy_time,fwd=0)

END

;######################################################################

FUNCTION get_pos_enspect, aspect_ratio, ivar, n_vars, x_as_y

  pos = FLTARR(10)  

  IF aspect_ratio LT 1.0 THEN BEGIN ; vertical
    lmargin = 6*!D.Y_CH_SIZE & rmargin = 8*!D.X_CH_SIZE
    bmargin = 4*!D.Y_CH_SIZE & tmargin = 1*!D.Y_CH_SIZE
    between = 0.5*!D.Y_CH_SIZE
    height=(!D.Y_SIZE-tmargin-bmargin-(n_vars-1)*between)/n_vars
    width = !D.X_SIZE-rmargin-lmargin
    IF x_as_y THEN $
      IF (aspect_ratio LE FLOAT(height)/width) THEN height = width*aspect_ratio $
      ELSE width = height/aspect_ratio

    res = convert_coord([lmargin,lmargin+width],$
                        [!D.Y_SIZE-tmargin-ivar*between-(ivar+1.0)*height,$
                         !D.Y_SIZE-tmargin-ivar*(height+between)],$
                        /device,/to_normal)
    pos[0] = res[0,0] & pos[1] = res[1,0] ; CONTOUR left, bottom
    pos[2] = res[0,1] & pos[3] = res[1,1] ; CONTOUR right, top  
    res = convert_coord([2*!D.X_CH_SIZE],$
                        [!D.Y_SIZE-tmargin-ivar*(height+between)-!D.Y_CH_SIZE],$
                        /device,/to_normal)
    pos[4] = res[0,0] & pos[5] = res[1,0] ; Var_names
    res = convert_coord([lmargin+width+1*!D.X_CH_SIZE, lmargin+width+3*!D.X_CH_SIZE],$
                        [!D.Y_SIZE-tmargin-ivar*between-(ivar+1.0)*height,$
                         !D.Y_SIZE-tmargin-ivar*(height+between)],$
                        /device,/to_normal)
    pos[6] = res[0,0] & pos[7] = res[1,0] ; COLORBAR left, bottom
    pos[8] = res[0,1] & pos[9] = res[1,1] ; COLORBAR right, top         
    
  ENDIF ELSE BEGIN
    lmargin = 6*!D.Y_CH_SIZE & rmargin = 3*!D.X_CH_SIZE
    bmargin = 4*!D.Y_CH_SIZE & tmargin = 5*!D.Y_CH_SIZE  
    between = 8*!D.X_CH_SIZE
    height=!D.Y_SIZE-tmargin-bmargin
    width =(!D.X_SIZE-rmargin-lmargin-(n_vars-1)*between)/n_vars
    
    IF x_as_y THEN $
      IF (aspect_ratio LE FLOAT(height)/width) THEN height = FIX(width*aspect_ratio) $
      ELSE width = FIX(height/aspect_ratio)/n_vars
    
    res = convert_coord([lmargin+ivar*(width+between),$
                         lmargin+ivar*(width+between)+width],$
                        [!D.Y_SIZE-tmargin-height, !D.Y_SIZE-tmargin],$
                        /device,/to_normal)
    
    pos[0] = res[0,0] & pos[1] = res[1,0] ; CONTOUR left, bottom
    pos[2] = res[0,1] & pos[3] = res[1,1] ; CONTOUR right, top  
    res = convert_coord([lmargin-3*!D.X_CH_SIZE+ivar*(width+between)],$
                        [!D.Y_SIZE-2.5*!D.Y_CH_SIZE],$
                        /device,/to_normal)
    pos[4] = res[0,0] & pos[5] = res[1,0] ; Var_names           
    res = convert_coord([lmargin+ivar*(width+between),$
                         lmargin+ivar*(width+between)+width],$
                        [!D.Y_SIZE-3*!D.Y_CH_SIZE,$
                         !D.Y_SIZE-!D.Y_CH_SIZE],$
                        /device,/to_normal) 
    pos[6] = res[0,0] & pos[7] = res[1,0] ; COLORBAR left, bottom
    pos[8] = res[0,1] & pos[9] = res[1,1] ; COLORBAR right, top 
  ENDELSE
  
  RETURN, pos
    
END

;######################################################################

PRO energyspect_output, diag

  COMMON global_vars

  i = (*diag).internal_vars
  n_vars = 6  ;num_vars

  data_kx = time_avg((*i).data_kx,/avg,fwd=0,$
    tarr=tarr)
  data_ky = time_avg((*i).data_ky,/avg,fwd=0,$
    tarr=tarr)
  xind=(*i).xind
  yind=(*i).yind

  ;dsizex=size(xind)
  ;dsizey=size(yind)
  ;print,'dsizex',dsizex[1]
  ;print,'dsizey',dsizey[1]
  ;data_t=time_avg((*i).data_id,/avg,fwd=1,tarr=time_arr)
  ;print,SIZE(data_t)
  ;print,"number of time steps:", series.step_count
  ;allsum=FLTARR(series.step_count,n_vars)
  ;FOR var=0,n_vars-1 DO BEGIN
  ;  allsum[*,var]=TOTAL(TOTAL(data_t[*,*,var,*],2),1)
  ;ENDFOR
  ;print,"allsum out"
  ;FOR var=0,series.step_count-1 DO BEGIN
  ;  print,time_arr[var],allsum[var,0],allsum[var,1],allsum[var,2],allsum[var,1]-allsum[var,2] 
  ;ENDFOR

  time=store_step((*i).time_id,/get)
  ntime=N_ELEMENTS(time)
  ;print,"time",*time[0],*time[ntime-1]
  tmin=*time[0]
  tmax=*time[ntime-1]
  trange='t=['+STRING(tmin,format='(f6.2)')+','+STRING(tmax,format='(f6.2)')+']'


  IF par.arakawa_zv AND  NOT par.arakawa_cons_bc  THEN BEGIN
  en_var_string = '!6' + ['E!Dtot!N','dE/dt!9!!!6!Dnonconserve!N(Warning: Includes z,v poisson)',$
    'Q (= dE/dt!9!!!6!Ddrive!N)','C (= dE/dt!9!!!6!Dcoll!N)','D (hyp_z,v,x,y) (Warning: missing diss b.c.)',$
    'dE/dt!D NL !N', '1/E dE/dt !D NL !N']
  ENDIF ELSE BEGIN 
  en_var_string = '!6' + ['E!Dtot!N','dE/dt!9!!!6!Dnonconserve!N(Warning: Includes z,v, poisson)',$
    'Q (= dE/dt!9!!!6!Ddrive!N)','C (= dE/dt!9!!!6!Dcoll!N)','D (hyp_z,v,x,y)','dE/dt!D NL !N', '1/E dE/dt !D NL !N']
  ENDELSE

  
  count = 0

  x_as_y = 0
    
  ; identical scaling of x and y axes for sxsy, kxky
  x_as_y = 1

  x_axis = 'k!Dx!N '
  y_axis = 'k!Dy!N '
  
  xtitle = '!6' + x_axis + get_var_string(/rhostr)
  ytitle = '!6' + y_axis + get_var_string(/rhostr)

  csize = 1.41 - (n_vars < 4) * 0.125
  cthick = 3.0
   
  colorarr = [1,3,2,4,6,8] ; to be color consistent with older plots
  xtickformat = 'betterticks'
  ytickformat = 'betterticks'
  set_output, diag, /ps, multi=[0,1,1], ysize=20, xsize=14.4


;Output plot file here.
  ;xrange=[MIN(yind),MAX(yind)]
    IF (*i).ky_range[0] EQ 999 THEN BEGIN
      xrange=[MIN(yind),MAX(yind)]
    ENDIF ELSE BEGIN
      xrange=(*i).ky_range
    ENDELSE


  FOR var = 0, n_vars - 1 DO BEGIN    ; loop over all variables to plot

    IF par.arakawa_zv AND  NOT par.arakawa_cons_bc  THEN BEGIN
      out_string="Warning: dissipative boundary condition not included."
      ;XYOUTS, out_string, CHARSIZE=3.0,ALIGNMENT=0.0
      print,"Warning!!: Dissipation do not include dissipative parallel bc."
    ENDIF

    yrange=[MIN(data_ky[*,var]),MAX(data_ky[*,var])]
    ;print,'yrange',yrange
    ;print,'xrange',xrange
    PLOT, yind, data_ky[*,var], COLOR=colorarr[0], XTITLE=ytitle, $
        XLOG=0,  $
        XTICKLEN=1.0, POSITION=[0.15,0.4,0.9,0.9], $
        YLOG=0,  YRANGE=yrange, $
        TITLE=en_var_string[var], XRANGE=xrange 


    XYOUTS, 0.15,0.28,/NORMAL, COLOR=1,$
        trange, CHARSIZE=1.5*csize,$
        CHARTHICK=cthick

  ENDFOR

  ;OUTPUT nonlinear transfer rate:  NL/E
    yrange=[MIN(data_ky[*,5]/data_ky[*,0]),MAX(data_ky[*,5]/data_ky[*,0])]
    PLOT, yind, data_ky[*,5]/data_ky[*,0], COLOR=colorarr[0], XTITLE=ytitle, $
        XLOG=0,  $
        XTICKLEN=1.0, POSITION=[0.15,0.4,0.9,0.9], $
        YLOG=0,  YRANGE=yrange, $
        TITLE=en_var_string[6], XRANGE=xrange

    XYOUTS, 0.15,0.28,/NORMAL, COLOR=1,$
        trange, CHARSIZE=1.5*csize,$
        CHARTHICK=cthick

  ;xrange=[MIN(xind),MAX(xind)]
    IF (*i).kx_range[0] EQ 999 THEN BEGIN
      xrange=[MIN(xind),MAX(xind)]
    ENDIF ELSE BEGIN
      xrange=(*i).kx_range
    ENDELSE
  FOR var = 0, n_vars - 1 DO BEGIN    ; loop over all variables to plot

    yrange=[MIN(data_kx[*,var]),MAX(data_kx[*,var])]
    PLOT, xind, data_kx[*,var], COLOR=colorarr[0], XTITLE=xtitle, $
        XLOG=0,  $
        XTICKLEN=1.0, POSITION=[0.15,0.4,0.9,0.9], $
        YLOG=0,  YRANGE=yrange, $
        TITLE=en_var_string[var], XRANGE=xrange 

    XYOUTS, 0.15,0.28,/NORMAL, COLOR=1,$
        trange, CHARSIZE=1.5*csize,$
        CHARTHICK=cthick

    IF par.arakawa_cons_bc ne 1 THEN BEGIN
      out_string="Warning: dissipative boundary condition not included."
      ;XYOUTS, out_string, CHARSIZE=3.0
      ;XYOUTS, pos[4], pos[5],/NORMAL, COLOR=1,$
      ;  out_string, CHARSIZE=1.5*csize,$
      ;  CHARTHICK=cthick
      print,"Warning!!: dissipation does not include dissipative parallel bc."
    ENDIF

  ENDFOR

  ;OUTPUT nonlinear transfer rate:  NL/E
     yrange=[MIN(data_kx[*,5]/data_kx[*,0]),MAX(data_kx[*,5]/data_kx[*,0])]
    PLOT, xind, data_kx[*,5]/data_kx[*,0], COLOR=colorarr[0], XTITLE=xtitle, $
        XLOG=0,  $
        XTICKLEN=1.0, POSITION=[0.15,0.4,0.9,0.9], $
        YLOG=0,  YRANGE=yrange, $
        TITLE=en_var_string[6], XRANGE=xrange 

    XYOUTS, 0.15,0.28,/NORMAL, COLOR=1,$
        trange, CHARSIZE=1.5*csize,$
        CHARTHICK=cthick

  ;Output data file
  set_output, diag, dat=[[yind],[data_ky[*,0]],[data_ky[*,1]],[data_ky[*,2]], [data_ky[*,3]], $
              [data_ky[*,4]], [data_ky[*,5]]],commentlines='1) ky, 2) energy, 3) dedt_nc, 4) Q, 5) C, 6) Hyp'
  set_output, diag, dat=[[xind],[data_kx[*,0]],[data_kx[*,1]],[data_kx[*,2]], [data_kx[*,3]], $
              [data_kx[*,4]], [data_kx[*,5]]],commentlines='1) kx, 2) energy, 3) dedt_nc, 4) Q, 5) C, 6) Hyp',/append

  set_output, diag, /reset, eps=(series.step_count EQ 1)


END
