FUNCTION gamdiag_info

  RETURN, {$
    type      : 'mom',$ ;'mom_uni',$
    title     : 'GAM diag',$
    help_text : ['Draws a linear-linear and a log-linear plot of the time trace '+$'
    	'of the real, imaginary and absolute value of selected modes and variable. '+$
    	'The corresponding kx, ky and z values can be found in the legend. '+$
	'Data output will only be done for first choice of var, kxind and zind.'],$    
    ext_vars  : [$		 
    	['var','0','index of the fluctuating quantity to be plotted'+$
	 ' (see variable list)'],$
    	['kxind','0','index of kx modes to be plotted (integer '+$
	 'between 0 and nkx0 - 1; default: 1 if kx_center=0 else 0)'],$
    	['zind','0','index of z-value to be plotted (integer '+$
	 'between 0 and nz0 - 1; default: -1 (average))']]}
 		 
END

;######################################################################

PRO gamdiag_init, diag

  COMMON global_vars
  
  e = (*diag).external_vars

  IF N_ELEMENTS(*e.var)   LT 1 THEN *e.var   = 0
  IF N_ELEMENTS(*e.kxind) LT 1 THEN BEGIN
     IF (par.kx_center NE 0.0) THEN $
        *e.kxind = 0 ELSE *e.kxind = 1
  ENDIF
  IF N_ELEMENTS(*e.zind)  LT 1 THEN *e.zind  = -1 ; averaging
  
  i = set_internal_vars(diag,{$
    vars    : (*e.var)[0],$
    kxind   : *e.kxind,$
    zind    : *e.zind,$    
    time_id : PTR_NEW(),$
    data_id : PTR_NEW()})
    
  IF (par.x_local AND par.y_local) THEN fft_format, kxky=(*e.var)[0] $
  ELSE IF par.y_local THEN fft_format, sxky=(*e.var)[0] $
  ELSE fft_format, kxsy=(*e.var)[0]

END

;######################################################################

PRO gamdiag_loop, diag

  COMMON global_vars

  i = (*diag).internal_vars

  n_kxind = N_ELEMENTS((*i).kxind)
  n_zind = N_ELEMENTS((*i).zind)

  tt_data = DCOMPLEXARR(n_kxind,n_zind,gui.out.n_spec_sel,/NOZERO)
  
  IF (par.x_local AND par.y_local) THEN BEGIN
   FOR isp = 0, gui.out.n_spec_sel - 1 DO BEGIN
     FOR ikx = 0, n_kxind - 1 DO BEGIN
        FOR ins = 0, n_zind - 1 DO BEGIN
           IF (*i).zind[ins] EQ - 1 THEN BEGIN
             tt_data[ikx,ins,isp] = COMPLEX(0.0,0.0)
             FOR nsi = 0, par.nz0 - 1 DO BEGIN
               tt_data[ikx,ins,isp] += $
                 (*mom[isp,(*i).vars].kxky)[(*i).kxind[ikx],$
                   0,nsi]*(*series.geom).jac_norm[nsi]
             ENDFOR
             tt_data[ikx,ins,isp] = tt_data[ikx,ins,isp]/par.nz0
           ENDIF ELSE tt_data[ikx,ins,isp] = $
             (*mom[isp,(*i).vars].kxky)[(*i).kxind[ikx],0,(*i).zind[ins]]
         ENDFOR
       ENDFOR
   ENDFOR
  ENDIF ELSE BEGIN
   IF (par.y_local) THEN BEGIN
    FOR isp = 0, gui.out.n_spec_sel - 1 DO BEGIN
     FOR ikx = 0, n_kxind - 1 DO BEGIN
        FOR ins = 0, n_zind - 1 DO BEGIN
           IF (*i).zind[ins] EQ - 1 THEN BEGIN
             tt_data[ikx,ins,isp] = COMPLEX(0.0,0.0)
             FOR nsi = 0, par.nz0 - 1 DO BEGIN
               tt_data[ikx,ins,isp] += $
                 (*mom[isp,(*i).vars].sxky)[(*i).kxind[ikx],$
                 0,nsi]*(*series.geom).jacobian[ikx,nsi]
             ENDFOR
             tt_data[ikx,ins,isp] = tt_data[ikx,ins,isp]/TOTAL((*series.geom).jacobian[ikx,*])
           ENDIF ELSE tt_data[ikx,ins,isp] = $
             (*mom[isp,(*i).vars].sxky)[(*i).kxind[ikx],0,(*i).zind[ins]]
         ENDFOR
       ENDFOR
    ENDFOR  
   ENDIF ELSE BEGIN
    tt_data = COMPLEX(0.0,0.0)
    FOR isp = 0, gui.out.n_spec_sel - 1 DO BEGIN
     FOR ikx = 0, n_kxind - 1 DO BEGIN
        FOR ins = 0, n_zind - 1 DO BEGIN
           IF (*i).zind[ins] EQ -1 THEN BEGIN
              tt_data[ikx,ins,isp] += $
                 TOTAL((*mom[isp,(*i).vars].kxsy)[(*i).kxind[ikx],*,*]*$
                       (*series.geom).jacobian)/TOTAL((*series.geom).jacobian)
           ENDIF ELSE BEGIN
              FOR iy = 0, par.ny0 - 1 DO BEGIN
                 tt_data[ikx,ins,isp] += $
                    (*mom[isp,(*i).vars].kxsy)[(*i).kxind[ikx],iy,(*i).zind[ins]]/par.ny0
              ENDFOR
           ENDELSE
        ENDFOR
       ENDFOR
    ENDFOR  
   ENDELSE
  ENDELSE

  (*i).time_id = store_step((*i).time_id,mom_time)
  (*i).data_id = store_step((*i).data_id,tt_data)

END

;######################################################################
PRO gamdiag_plot_residual, h, A_R

  maxval = MAX(A_R, min=minval)
  maxxval = MAX(h)
    
; lin - lin plot
  PLOT, h, A_R, COLOR = 1, $
    /XSTYLE, YRANGE=[0,maxval], $
    XRANGE=[0,1.1*maxxval],PSYM=4,$
    TITLE=title, CHARSIZE=csize,$
    xtitle='h', ytitle='A!IR!N',$
    POSITION=[0.15,0.2,0.95,0.9615]

  h = INDGEN(20)/20.*1.1*maxxval
  A_R = 1./(1.+1.6/h)
  OPLOT, h, A_R, COLOR=2
      
END                          ; plot residual level vs. h=q^2/sqrt(r/R)
;######################################################################

PRO gamdiag_plot_result_vs_x, xaxis, fit_res, theo_res

  title = ['A!IR!N','!7c!6!IG!N','!7x!6!IG!N']

  FOR p=0, N_ELEMENTS(fit_res[*,0])-1 DO BEGIN
      
      maxval = MAX([fit_res[p,*],theo_res[p,*]], min=minval)
      
; lin - lin plot
      PLOT, xaxis, fit_res[p,*], COLOR = 1, $
        /XSTYLE, YRANGE=[0,maxval], PSYM=4,$
        YTITLE=title[p], CHARSIZE=csize,$
        xtitle='x/a', title='',$
        POSITION=[0.15,0.2,0.95,0.9615]
      
      OPLOT, xaxis,theo_res[p,*], COLOR=2
      
      plot_legend, [1,2],['GENE','theory'], per_line=2,$
        x_offset=[0.15,0.0,0.95,0.12], psym=[4,0]
      
  ENDFOR
  
END

;###########################################################################

PRO gamdiag_plot_data, diag, time, tt_data, title, legend_str, text_tit,$
                       A=A, B=B

  COMMON global_vars
  i = (*diag).internal_vars

  n_kxind = N_ELEMENTS(tt_data[*,0,0])
  n_zind  = N_ELEMENTS(tt_data[0,*,0])
  n_tind  = N_ELEMENTS(tt_data[0,0,*])

  maxval = MAX(tt_data, min=minval)
  show_theory = (n_kxind*n_zind EQ 1)
    
; lin - lin plot
  PLOT, time, tt_data[0,0,*], COLOR = 1, $
    /XSTYLE, YRANGE=[minval,maxval], $
    TITLE=title,$ ;CHARSIZE=csize,$
    POSITION=[0.15,0.2,0.95,0.9615]
      
  count = 0

  FOR ikx=0, n_kxind-1 DO BEGIN
    FOR ins=0, n_zind-1 DO BEGIN   
     OPLOT, time, tt_data[ikx,ins,*], COLOR = 1 + 3*count
     IF (show_theory) THEN BEGIN
        IF KEYWORD_SET(A) THEN BEGIN
           gamfunct,time,A[*,count],F
           OPLOT, time, F, COLOR = 2 + 3*count, linestyle=2
        ENDIF
        IF KEYWORD_SET(B) THEN BEGIN
           gamfunct,time,B[*,count],F
           OPLOT, time, F, COLOR = 3 + 3*count, linestyle=2     
        ENDIF
     ENDIF
     
;     omega_f = sqrt2/par.major_R
;     gamma_f = omega_f*exp(-q_^2)
;     OPLOT,time,(1-A_R)*exp(-gamma_f*time)*cos(omega_f*time)+A_R, $
;           COLOR = 5 + 4*count,linestyle=2

     count = count + 1 	
   ENDFOR   
  ENDFOR
  
  ; add base line  

  IF !Y.CRANGE[0] LE 0 AND !Y.CRANGE[1] GE 0 THEN $
    oplot, !X.CRANGE, [0,0], color=1

  ; print legend
  IF (show_theory) THEN BEGIN
     plot_legend, INDGEN(3)+1,['Sim.','[Sugama06]','Fit'], $
      x_offset=[0.15,0.0,0.95,0.12]
  ENDIF ELSE plot_legend, INDGEN(n_kxind*n_zind) + 1, $
      legend_str, x_offset=[0.15,0.0,0.95,0.12]
  plot_info_str, diag

END ; plot time_trace

;######################################################################

PRO gamdiag_output, diag

  COMMON global_vars

  i = (*diag).internal_vars
  
  n_kxind = N_ELEMENTS((*i).kxind)
  n_zind = N_ELEMENTS((*i).zind)
  
  time = store_step((*i).time_id,/get,/reg_array)
  full_data = REFORM(store_step((*i).data_id,/get,/reg_array),[n_kxind,$
    n_zind,gui.out.n_spec_sel,series.step_count],/OVERWRITE)

  tt_data = DOUBLE(full_data)

  csize = 1.42

 ; build legend_str  
  legend_str = STRARR(n_kxind*n_zind)   
  
  fit_res = DBLARR(3,n_kxind*n_zind,/NOZERO)
  theo_res = DBLARR(3,n_kxind*n_zind,/NOZERO)
  coord_arr = DBLARR(2,n_kxind*n_zind,/NOZERO)

  FOR isp = 0, gui.out.n_spec_sel - 1 DO BEGIN
     sp=(*gui.out.spec_select)[isp]

    count = 0
    FOR ikx = 0, n_kxind-1 DO BEGIN
      FOR ins = 0, n_zind-1 DO BEGIN

      ;normalize to t=0 value
      tt_data[ikx,ins,isp,*] = tt_data[ikx,ins,isp,*]/tt_data[ikx,ins,isp,0]

      ;construct some values
      IF (par.x_local) THEN BEGIN
         x_R = par.trpeps
         q_ = par.q0
         IF ((*i).zind[ins] GE 0) THEN $
            Bfield = (*series.geom).Bfield[ins] $
         ELSE $
            Bfield = TOTAL((*series.geom).Bfield*(*series.geom).jac_norm)/par.nz0
         kx=(*series.kx)[ikx,0]
      ENDIF ELSE BEGIN
         x_R = (0.5 + (-series.lx/2.+(*i).kxind[ikx]*series.dx)*par.rhostar)*$
               par.minor_r/par.major_R
         q_ = (*series.geom).q((*i).kxind[ikx])
         IF ((*i).zind[ins] GE 0) THEN $
            Bfield = (*series.geom).Bfield[ikx,ins] $
         ELSE Bfield = TOTAL((*series.geom).Bfield*(*series.geom).$
         jacobian[ikx,*])/TOTAL((*series.geom).jacobian[ikx,*])
         kx=2.0*!PI/par.lx
      ENDELSE

      ;get Sugama's result
      get_sugama_res,q_,x_R*par.major_R,par.major_R,kx,Bfield,A

      ;use Sugama result for fit
      B = A
      weights = FLTARR(series.step_count)+1.0
      yfit = CURVEFIT(time,REFORM(tt_data[ikx,ins,isp,*]),weights,$
                      B, SIGMA, FUNCTION_NAME='gamfunct')

      ; print comparison
      print, 'x/R=',rm0es(x_R), ', q=',rm0es(q_)
      print_result, A, B, TOTAL(tt_data[ikx,ins,isp,*])/series.step_count,$
        MEDIAN(tt_data[ikx,ins,isp,*])

      ;store fit results
      fit_res[*,count] = B
      coord_arr[0,count] = q_
      coord_arr[1,count] = x_R
      theo_res[*,count] = A

      IF par.x_local THEN legend_str[count] = $
        '('+rm0es((*series.kx)[(*i).kxind[ikx],0],prec=3) $
      ELSE legend_str[count] = $
        '('+rm0es((*series.lxarr)[0]*(-0.5+(*i).kxind[ikx]/$
         par.nx0),prec=3)
	
      IF (*i).zind[ins] EQ -1 THEN $
        legend_str[count] = legend_str[count] + ')' $
      ELSE legend_str[count] = legend_str[count] + ',' + $
        rm0es((*par.z)[(*i).zind[ins]],prec=2)+')'

      count += 1
    ENDFOR
  ENDFOR  
     
  IF series.step_count GT 1 THEN BEGIN
    set_output, diag, sp, /ps, multi=[0,1,1], charsize=csize
   
;   %%%%%%%%%% Real part %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    title = get_var_string((*i).vars,/fancy)+'(t)/'+$
            get_var_string((*i).vars,/fancy)+'(t=0)' 
    text_tit = get_var_string((*i).vars)+'(t)/'+$
            get_var_string((*i).vars)+'(t=0)' 

    gamdiag_plot_data, diag, time, TOTAL(tt_data[*,*,isp,*],3),$
      title, legend_str, text_tit,$
      A=theo_res,B=fit_res

    IF (n_kxind*n_zind GT 1) THEN BEGIN
        gamdiag_plot_residual, (sqrt(coord_arr[1,*])/$
            coord_arr[0,*]^2), fit_res[0,*]
        gamdiag_plot_result_vs_x, coord_arr[1,*]*par.major_R/par.minor_r,$
            fit_res, theo_res
      
    ENDIF
    set_output, diag, sp, /reset

;  %%%%%%%%%% DATA output only for one set of var, kx, ky, z %%%%%%%%%%%%%%%%%%%%%%
   xind = n_kxind/2
   real_temp = REFORM(DOUBLE(full_data[xind,0,isp,*]));/$
;     MAX(ABS(full_data[xind,0,isp,*]))
   imag_temp = REFORM(IMAGINARY(full_data[xind,0,isp,*]));/$
;     MAX(ABS(full_data[xind,0,isp,*]))

   set_output, diag, sp, header = ['t','Re['+get_var_string((*i).vars[0])+']',$
     'Im['+get_var_string((*i).vars[0])+']'], dat=[[time],[real_temp],[imag_temp]],$
     commentline=['xind= '+rm0es(xind)+', x/a= '+rm0es(coord_arr[1,xind]*par.major_R/par.minor_r)+$
      ', q= '+rm0es(coord_arr[0,xind]),$
      'Fit:    '+rm0es(fit_res[0,xind])+','+rm0es(fit_res[1,xind])+','$
       +rm0es(fit_res[2,xind]),$
      'Analytic: '+rm0es(theo_res[0,xind])+','+rm0es(theo_res[1,xind])+','$
       +rm0es(theo_res[2,xind])]

   IF (n_kxind*n_zind GT 1) THEN BEGIN
       set_output, diag, sp, header = ['x/a','q','A_R','gamma','omega','A_R(theo)',$
         'gam(theo)','om(theo)'], dat=[[REFORM(coord_arr[1,*])*par.major_R/par.minor_r],$
         [REFORM(coord_arr[0,*])],$                              
         [REFORM(fit_res[0,*])],[REFORM(fit_res[1,*])],[REFORM(fit_res[2,*])],$
         [REFORM(theo_res[0,*])],[REFORM(theo_res[1,*])],[REFORM(theo_res[2,*])]],/append
   ENDIF

   ENDIF ELSE BEGIN
     print, 'gamdiag: cannot create plots with one time step'
   ENDELSE
   

  ENDFOR ;-- isp
END

;######################################################################

PRO gamfunct, X, A, F, pder
  
  f1x = EXP(- A[1] * X)
  f2x = COS( A[2] * X)
  F = A[0] + (1 - A[0]) * f1x * f2x
  
;If the procedure is called with four parameters, calculate the  
;partial derivatives.  
  IF N_PARAMS() GE 4 THEN $  
    pder = [[1.-f1x*f2x], [( A[0] - 1.) * X * f1x*f2x], [( A[0] - 1.) * X * SIN( A[2] * X) *f1x ]]  
END  

;######################################################################

PRO get_sugama_res, q_, x, R0, kx, Bfield, res
; see J. Plasma Phys, 72 (2006), pp.825-828
; and Erratum: J. Plasma Phys. 
; input: q, x, R0, Bfield
; output: array res=[residual, gamma, omega]

  erratum = 0  ;the non-corrected version seems to fit better???
  L_par = q_*R0
  sqrt2 = sqrt(2.0)
  TeTi = 1.0
  
  help_fac = 1 + 2.*(23.+16.*TeTi+4.*TeTi^2)/(q_*(7+4*TeTi))^2
  omega_hat = 0.5*sqrt(7.+4.*TeTi)*q_*sqrt(help_fac)
  omega_f = omega_hat*sqrt2/L_par

  if (erratum) then begin
     gamma_f = sqrt(!PI)/2.*q_^2*(sqrt2/L_par)/help_fac*$
               (exp(-omega_hat^2)*omega_hat^2*(omega_hat^2+1+2*TeTi)+$
                0.25*(sqrt2*kx*q_/Bfield)^2*exp(-0.25*omega_hat^2)*omega_hat^2*$
                (omega_hat^4/128.+(1.+TeTi)/16.*omega_hat^2+$
                 (3./8.+7./16.*TeTi+5./32.*TeTi^2)))
;; maybe, the erratum is missing a factor of 2.0 in the last major bracket??
  endif else begin
     gamma_f = sqrt(!PI)/2.*q_^2*(sqrt2/L_par)/help_fac*$
               (exp(-omega_hat^2)*omega_hat^2*(omega_hat^2+1+2*TeTi)+$
                0.25*(sqrt2*kx*q_/Bfield)^2*exp(-0.25*omega_hat^2)*omega_hat^2*$
                (omega_hat^4/64.+(1.+3.*TeTi/8.)*(omega_hat^2/8.+3./4.)))
  endelse

  A_R = 1./(1.+1.6*q_^2/sqrt(x/R0))

  res=[A_R,gamma_f,omega_f]
  
END

;######################################################################

PRO print_result, A, B, avg, median
    print, '        Analytic   Fit      Analytic/Fit   Average         Median'
    print, 'A_R     ',STRING(A[0],FORMAT='(F6.4)'),'     ',$
           STRING(B[0],FORMAT='(F6.4)'),'   ',$
           STRING(A[0]/B[0],FORMAT='(F6.4)')+'         ',$
           STRING(avg,FORMAT='(F6.4)')+'         ',$
           STRING(median,FORMAT='(F6.4)')
    print, 'gamma   ',STRING(A[1],FORMAT='(F6.4)'),'     ',$
           STRING(B[1],FORMAT='(F6.4)'),'   ',$
           STRING(A[1]/B[1],FORMAT='(F6.4)')
    print, 'omega   ',STRING(A[2],FORMAT='(F6.4)'),'     ',$
           STRING(B[2],FORMAT='(F6.4)'),'   ',$
           STRING(A[2]/B[2],FORMAT='(F6.4)')
END
