FUNCTION spectra_info

  RETURN, {$
    type      : 'mom',$ ;'mom_uni',$
    title     : 'Amplitude spectra (ky/kx)',$
    help_text : ['Plots amplitude spectra in ky/kx space, '+$
        'averaged over the remaining spatial coordinates and time.'],$
    ext_vars  : [$		 
        ['vars','0','index of the fluctuating quantity to be plotted '+$
         '(see variable list); default: all; negative values for '+$
         'multiplication of variable (-var-1) with k_[x,y]'],$
        ['kxind','0','kx index at which to evaluate ky spectrum; '+$
         'default: average (-1)'],$
        ['kyind','0','ky index at which to evaluate kx spectrum; '+$
         'default: average (-1)'],$
        ['minmax','0','plot range; default: [1e-4,1e12]'],$
        ['fit','0','relative fit range (2d array); default: no fit']]}

END

;######################################################################

PRO spectra_init, diag

  COMMON global_vars

  e = (*diag).external_vars

  IF NOT KEYWORD_SET(*e.minmax) THEN *e.minmax = [1e-4,1e12]
  IF N_ELEMENTS(*e.vars) LT 1 THEN $
    IF par.beta GT 0 THEN *e.vars = INDGEN(par.n_moms+par.n_fields) $
    ELSE *e.vars = [0,INDGEN(par.n_moms)+par.n_fields]

  ; look for negative variables, set k multiplication
  neg_var = WHERE(*e.vars LT 0)
  kmult_var = BYTARR(N_ELEMENTS(*e.vars))
  IF neg_var[0] NE -1 THEN BEGIN
    (*e.vars)[neg_var] = - (*e.vars)[neg_var] - 1
    kmult_var[neg_var] = 1B
  ENDIF

  IF N_ELEMENTS(*e.kxind) LT 1 THEN $
    IF (par.kx0_ind EQ 0) THEN *e.kxind = -1 $
    ELSE *e.kxind = 0
  IF N_ELEMENTS(*e.kyind) LT 1 THEN $
    IF par.adapt_lx THEN *e.kyind = par.ky0_ind $
    ELSE *e.kyind = -1

  IF par.adapt_lx THEN BEGIN
    PRINT, (*diag).name + ' warning: using only kx=0 mode ' + $
      '(without conn.) for ky spectrum'
    *e.kxind = 0
  ENDIF

  IF N_ELEMENTS(*e.fit) NE 2 THEN fit = [0.0,0.0] ELSE fit = *e.fit

  i = set_internal_vars(diag,{$
    vars        : *e.vars,$
    kmult_var   : kmult_var,$
    kxind       : *e.kxind,$
    kyind       : *e.kyind,$  
    xsumspec_id : PTR_NEW(),$
    ysumspec_id : PTR_NEW(),$
    fit     	: fit,$
    minmax      : *e.minmax})

  fft_format, kxky=*e.vars

END

;######################################################################

PRO spectra_loop, diag

  COMMON global_vars

  i = (*diag).internal_vars

  nkxind = N_ELEMENTS((*i).kxind)
  nkyind = N_ELEMENTS((*i).kyind)
  nkxpos = par.nkx0 / 2 > 1
  nvars = N_ELEMENTS((*i).vars)

  xsumspec = FLTARR(nkxpos,nkyind,nvars,gui.out.n_spec_sel,/NOZERO)
  ysumspec = FLTARR(par.nky0,nkxind,nvars,gui.out.n_spec_sel,/NOZERO)
  xtempdata = FLTARR(nkxpos,par.nz0,/NOZERO)
  ytempdata = FLTARR(par.nky0,par.nz0,/NOZERO)

  FOR isp = 0, gui.out.n_spec_sel - 1 DO BEGIN
    FOR ivar = 0, nvars - 1 DO BEGIN
      FOR iky = 0, nkyind - 1 DO BEGIN
        IF (*i).kyind[iky] EQ -1 THEN BEGIN ; x spectrum averaged over ky
          IF par.x_local THEN BEGIN
            xtempdata[0,*] = 2.0 * TOTAL(ABS($
              (*mom[isp,(*i).vars[ivar]].kxky)[0,1:par.nky0-1,*])^2,2) + $
              TOTAL(ABS((*mom[isp,(*i).vars[ivar]].kxky)[0,0,*])^2,2)
            xtempdata[1:nkxpos-1,*] = TOTAL(ABS($
              (*mom[isp,(*i).vars[ivar]].kxky)[1:nkxpos-1,1:par.nky0-1,*])^2,2) + $
              REVERSE(TOTAL(ABS((*mom[isp,(*i).vars[ivar]].kxky)$
              [par.nkx0-nkxpos+1:par.nkx0-1,1:par.nky0-1,*])^2,2),1) + $
              TOTAL(REFORM(ABS((*mom[isp,(*i).vars[ivar]].kxky)[1:nkxpos-1,0,*])^2,$
              [nkxpos-1,1,par.nz0]),2)
          ENDIF ELSE BEGIN
            xtempdata[0,*] = 2.0 * (TOTAL(ABS($
              (*mom[isp,(*i).vars[ivar]].kxky)[0,1:par.nky0-1,*])^2,2) + $
              TOTAL(ABS((*mom[isp,(*i).vars[ivar]].kxky)[0,0,*])^2,2)) * $
              (*series.geom).jac_norm[0,*]
            xtempdata[1:nkxpos-1,*] = TOTAL(ABS($
              (*mom[isp,(*i).vars[ivar]].kxky)[1:nkxpos-1,1:par.nky0-1,*])^2*$
              REBIN(REFORM((*series.geom).jac_norm[1:nkxpos-1,*],$
              [nkxpos-1,1,par.nz0]),[nkxpos-1,par.nky0-1,par.nz0]),2) + $
              REVERSE(TOTAL(ABS((*mom[isp,(*i).vars[ivar]].kxky)$
              [par.nkx0-nkxpos+1:par.nkx0-1,1:par.nky0-1,*])^2*$
              REBIN(REFORM((*series.geom).jac_norm[par.nkx0-nkxpos+1:par.nkx0-1,*],$
              [nkxpos-1,1,par.nz0]),[nkxpos-1,par.nky0-1,par.nz0]),2),1) + $
              TOTAL(ABS((*mom[isp,(*i).vars[ivar]].kxky)[1:nkxpos-1,0,*])^2*$
              (*series.geom).jac_norm[1:nkxpos-1,*],2)
          ENDELSE
        ENDIF ELSE BEGIN ; x spectrum for certain ky modes
          IF (par.ky0_ind EQ 0) AND ((*i).kyind[iky] EQ 0) THEN BEGIN ; --- zero ky mode
            xtempdata[0,*] = ABS((*mom[isp,(*i).vars[ivar]].kxky)[0,0,*])^2
            IF nkxpos GT 1 THEN xtempdata[1:nkxpos-1,*] = $
              ABS((*mom[isp,(*i).vars[ivar]].kxky)[1:nkxpos-1,0,*])^2
          ENDIF ELSE BEGIN
            xtempdata[0,*] = ABS((*mom[isp,(*i).vars[ivar]].kxky)[0,(*i).kyind[iky],*])^2
            IF nkxpos GT 1 THEN xtempdata[1:nkxpos-1,*] = $
              ABS((*mom[isp,(*i).vars[ivar]].kxky)[1:nkxpos-1,(*i).kyind[iky],*])^2 + $
              ABS(REVERSE((*mom[isp,(*i).vars[ivar]].kxky)$
              [par.nkx0-nkxpos+1:par.nkx0-1,(*i).kyind[iky],*],1))^2
          ENDELSE
        ENDELSE

        IF par.x_local THEN xtempdata *= REBIN(REFORM($
          (*series.geom).jac_norm,[1,par.nz0]),[nkxpos,par.nz0])

        xsumspec[*,iky,ivar,isp] = par.nz0 GT 1 ? $
          TOTAL(xtempdata,2) / par.nz0 : xtempdata
      ENDFOR ; --- iky loop
      FOR ikx = 0, N_ELEMENTS((*i).kxind) - 1 DO BEGIN
        IF (*i).kxind[ikx] EQ -1 THEN ytempdata[0,0] = $
          TOTAL(ABS((*mom[isp,(*i).vars[ivar]].kxky))^2,1) ELSE $
          ytempdata[0,0] = $
          ABS(REFORM((*mom[isp,(*i).vars[ivar]].kxky)[(*i).kxind[ikx],*,*]))^2

        IF par.x_local THEN ytempdata *= REBIN(REFORM($
          (*series.geom).jac_norm,[1,par.nz0]),[par.nky0,par.nz0])

        ysumspec[*,ikx,ivar,isp] = par.nz0 GT 1 ? $
          TOTAL(ytempdata,2) / par.nz0 : ytempdata
      ENDFOR ; --- ikx loop
    ENDFOR ; --- ivar
  ENDFOR ; --- species loop

  (*i).xsumspec_id = time_avg((*i).xsumspec_id,xsumspec,mom_time,$
                              fwd=gui.out.res_steps)
  (*i).ysumspec_id = time_avg((*i).ysumspec_id,ysumspec,mom_time,$
                              fwd=gui.out.res_steps)

END

;######################################################################

PRO spectra_plot, xdata, ydata, log=log, position=position, $
  title=title, xtitle=xtitle, ytitle=ytitle, fit=fit, $
  colorarr=colorarr, linestyle=linestyle, minmax=minmax, vars=vars, $
  kmult=kmult, label=label, xname=xname, warn=warn

  IF NOT KEYWORD_SET(label) THEN label = xtitle
  IF N_ELEMENTS(warn) LT 1 THEN warn = 0

  n_vars = N_ELEMENTS(ydata[0,*])
  IF N_ELEMENTS(kmult) NE n_vars THEN kmult = BYTARR(n_vars)

  rho_str = ' ' + get_var_string(/rhostr)

  ndata = N_ELEMENTS(xdata)

  mymax = MAX(ydata, min=mymin)
  IF mymax GT minmax[1] THEN BEGIN
    mymax = minmax[1]
    IF warn THEN PRINT, 'suppressed values > ' + rm0es(mymax) + $
      ' in ' + xname + ' spectra'
  ENDIF

  IF log THEN BEGIN
    posv = WHERE(xdata GT 0.0)
    tickformat = 'betterticks'
    IF mymin LT minmax[0] THEN BEGIN
      mymin = minmax[0]
      IF warn THEN PRINT, 'suppressed values < ' + rm0es(mymin) + $
        ' in ' + xname + ' spectra'
    ENDIF
  ENDIF ELSE BEGIN
    posv = INDGEN(ndata)
    tickformat = ''
  ENDELSE
  
  PLOT, xdata[posv], ydata[posv,0], COLOR=1, $
    XLOG=log, XTICKFORMAT=tickformat, XTITLE=xtitle+rho_str,$
    YLOG=log, YTICKFORMAT=tickformat, YTITLE=ytitle,$
    YMARGIN=[0,1], YRANGE=[mymin,mymax], XTICKLEN=1.0, $ ;/YSTYLE, $
    /XSTYLE, TITLE=title, POSITION=position, $
    CHARSIZE=1.0,/NODATA

  FOR ivar = 0, n_vars - 1 DO BEGIN
    ydata_mod = (kmult[ivar] ? xdata[posv]^2 : 1) * ydata[posv,ivar]
    OPLOT, xdata[posv], ydata_mod, COLOR=colorarr[ivar], $
      LINESTYLE=linestyle[ivar], PSYM=-2*kmult[ivar]
  ENDFOR

  IF log AND (TOTAL(fit) NE 0) THEN BEGIN
    PRINT, 'fit results for ' + rm_idl_fmt(title) + ' in range ' + $
      xname + '=[' + rm0es(xdata[fit[0]*ndata]) + ',' + $
      rm0es(xdata[fit[1]*ndata]) + ']:'
    FOR ivar = 0, n_vars - 1 DO BEGIN
      fitres = LINFIT(ALOG10(xdata[fit[0]*ndata:fit[1]*ndata]),$
        ALOG10(ydata[fit[0]*ndata:fit[1]*ndata,ivar]))
      OPLOT, xdata[0.75*fit[0]*ndata:(1.25*fit[1]<1.0)*ndata],$
        10^fitres[0]*xdata[0.75*fit[0]*ndata:(1.25*fit[1]<1.0)*ndata]^fitres[1],$
        COLOR=colorarr[ivar], linestyle=3
      xmid =10^(0.5*(ALOG10(xdata[fit[1]*ndata])+ALOG10(xdata[fit[0]*ndata])))
      XYOUTS, xmid, 10^(0.5*fitres[0])*xmid^fitres[1], '!9'+STRING(63B)+' !6'+$
        label+'!E'+rm0es(fitres[1],prec=2)+'!N',color=colorarr[ivar]

      PRINT, '<|' + get_var_string(vars[ivar]) + '|^2> ~ ' + $
        xname + '^' + rm0es(fitres[1],prec=2)
    ENDFOR
  ENDIF

END

;######################################################################

PRO spectra_output, diag

  COMMON global_vars

  i = (*diag).internal_vars

  nkxind = N_ELEMENTS((*i).kxind)
  nkyind = N_ELEMENTS((*i).kyind)
  nvars = N_ELEMENTS((*i).vars)
  nsp = gui.out.n_spec_sel  
  nkxpos = par.nkx0 / 2 > 1
  n_res_steps = gui.out.res_steps * (series.step_count - 1) + 1

  xsumspec = REFORM(time_avg((*i).xsumspec_id,/avg,fwd=gui.out.res_steps,$
    tarr=time),[nkxpos,nkyind,nvars,nsp,n_res_steps])
  ysumspec = REFORM(time_avg((*i).ysumspec_id,/avg,fwd=gui.out.res_steps,$
    tarr=time),[par.nky0,nkxind,nvars,nsp,n_res_steps])

  ; this part ensures that variables have always the same color
  ; (if par.n_fields = const.)
  colorarr = INDGEN(par.n_moms+par.n_fields) + 1
  linestyle = INTARR(par.n_moms+par.n_fields)
  FOR var = 6, par.n_moms - 1 DO BEGIN
     colorarr[var+par.n_fields] = colorarr[par.n_fields+(var MOD 6)]
     linestyle[var+par.n_fields] = var / 6
  ENDFOR

  colorarr = colorarr[(*i).vars]
  linestyle = linestyle[(*i).vars]
  psym = INTARR(nvars)
  kx_str = STRARR(nvars)
  ky_str = STRARR(nvars)

  IF TOTAL((*i).kmult_var) GT 0 THEN BEGIN
    kmult_inds = WHERE((*i).kmult_var)
    psym[kmult_inds] = -2
    kx_str[kmult_inds] = '!6k!Dx!N'
    ky_str[kmult_inds] = '!6k!Dx!N'
  ENDIF

  rho_str = ' ' + get_var_string(/rhostr)

  ky = (*series.ky)
  kx = (*series.kx)[0:par.nkx0/2-1,0]

  FOR isp = 0, gui.out.n_spec_sel - 1 DO BEGIN
    sp = (*gui.out.spec_select)[isp]

    set_output, diag, sp, /ps, multi=[0,1,2], coltable=41

    FOR ikx = 0, N_ELEMENTS((*i).kxind) - 1 DO BEGIN
      IF (*i).kxind[ikx] EQ -1 THEN kx_str = 'k!Dx!N avg.' $
      ELSE kx_str = 'k!Dx!N' + get_var_string(/rhostr) + '=' + $
                    rm0es(kx[(*i).kxind[ikx]],prec=3)

      maxy = MAX(ysumspec[*,ikx,*,isp,*],MIN=miny)
      logminy = miny
      IF maxy GT (*i).minmax[1] THEN BEGIN
        maxy = (*i).minmax[1]
        PRINT, 'suppressed values > ' + rm0es(maxy) + ' in ky spectra'
      ENDIF
      IF miny LT (*i).minmax[0] THEN BEGIN
        logminy = (*i).minmax[0]
        PRINT, 'suppressed values < ' + rm0es(logminy) + ' in ky spectra'
      ENDIF

      ; --- ky spectra ---
      ; plot k_y spectrum (logarithmic)
      title = '!6k!Dy!N spectrum (' + kx_str + ')'
   
      FOR n = 0, n_res_steps - 1 DO BEGIN 
        IF par.nky0 GT (1 + (par.ky0_ind EQ 0)) THEN BEGIN    
          spectra_plot, ky, REFORM(ysumspec[*,ikx,*,isp,n]), $
            title=title, ytitle='!12<!9!!!6A!9!!!6!U2!N!12>!6', $
            xtitle='!6k!Dy!N', minmax=(*i).minmax, $
            fit=(*i).fit, colorarr=colorarr, linestyle=linestyle, $
            vars=(*i).vars, kmult=(*i).kmult_var, xname='ky', $
            position=[0.15,0.596,0.95,0.9615], /log, /warn

          linplot_end = 0.538
          title = ''
        ENDIF ELSE linplot_end = 0.9615

        ; plot k_y spectrum (linear)
        IF par.nky0 GT 1 THEN BEGIN
          spectra_plot, ky, REFORM(ysumspec[*,ikx,*,isp,n]), $
            title=title, ytitle='!12<!9!!!6A!9!!!6!U2!N!12>!6', $
            xtitle='!6k!Dy!N', minmax=(*i).minmax, $
            fit=(*i).fit, colorarr=colorarr, linestyle=linestyle, $
            vars=(*i).vars, kmult=(*i).kmult_var, xname='ky', $
            position=[0.15,0.231,0.95,linplot_end], log=0

          plot_legend, colorarr, ky_str + get_var_string((*i).vars,/fancy), $
            x_offset=[0.15,0.0,0.95,0.12], linestyles=linestyle, psym=psym
        ENDIF

        plot_info_str, diag, time=(gui.out.res_steps ? time[n] : undef)

        IF N_ELEMENTS(ky) EQ 1 THEN BEGIN
           set_output, diag, sp, header=['ky',get_var_string((*i).vars)],append=(ikx+n GT 0),$
             dat=[ky,REFORM(ysumspec[*,ikx,*,isp,n])], commentline='kxind = '+$
             rm0es((*i).kxind[ikx])+(gui.out.res_steps ? ', time = '+rm0es(time[n]) : '') 
        ENDIF ELSE BEGIN
           set_output, diag, sp, header=['ky',get_var_string((*i).vars)],append=(ikx+n GT 0), $
             dat=[[ky],[REFORM(ysumspec[*,ikx,*,isp,n])]], commentline='kxind = '+$
             rm0es((*i).kxind[ikx])+(gui.out.res_steps ? ', time = '+rm0es(time[n]) : '')
        ENDELSE

      ENDFOR ; time loop
    ENDFOR ; kx loop
  
    ; --- kx spectra ---
    FOR iky = 0, N_ELEMENTS((*i).kyind) - 1 DO BEGIN
      kx = (*series.kx)[0:par.nkx0/2-1,iky]

      IF (*i).kyind[iky] EQ -1 THEN ky_str = 'k!Dy!N avg.' $
        ELSE ky_str = 'k!Dy!N' + get_var_string(/rhostr) + '=' + $
        rm0es(ky[(*i).kyind[iky]],prec=3)

      ; plot k_x spectrum (logarithmic)
      title = '!6k!Dx!N spectrum (' + ky_str + ')'

      FOR n = 0, gui.out.res_steps * (series.step_count - 1) DO BEGIN
        IF par.nkx0 / 2 GT 1 THEN BEGIN
          spectra_plot, kx, REFORM(xsumspec[*,iky,*,isp,n]), $
            title=title, ytitle='!12<!9!!!6A!9!!!6!U2!N!12>!6', $
            xtitle='!6k!Dx!N', minmax=(*i).minmax, $
            fit=(*i).fit, colorarr=colorarr, linestyle=linestyle, $
            vars=(*i).vars, kmult=(*i).kmult_var, xname='kx', $
            position=[0.15,0.596,0.95,0.9615], /log, /warn

          linplot_end = 0.538
          title = ''
        ENDIF ELSE linplot_end = 0.9615

        ; plot time k_x spectrum (linear)
        IF par.nkx0 / 2 GT 1 THEN BEGIN
          spectra_plot, kx, REFORM(xsumspec[*,iky,*,isp,n]), $
            title=title, ytitle='!12<!9!!!6A!9!!!6!U2!N!12>!6', $
            xtitle='!6k!Dx!N', minmax=(*i).minmax, $
            fit=(*i).fit, colorarr=colorarr, linestyle=linestyle, $
            vars=(*i).vars, kmult=(*i).kmult_var, xname='kx', $
            position=[0.15,0.231,0.95,linplot_end], log=0

          plot_legend, colorarr, kx_str + get_var_string((*i).vars,/fancy), $
            x_offset=[0.15,0.0,0.95,0.12], linestyles=linestyle, psym=psym
        ENDIF

        plot_info_str, diag, time=(gui.out.res_steps ? time[n] : undef)

        set_output, diag, sp, header=['kx',get_var_string((*i).vars)], $
                    dat=[[kx],[REFORM(xsumspec[*,iky,*,isp,n])]], commentline='kyind = '+$
                    rm0es((*i).kyind[iky])+(gui.out.res_steps ? ', time = '+rm0es(time[n]) : ''),$
                    /append
        ENDFOR ; time loop
    ENDFOR ; ky loop

    set_output, diag, sp, /reset
  ENDFOR ; species loop

END
