FUNCTION vsp1_info

  RETURN, {$
    type      : 'vsp',$
    title     : 'velocity space',$
    help_text : ['vsp diag'],$
    ext_vars  : [['var','0','variable to be plotted '+$
    	    	  '0 = G_es, 1 = G_em, 2 = Q_es, 3 = Q_em, 4 = abs(f_); '+$
		  'default: 4'],$
    	    	['zind','0','index of parallel slices to be plotted; '+$
		 'default: [0,nz0/2,nz0-1]'],$
		['log','1','switch on/off logarithmic color encoding']]}

END

;######################################################################

PRO vsp1_init, diag

  COMMON global_vars
  
  e = (*diag).external_vars
  
  IF N_ELEMENTS(*e.var) NE 1 THEN *e.var = 4 ; 0 = G_es, 1 = G_em, 2 = Q_es, 3 = Q_em, 4 = abs(f_)
  IF N_ELEMENTS(*e.zind) LT 1 THEN *e.zind = [0,par.nz0/2,par.nz0-1]
  IF N_ELEMENTS(*e.log) NE 1 THEN *e.log = 0

  i = set_internal_vars(diag,{$
    var        : *e.var,$
    zind       : *e.zind,$
    log        : *e.log,$
    time_id    : PTR_NEW(),$
    average_id : PTR_NEW(),$
    data_id    : PTR_NEW()})

END

;######################################################################

PRO vsp1_loop, diag

  COMMON global_vars
  
  i = (*diag).internal_vars  

  (*i).average_id = time_avg((*i).average_id,(*vsp)$
      [(*i).zind,*,*,*,(*i).var],vsp_time)
    
  (*i).time_id = store_step((*i).time_id,vsp_time)
  (*i).data_id = store_step((*i).data_id,(*vsp)[(*i).zind,*,*,*,(*i).var])
      
END

;######################################################################

PRO vsp1_output, diag

  COMMON global_vars

  i = (*diag).internal_vars

  average = time_avg((*i).average_id,/avg)
  time = store_step((*i).time_id,/get)
  data = store_step((*i).data_id,/get)

  vpar = - par.lv + INDGEN(par.nv0) * 2.0 * par.lv / (par.nv0 - 1.0)
  GetMuWeightsAndKnots, mu_weights, mu
  
  c_levels = 40
  nzind = N_ELEMENTS((*i).zind)

  FOR isp = 0, gui.out.n_spec_sel - 1 DO BEGIN
    sp = (*gui.out.spec_select)[isp]
    
    set_output, diag, sp, /ps, coltable=33, multi=[0,1,1]

    first = 1
    FOR n = 0, series.step_count - 1 DO BEGIN
      FOR iz = 0, nzind - 1 DO BEGIN
        pos = [0.1,0.1+0.9/nzind*(nzind-1-iz),0.925,1.0-0.9/nzind*iz]
        dpos_x = pos[2] - pos[0]
        dpos_y = pos[3] - pos[1]
        dy = 0.15
      
        IF (WHERE((*data[n])[iz,*,*,isp,0] NE 0.0))[0] NE -1 THEN BEGIN
          lev_col = contour_levels((*data[n])[iz,*,*,isp,0],c_levels,log=(*i).log)
	  
          CONTOUR, (*data[n])[iz,*,*,isp,0], vpar, mu, $
            LEVELS=lev_col[*,0], C_COLORS=lev_col[*,1], $
            /XSTYLE, /YSTYLE, /FILL, XTITLE='!6v!I!9#!6!N', $
            YTITLE='!7l!6', TITLE=get_vsp_string((*i).var,/fancy)+' @ z/qR = '+$
            rm0es((*par.z)[(*i).zind[iz]],prec=3), NOERASE=(iz GT 0),$
            POSITION=[pos[0]+0.01*dpos_x,pos[1]+dy*dpos_y,pos[0]+0.9*dpos_x,$
	    pos[3]-dy*dpos_y]
	    


          plot_trapped_passing_bnd, (*i).zind[iz], vpar, mu

          plot_colorbar, lev_col, prec=3, log=(*i).log, $
	    POSITION=[pos[0]+0.95*dpos_x,pos[1]+dy*dpos_y,pos[2],pos[3]-dy*dpos_y]

          XYOUTS, 0.05, 0.005, 'time: ' + rm0es(*time[n]) + ' ' + $
            get_var_string(1,/time,/ounit) + ', ' + spec[sp].name, /NORMAL, CHARSIZE=1.0

          set_output, diag, sp, dat=[[REFORM((*data[n])[iz,*,*,isp,0])]], $
	    append=(first EQ 0), commentline='time = '+rm0es(*time[n])+$
	    ', var = '+get_vsp_string((*i).var)+' , z = '+rm0es((*par.z)[(*i).zind[iz]])
        ENDIF ELSE BEGIN
          PRINT, 'vsp: empty plot'
          lev_col = contour_levels((*data[n])[iz,*,*,isp,0])	  
          CONTOUR, (*data[n])[iz,*,*,isp,0], vpar, mu, $
            LEVELS=lev_col[*,0], C_COLORS=lev_col[*,1], NOERASE=(iz GT 0),$
            /XSTYLE, /YSTYLE, /NODATA, XTITLE='!6v!I!9#!6!N', $
            YTITLE='!7l!6', TITLE='z/qR = '+rm0es((*par.z)[(*i).zind[iz]],prec=3)	  
        ENDELSE
	first = 0
      ENDFOR
    ENDFOR
    IF sp EQ gui.out.n_spec_sel - 1 THEN PTR_FREE, data, time

    IF series.step_count GT 1 THEN BEGIN ; --- plot time average      
      FOR iz = 0, N_ELEMENTS((*i).zind) - 1 DO BEGIN
        pos = [0.1,0.1+0.9/nzind*(nzind-1-iz),0.925,1.0-0.9/nzind*iz]
        dpos_x = pos[2]-pos[0] & dpos_y = pos[3]-pos[1]
        dy = 0.15      
        IF (WHERE(average[iz,*,*,isp,0] NE 0.0))[0] NE -1 THEN BEGIN      
          lev_col = contour_levels(average[iz,*,*,isp,0],c_levels,log=(*i).log)
          CONTOUR, average[iz,*,*,isp,0], vpar, mu, $
            LEVELS=lev_col[*,0], C_COLORS=lev_col[*,1], $
            /XSTYLE, /YSTYLE, /FILL, XTITLE='!6v!I!9#!6!N', $
            YTITLE='!7l!6', TITLE='time average of '+get_vsp_string((*i).var,/fancy)+$
	      ' @ z/qR = '+rm0es((*par.z)[(*i).zind[iz]],prec=3), $
            POSITION=[pos[0]+0.01*dpos_x,pos[1]+dy*dpos_y,pos[0]+0.9*dpos_x,$
	    pos[3]-dy*dpos_y], NOERASE=(iz GT 0)

          plot_trapped_passing_bnd, (*i).zind[iz], vpar, mu

          plot_colorbar, lev_col, prec=3, log=(*i).log, $
	    POSITION=[pos[0]+0.95*dpos_x,pos[1]+dy*dpos_y,pos[2],pos[3]-dy*dpos_y]

          XYOUTS, 0.05, 0.005, 'time average, ' + spec[sp].name, /NORMAL, CHARSIZE=1.0

          set_output, diag, sp, dat=[[REFORM(average[iz,*,*,isp,0])]], $
	    append=(first EQ 0), commentline='time averaged, var = '+$
            get_vsp_string((*i).var)+' , z = '+rm0es((*par.z)[(*i).zind[iz]])
        ENDIF ELSE BEGIN
          PRINT, 'vsp: empty plot'
          lev_col = contour_levels(average[iz,*,*,isp,0])
          CONTOUR, average[iz,*,*,isp,0], vpar, mu, $
            LEVELS=lev_col[*,0], C_COLORS=lev_col[*,1],NOERASE=(iz GT 0),$
            /XSTYLE, /YSTYLE, /NODATA, XTITLE='!6v!I!9#!6!N', $
            YTITLE='!7l!6', TITLE='z/qR = '+rm0es((*par.z)[(*i).zind[iz]],prec=3)
        ENDELSE
      ENDFOR
    ENDIF
    
    set_output, diag, sp, /reset
  ENDFOR ; --- isp loop 

END

;################################################################################

PRO plot_trapped_passing_bnd, zind, vpar, mu, Blev=Blev
  COMMON global_vars

  IF NOT KEYWORD_SET(Blev) THEN Blev = max((*series.geom).Bfield)

  vpar_int = INTERPOL(vpar,512>par.nv0)
  mu_int = mu ;INTERPOL(mu,64>par.nw0)

  show_grid = 0
  IF (show_grid) THEN BEGIN
    FOR ivp=0, N_ELEMENTS(vpar)-1 DO BEGIN
      FOR imu = 0, N_ELEMENTS(mu)-1 DO BEGIN
         PLOTS, vpar[ivp], mu[imu], PSYM=4, color=10, SYMSIZE=0.5
      ENDFOR
    ENDFOR
  ENDIF

  FOR imu = 0, N_ELEMENTS(mu_int)-1 DO BEGIN
     energy_old= vpar_int[0]^2 + mu_int[imu]*(*series.geom).Bfield[zind]
     FOR ivp = 1, N_ELEMENTS(vpar_int)-1 DO BEGIN
        energy = vpar_int[ivp]^2 + mu_int[imu]*(*series.geom).Bfield[zind]
        IF ((energy_old-mu_int[imu]*Blev)*(energy-mu_int[imu]*Blev) LE 0) THEN $
           PLOTS, vpar_int[ivp], mu_int[imu], PSYM=4, color=3, SYMSIZE=0.5
        energy_old = energy
     ENDFOR
  ENDFOR

  ; separation line of trapped and passing particles (analytic result
  IF (par.magn_geometry EQ '') OR (par.magn_geometry EQ 's-alpha') $
     OR (par.magn_geometry EQ 's_alpha') AND par.n_pol EQ 1 THEN $
        OPLOT, vpar_int, (1.0 - par.trpeps) * vpar_int^2 / (par.trpeps * $
        (1.0 + COS(-!PI+zind*par.dz))), COLOR=3
END
