###########################################################################
### architecture dependent GENE makefile for EPFL BlueGene/P            ###
###########################################################################

###########################################################################
### SWITCHES                                                            ###
###########################################################################

# FFTLIB - needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB = essl
PRECISION = double
OPENMP = no

SLEPC= no
WITHNAG = no
USE_PERFLIB = none
MASSLIB = no
DEBUG = no
PRODRUN = no
SCALAPACK = no

FUTILS = yes

#memory per core
PREPROC= -WF,-D'MB_PER_CORE=700'

###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
BGP_SYS = /bgsys/drivers/ppcfloor/comm
INCPATHS = -I$(OBJDIR) -I. -I$(SRCDIR) -I$(BGP_SYS)/include

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
LIBS = -L$(BGP_SYS)/lib

#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 INCPATHS +=
 LIBS +=
endif
ifeq ($(FFTLIB),essl)
 INCPATHS +=
 LIBS += -L/bgsys/ibm_essl/sles10/prod/opt/ibmmath/lib -lessl
 PREPROC += -WF,-DWITHESSL
endif
ifeq ($(FFTLIB),fftw)
 INCPATHS +=
 LIBS +=
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(WITHNAG),yes)
 INCPATHS +=
 LIBS +=
#DEFAULT: use NAG-FORTRAN 90
 PREPROC +=-WF,-DWITHNAGF90
#Alternative: use NAG_FORTRAN-F77
#PREPROC +=-WF,-DWITHNAGF77 
endif

ifeq ($(SLEPC),yes)
  PETSC_DIR=
  SLEPC_DIR=
  PETSC_ARCH =
  include $(SLEPC_DIR)/conf/slepc_common
  INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
  #LIBS +=$(PETSC_LIB) $(SLEPC_LIB)
  LIBS += $(SLEPC_LIB)
  PREPROC += -WF,-DWITHSLEPC
endif

ifeq ($(SCALAPACK),yes)
  LIBS += -L/usr/local/lib -lscalapack -llapack -lblacsF77init -lblacs 
  LIBS += -L/opt/ibmmath/lib -lessl
  PREPROC += -WF,-DWITHSCAL
endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -WF,-DWITHPERF
endif
ifeq ($(USE_PERFLIB),perf)
 LIBS += -L/usr/local/lib
endif

ifeq ($(MASSLIB),mass)
 MASS_LIB = -lmass
endif

ifeq ($(FUTILS),yes)
 #  FUTILS and HDF5
 FUTILSDIR = /home/ttran/Utils/futils/src/

 HDF5_LIBPATH  = -L/bgsys/lib/ -L$(FUTILSDIR)
 INCPATHS +=  -I/bgsys/include/hdf5
 
 HDF5_LIBS = -lfutils -lhdf5_fortran -lhdf5 -lz


 LIBS += $(HDF5_LIBPATH) $(HDF5_LIBS)
 PREPROC += -WF,-DWITHFUTILS
 INCPATHS += -I$(FUTILSDIR)
endif

###########################################################################
### COMPILER & COMPILER FLAGS       				  	###
###########################################################################

#FORTRAN COMPILER
MPFC = mpixlf90_r
ARCHIVE = ar r 

#FORTAN COMPILER FLAGS
ifeq ($(PRODRUN),yes)
 OPTLEVEL = 4
 IPA= -qipa
else
 OPTLEVEL = 3
endif
ifeq ($(DEBUG),yes)
 OPT= -qcheck -g -O0 -C
else
 OPT= -O$(OPTLEVEL) $(IPA) -Q -qarch=450d -qtune=450\
	-qunroll -qalias=nopteovrlp:nointptr:std
endif
FFLAGS = $(OPT) -qmaxmem=-1 -qmoddir=$(OBJDIR) -qflag=I:I

#PRECOMILER SWITCHES
PREPROC += -WF,-DAIX
PREPROC += -WF,$(SVNDEF)

#LINKER (usually same as MPFC)
MPLD =$(MPFC)

#LDFLAGS
LDFLAGS =

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################

#OPENMP
ifeq ($(OPENMP),yes)
FFLAGS +=
LDFLAGS += 
PREPROC += -WF,-DWITHOMP
endif

#Specify compiler flag for conversion to double precision
#(e.g. -r8, -qrealsize=8, etc.)
ifeq ($(PRECISION),double)
FFLAGS += -qautodbl=dbl4
PREPROC += -WF,-DDOUBLE_PREC
endif


###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

#Use mpi library wrapper
$(OBJDIR)/comm.o: $(OBJDIR)/mpimod.o
$(OBJDIR)/storefullmatrix.o: $(OBJDIR)/mpimod.o


#------- mpirun
run:	$(EXECDIR)/$(EXEC)
	cd $(RUNDIR);\
	OMP_NUM_THREADS=$(OMP_NUM_THREADS)\
	mpirun -np $(N_PES) -mode VN ./$(EXEC)

###########################################################################
### Machine dependet compiling rules                                    ###
###########################################################################

# The following compling rule is due to a compiler bug. If used -O3 
# the compiler does make an error so we need -qstrict also.
# This rule can be deleted, if the compiler has been repaired.
$(OBJDIR)/vel_space.o: $(SRCDIR)/vel_space.F90
	@echo "Special compiling rule for " $<
	$(MPFC) $(FFLAGS) -qstrict $(INCPATHS) $(PREPROC) -c -o $@ $<

