###########################################################################
### architecture dependent GENE makefiles; Cobra version                ###
###########################################################################

###########################################################################
### SWITCHES                                                            ###
###########################################################################

# FFTLIB - needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB=mkl
PRECISION=double
OPENMP=no

SLEPC=no
DEBUG=no
USE_PERFLIB=none
SCALAPACK=no

#TO DO
#insert memory per core (hardware mem. minus OS consumpt. minus MPI consumpt.)
PREPROC= -D'MB_PER_CORE=2500'

###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
INCPATHS += -I$(OBJDIR) -I. -I$(SRCDIR)

MKLPATH=/opt/intel/Compiler/11.1.056/mkl
MKLLIBPATH=$(MKLPATH)/lib/em64t

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
LIBS +=
#-L/usr/lib64 -lblas

#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 INCPATHS += 
#-I$(MKLPATH)/include -I$(MKLPATH)/mkl/include/fftw
 LIBS += 
#-L$(MKLPATH)/lib/em64t  -lmkl_em64t -Bstatic -lguide -Bdynamic -lpthread -lm
endif
ifeq ($(FFTLIB),fftw)
 INCPATHS += 
 INCPATHS += 
 LIBS += -L/opt/fftw_3.2.2/lib
  ifeq ($(PRECISION),double)
  LIBS += 
  else
  LIBS += 
  endif
endif

ifeq ($(SCALAPACK),yes)
  LIBS += -Wl,--start-group $(MKLLIBPATH)/libmkl_scalapack_lp64.a \
$(MKLLIBPATH)/libmkl_blacs_intelmpi_lp64.a $(MKLLIBPATH)/libmkl_intel_lp64.a \
$(MKLLIBPATH)/libmkl_sequential.a $(MKLLIBPATH)/libmkl_core.a -Wl,--end-group \
-lpthread 
  PREPROC += -DWITHSCAL
else
#Insert BLAS library                                                                       
LIBS += -L$(MKLLIBPATH)
LIBS += -Wl,--start-group $(MKLLIBPATH)/libmkl_intel_lp64.a \
$(MKLLIBPATH)/libmkl_sequential.a $(MKLLIBPATH)/libmkl_core.a \
-Wl,--end-group -lpthread
endif


###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(SLEPC),yes)
 PETSC_ARCH = linux-gnu-c-opt
 PETSC_DIR = /home/malbrech/malbrech/INSTALL_PERSO/petsc-3.0.0-p12
 SLEPC_DIR = /home/malbrech/malbrech/INSTALL_PERSO/slepc-3.0.0-p7
 include $(SLEPC_DIR)/conf/slepc_common
 INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
 LIBS +=$(PETSC_LIB) $(SLEPC_LIB)
 PREPROC += -DWITHSLEPC
endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -DWITHPERF
endif

###########################################################################
### COMPILER & COMPILER FLAGS       				  	###
###########################################################################

#FORTRAN COMPILER
MPFC =/home/malbrech/malbrech/INSTALL_PERSO/openmpi-1.4.1/bin/mpif90
####MPFC=mpif90


#ARCHIVE command
ARCHIVE = ar r

#FORTAN COMPILER FLAGS
FFLAGS = -O3 -module $(OBJDIR)
# -xP -rcd -static-intel

#PRECOMPILER SWITCHES
PREPROC += -DLINUX -DMPI  $(SVNDEF)

#LINKER (usually same as MPFC)
MPLD =$(MPFC)

#LDFLAGS 
LDFLAGS = -lsvml

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################

#OPENMP
ifeq ($(OPENMP),yes)
FFLAGS += -openmp -fpp
LDFLAGS += -openmp
PREPROC += -DWITHOMP
endif

MKLFLAGS = $(FFLAGS)

#Specify compiler flag for conversion to double precision
#(e.g. -r8, -qrealsize=8, etc.)
ifeq ($(PRECISION),double)
FFLAGS += -r8
PREPROC += -DDOUBLE_PREC
endif

###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

#Use mpi library wrapper
MPIDEPOBJ = \
	$(OBJDIR)/BoundaryDescription.o\
	$(OBJDIR)/boundary_exchange_general.F90\
	$(OBJDIR)/boundary.o\
	$(OBJDIR)/checkpoint.o\
	$(OBJDIR)/comm.o\
	$(OBJDIR)/field_solve_xky.o\
	$(OBJDIR)/gene.o\
	$(OBJDIR)/gyro_average_xky.o\
	$(OBJDIR)/storebandedmatrix.o\
	$(OBJDIR)/storefullmatrix.o\
	$(OBJDIR)/storevector.o

$(MPIDEPOBJ): $(OBJDIR)/mpimod.o

#------- mpirun
run:	$(EXECDIR)/$(EXEC)
	ulimit -s unlimited;\
	cd $(RUNDIR);\
	export OMP_NUM_THREADS=$(OMP_NUM_THREADS);\
	export MKL_SERIAL=yes;\
	mpirun -np $(N_PES) ./$(EXEC)



###########################################################################
### Machine dependent compiling rules                                   ###
###########################################################################
#additional compilations, e.g., of MKL library files, might be 
#necessary on some machines (see for instance bob_cluster)
#
ifeq ($(FFTLIB),mkl)
 $(OBJDIR)/par_mod.o:		$(OBJDIR)/mkl_dfti.o
 $(OBJDIR)/fourier_mkl.o:	$(OBJDIR)/mkl_dfti.o
 $(OBJDIR)/mkl_dfti.o: 		$(MKLPATH)/include/mkl_dfti.f90
	@echo "dfti-ing" $<
	$(MPFC) $(MKLFLAGS) -c -o $@ $< -module $(OBJDIR)
endif
