###########################################################################
### architecture dependent GENE makefile, Regatta version               ###
###########################################################################

###########################################################################
### SWITCHES                                                            ###
###########################################################################

# FFTLIB defines the library used for Fourier transformations. Needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB = essl
PRECISION= double
OPENMP = no

SLEPC= yes
USE_PERFLIB = none
#USE_PERFLIB = none, hpm or perf
SCALAPACK = no

DEBUG = no
OUTPUT=no
PRODRUN = no

BITMODE = 64

#memory per core (32 GB per node minus approx. value consumed by OS)
PREPROC= -WF,-D'MB_PER_CORE=3500'

###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
INCPATHS = -I$(OBJDIR) -I. -I$(SRCDIR)

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
LIBS = -lxlf90 -lxlsmp -lm -lc -lpthreads -lxlfpmt4

#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 INCPATHS +=
 LIBS +=
endif
ifeq ($(FFTLIB),essl)
 INCPATHS +=
 LIBS += -lesslsmp
 PREPROC+=-WF,-DWITHESSL
endif
ifeq ($(FFTLIB),fftw)
 INCPATHS += $(FFTW)
 LIBS += $(FFTW) -lfftw3
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(SLEPC),yes)
#module load petsc/2.3.3_O_complex
#module load slepc
 ifeq ($(PRECISION),double)
#  include $(SLEPC_DIR)/conf/slepc_common
  include $(SLEPC_DIR)/bmake/slepc_common
  INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
  LIBS +=$(PETSC_LIB) $(SLEPC_LIB)
  PREPROC += -WF,-DWITHSLEPC 
 else
  #PETSC_ARCH = AIX_SCOMPLEX
  SLEPC=no
 endif	

endif


ifeq ($(SCALAPACK),yes)
  LIBS += -lscalapack -llapack-essl
  PREPROC += -WF,-DWITHSCAL
endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -WF,-DWITHPERF
endif
ifeq ($(USE_PERFLIB),hpm)
 HPM_DIR = /afs/rzg/@sys/lib/HPM_V2_4
 INCPATHS += -I$(HPM_DIR)
 LIBS += -L(HPM_DIR)/lib -lhpm_r -lperfhpm_r -lpmapi -lm
endif
ifeq ($(USE_PERFLIB),perf)
 LIBS += -lperf_r -lpmapi
endif

###########################################################################
### COMPILER & COMPILER FLAGS                                           ###
###########################################################################

#FORTRAN COMPILER
MPFC = mpxlf95_r
ARCHIVE = ar -X$(BITMODE) r

#FORTAN COMPILER FLAGS
FFLAGS = -q$(BITMODE) -qmaxmem=-1 -qmoddir=$(OBJDIR) -qflag=I:I -qsuffix=cpp=F90

#PRECOMILER SWITCHES
PREPROC += -WF,-DAIX
PREPROC += -WF,$(SVNDEF)

#LINKER (usually same as MPFC)
MPLD =$(MPFC)

#LDFLAGS
LDFLAGS = -q$(BITMODE) -bh:4 -b64

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################

MACHINE_OPT = -qarch=auto -qtune=auto

ifeq ($(PRODRUN),yes)
 OPTLEVEL = 4
 IPA= -qipa=level=0 -qhot 
 #IPA = -qhot
else
 OPTLEVEL = 3
endif

ifeq ($(DEBUG),yes)
 FLTTRAP= -qflttrap=overflow:enable:invalid:zerodivide
 OPT= -qcheck -qsigtrap -g -C $(FLTTRAP)
else
 OPT= -O$(OPTLEVEL) $(IPA) -Q $(MACHINE_OPT) -qalign=4k\
	-qunroll -qalias=:nopteovrlp:nointptr:std
endif

SUPPRESS = -qsuppress=1500-036:1518-061:1518-128:1518-051:1501-245
FFLAGS += $(OPT) $(SUPPRESS)
LDFLAGS += $(OPT) $(SUPPRESS)

ifeq ($(BITMODE),32)
 LDFLAGS  += -bmaxdata:2147483648
endif


#Specify compiler flag for conversion to double precision
#(e.g. -r8, -qrealsize, etc.)
ifeq ($(PRECISION),double)
 FFLAGS += -qautodbl=dbl4
#-qrealsize=8
 PREPROC+=-WF,-DDOUBLE_PREC
endif

#OPENMP
ifeq ($(OPENMP),yes)
 FFLAGS += -qsmp=omp
 PREPROC+=-WF,-DWITHOMP
 LDFLAGS += 
endif

###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

#------- mpirun
run:	$(EXECDIR)/$(EXEC)
	cd $(RUNDIR);\
	OMP_NUM_THREADS=1\
	poe ./gene -tasks_per_node 8 -nodes 1 -retry_count 600 -retry 2

check:	$(F90FULLSRC)
	forchk -l listfile  -f03 -I /afs/ipp/common/soft/mpich2/1.0.5p2/i686/f95i.9.1/include \
	-ff $(SRCDIR)/mpimod.F90 $(F90FULLSRC)

###########################################################################
### Machine dependent compiling rules                                   ###
###########################################################################
