###########################################################################
### architecture dependent GENE makefile for Cadmos BlueGene/P at EPFL  ###
### see http://bluegene.epfl.ch                                         ###
###########################################################################

###########################################################################
### SWITCHES                                                            ###
###########################################################################

# FFTLIB - needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB = fftw
PRECISION = double
OPENMP = no

SLEPC= yes
# USE_PERFLIB set to perf, none, hpm or scalasca
USE_PERFLIB = none
MASSLIB = no
DEBUG = no
PRODRUN = no
SCALAPACK = yes

# CHEASE interface, HDF5 and futils are necessary
FUTILS=yes

#memory per core
PREPROC= -WF,-D'MB_PER_CORE=450'
###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
BGP_SYS = /bgsys/drivers/ppcfloor/comm
INCPATHS = -I$(OBJDIR) -I. -I$(SRCDIR) -I$(BGP_SYS)/include

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
LIBS = -L$(BGP_SYS)/lib -L/bgsys/lib

#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 INCPATHS +=
 LIBS +=
endif
ifeq ($(FFTLIB),essl)
 INCPATHS +=
 LIBS += -L/bgsys/ibm_essl/sles10/prod/opt/ibmmath/lib -lesslbg
 PREPROC += -WF,-DWITHESSL
endif
ifeq ($(FFTLIB),fftw)
 FFTW_HOME=/bgsys/local/fftw-3.2.2
 INCPATHS += -I$(FFTW_HOME)/include
 LIBS += -L$(FFTW_HOME)/lib -lfftw3
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(SLEPC),yes)
#define the following three variables!
# Note that on cadmos PETSC and SLEPSC
# have to be installed on the user 
# local machine
  PETSC_DIR=$(HOME)/petsc-3.1-p6
  SLEPC_DIR=$(HOME)/slepc-3.1-p5
  PETSC_ARCH =linux-gnu-cxx-opt

  include $(SLEPC_DIR)/conf/slepc_common
  INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
  #LIBS +=$(PETSC_LIB) $(SLEPC_LIB)
  LIBS += $(SLEPC_LIB)
  PREPROC += -WF,-DWITHSLEPC
endif

ifeq ($(SCALAPACK),yes)
  LIBS += -lscalapack -llapack
#-lblacsF77init -lblacs 
  LIBS += -L/bgsys/ibm_essl/sles10/prod/opt/ibmmath/lib -lesslbg
  PREPROC += -WF,-DWITHSCAL
endif

ifeq ($(FUTILS),yes)
 #  FUTILS and HDF5
 FUTILSDIR = $(EXTDIR)/$(MACHINE)/futils/src
 HDF5PATH = /bgsys
 HDF5_LIBPATH  = -L$(HDF5PATH)/lib -L$(FUTILSDIR)
 HDF5_LIBS = -lfutils -lhdf5_fortran -lhdf5 -lz 

 LIBS += $(HDF5_LIBPATH) $(HDF5_LIBS)
 PREPROC +=  -WF,-DWITHFUTILS
 HDF5VAR=/hdf5
 INCPATHS += -I$(HDF5PATH)/include$(HDF5VAR) -I$(FUTILSDIR)

endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -WF,-DWITHPERF
endif
ifeq ($(USE_PERFLIB),perf)
 LIBS += -L/usr/local/lib
endif

ifeq ($(MASSLIB),mass)
 MASS_LIB = -lmass
endif

###########################################################################
### COMPILER & COMPILER FLAGS       				  	###
###########################################################################

#FORTRAN COMPILER
MPFC = mpixlf90_r

CC = mpixlc_r
ARCHIVE = ar r 

#FORTAN COMPILER FLAGS
ifeq ($(PRODRUN),yes)
 OPTLEVEL = 4
 IPA= -qipa
else
 OPTLEVEL = 3
endif
ifeq ($(DEBUG),yes)
 FOPT= -qcheck -g -O0 -C
else
# COMMON_OPT= -O$(OPTLEVEL) $(IPA) 
COMMON_OPT = -Q -qarch=450d -qtune=450 -qunroll 
 FOPT = -qalias=:nopteovrlp:nointptr:std
endif
BGP_SYS_ARCH = /bgsys/drivers/ppcfloor/arch
COMMON_FLAGS = $(COMMON_OPT) -qmaxmem=-1 -qflag=I:I -I$(BGP_SYS_ARCH)/include
FFLAGS =  $(FOPT) -qmoddir=$(OBJDIR) 

#PRECOMILER SWITCHES
PREPROC += -WF,-DBGP
PREPROC += -WF,$(SVNDEF)
CPREPROC += -Wp,-DBGP

#LINKER (usually same as MPFC)
MPLD =$(MPFC)

#LDFLAGS
LDFLAGS =

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################

#OPENMP
ifeq ($(OPENMP),yes)
COMMON_FLAGS += -qsmp
LDFLAGS += 
PREPROC += -WF,-DWITHOMP
endif

#Specify compiler flag for conversion to double precision
#(e.g. -r8, -qrealsize=8, etc.)
ifeq ($(PRECISION),double)
FFLAGS += -qautodbl=dbl4
PREPROC += -WF,-DDOUBLE_PREC
endif

FFLAGS += $(COMMON_FLAGS)
CFLAGS += $(COMMON_FLAGS)
###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

$(EXECDIR)/BGP_test: $(OBJDIR)/BGP_personality_check.o
	$(MPLD) -o $@ $<

#------- mpirun
run:	$(EXECDIR)/$(EXEC)
	cd $(RUNDIR);\
	OMP_NUM_THREADS=$(OMP_NUM_THREADS)\
	mpirun -np $(N_PES) -mapfile TXYZ -mode VN ./$(EXEC)

###########################################################################
### Machine dependent compiling rules                                   ###
###########################################################################

#On BlueGene we need to use the mpimod wrapper:
MPIDEPOBJ = \
	$(OBJDIR)/BoundaryDescription.o\
	$(OBJDIR)/boundary_exchange_general.F90\
	$(OBJDIR)/boundary.o\
	$(OBJDIR)/checkpoint.o\
	$(OBJDIR)/comm.o\
	$(OBJDIR)/field_solve_xky.o\
	$(OBJDIR)/gene.o\
	$(OBJDIR)/gyro_average_xky.o\
	$(OBJDIR)/storebandedmatrix.o\
	$(OBJDIR)/storefullmatrix.o\
	$(OBJDIR)/storevector.o

$(MPIDEPOBJ): $(OBJDIR)/mpimod.o

# The following compling rule is due to a compiler bug. If used -O3 
# the compiler does make an error so we need -qstrict also.
# This rule can be deleted, if the compiler has been repaired.

$(OBJDIR)/boundary.o:$(SRCDIR)/boundary.F90
	@echo "Special compiling rule for " $<
	$(MPFC) $(FFLAGS) -qstrict $(INCPATHS) $(PREPROC) -c -o $@ $<

$(OBJDIR)/BGP_personality_check.o: $(SRCDIR)/BGP_personality_check.c
	@echo "Compiling " $<
	$(CC) $(CFLAGS) $(INCPATHS) -I/bgsys/drivers/ppcfloor/arch/include -c -o $@ $<
