###########################################################################
### template for architecture dependent GENE makefiles                  ###
###########################################################################

###########################################################################
### SWITCHES                                                            ###
###########################################################################

# FFTLIB - needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB = mkl
PRECISION= double
OPENMP = no
DEBUG = no

SLEPC=no
SCALAPACK = no
USE_PERFLIB = none

# FUTILS and HDF5 are required for some geometry interfaces
FUTILS = no

#insert memory per core and uncomment the following line
PREPROC= -D'MB_PER_CORE=800'

###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
INCPATHS = -I$(OBJDIR) -I. -I$(SRCDIR)

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
LIBS = 

#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 INCPATHS += -I$(INCLUDE)
 MKLPATH=$(INCLUDE)
 LIBS += -L$(LIBRARY_PATH) -Wl,--start-group $(LIBRARY_PATH)/libmkl_intel_lp64.a \
$(LIBRARY_PATH)/libmkl_sequential.a $(LIBRARY_PATH)/libmkl_core.a \
-Wl,--end-group -lpthread
endif
ifeq ($(FFTLIB),essl)
 INCPATHS +=
 LIBS +=
 PREPROC += -DWITHESSL
endif
ifeq ($(FFTLIB),fftw)
 INCPATHS +=
 LIBS +=
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(SLEPC),yes)
 PETSC_ARCH = DOUBLE_COMPLEX
 PETSC_DIR = /c3se/users/merz/software/petsc-3.0.0-p8
 SLEPC_DIR = /c3se/users/merz/software/slepc-3.0.0-p5
#slepc version < 3.0
# include $(SLEPC_DIR)/bmake/slepc_common
 include $(SLEPC_DIR)/conf/slepc_common
 INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
 LIBS +=$(PETSC_LIB) $(SLEPC_LIB)
 PREPROC += -DWITHSLEPC
endif

#SCALAPACK is only necessary if you want to use the direct eigenvalue solver
ifeq ($(SCALAPACK),yes)
  LIBS +=
  PREPROC += -DWITHSCAL
endif

ifeq ($(FUTILS),yes)
 #  FUTILS and HDF5
 FUTILSDIR = $(EXTDIR)/$(MACHINE)/futils/src
 HDF5PATH = 
 HDF5_LIBPATH  = -L$(HDF5PATH)/lib -L$(FUTILSDIR)
 HDF5_LIBS = -lfutils -lhdf5_fortran -lhdf5 -lz 

 LIBS += $(HDF5_LIBPATH) $(HDF5_LIBS)
 PREPROC +=  -DWITHFUTILS
 INCPATHS += -I$(FUTILSDIR)

endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -DWITHPERF
endif

###########################################################################
### COMPILER & COMPILER FLAGS       				  	###
###########################################################################

#FORTRAN COMPILER
MPFC = mpif90

#ARCHIVE command
#Note: On IBM machines you might need to add -X$(BITMODE) to the "ar" call
ARCHIVE = ar r

ifeq ($(DEBUG),yes)
  OPT = -O0
  FFLAGS += -C -traceback -debug inline-debug-info -warn all
  INCLUDE_SYMBOLS=yes
else
  OPT= -O3 -ip -xSSSE3
endif

#FORTAN COMPILER FLAGS
FFLAGS = $(OPT) -module $(OBJDIR)

#PRECOMILER SWITCHES
PREPROC += -DMPI $(SVNDEF)

#LINKER (usually same as MPFC)
MPLD =$(MPFC)

#LDFLAGS
LDFLAGS =

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################

#OPENMP
ifeq ($(OPENMP),yes)
FFLAGS +=
LDFLAGS += 
PREPROC +=-DWITHOMP
endif

#Specify compiler flag for conversion to double precision
#(e.g. -r8, -qrealsize=8, etc.)
ifeq ($(PRECISION),double)
FFLAGS += -r8
PREPROC +=-DDOUBLE_PREC
endif


###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

#------- mpirun 
run:	$(EXECDIR)/$(EXEC)
	cd $(RUNDIR);\
	MKL_SERIAL=YES\
	mpiexec -n $(N_PES) ./$(EXEC)

###########################################################################
### Machine dependent compiling rules                                   ###
###########################################################################

#additional compilations, e.g., of MKL library files, might be 
#necessary on some machines (see for instance bob_cluster)
#
#Furthermore, optimization options might be adapted for certain files here
ifeq ($(FFTLIB),mkl)
 $(OBJDIR)/par_mod.o:		$(OBJDIR)/mkl_dfti.o
 $(OBJDIR)/fourier_mkl.o:	$(OBJDIR)/mkl_dfti.o
 $(OBJDIR)/mkl_dfti.o: $(MKLPATH)/mkl_dfti.f90
	$(MPFC) $(MKLFLAGS) -c -o $@ $<  -module $(OBJDIR)
endif
