###########################################################################
### architecture dependent GENE makefiles, CRPP-CLUSTER version         ###
###########################################################################

##########################################################################
## The following modules need to be loaded :                            ##
## intel_comp/11.0e   (mkl is included with the compiler from 11.0 on)  ##
## mpich2/1.0.8-intel11.0						##
## hdf5									##
## if slepc = yes							##
## slepc/3.0.0								##
## petsc/3.0.0-complex-opt						##
## note : enter command 'module list' in pleiades2 terminal to display  ##
## loaded modules              						##
##########################################################################

###########################################################################
### SWITCHES                                                            ###
###########################################################################

# FFTLIB - needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB = fftw
PRECISION= double
OPENMP = no

# has to be switched on
SCALAPACK = yes

SLEPC= yes
USE_PERFLIB = none

# switching on extensive debugging -- slows down the code significantly
DEBUG = no
# switching on just source including and traceback -- code is still fast
# enough for doing longer runs
SOURCEINC = no

# CHEASE interface, HDF5 and futils are necessary
FUTILS=yes

#memory per core
PREPROC= -D'MB_PER_CORE=1900'
###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
INCPATHS = -I$(OBJDIR) -I. -I$(SRCDIR)

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
#LIBS = -L/usr/X11/lib -L$(MKL) -lsvml
#LIBS +=  -lmkl_scalapack -lmkl_lapack -lmkl -lguide -lpthread 
#LIBS += -Wl,--start-group $(MKL)/libmkl_intel_lp64.a \
#$(MKL)/libmkl_sequential.a $(MKL)/libmkl_core.a \
#-Wl,--end-group -lpthread


#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 INCPATHS +=
endif
ifeq ($(FFTLIB),essl)
 INCPATHS +=
 LIBS +=
 PREPROC += -DWITHESSL
endif
ifeq ($(FFTLIB),fftw)
 INCPATHS += -I/usr/local/fftw-3.0.1/include
 LIBS += -L/usr/local/fftw-3.0.1/lib 
 ifeq ($(PRECISION),double)
	LIBS += -lfftw3
 else
	LIBS += -lfftw3f
 endif
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(FUTILS),yes)
 #  FUTILS and HDF5
 FUTILSDIR = $(EXTDIR)/$(MACHINE)/futils/src
 HDF5PATH = $(HDF5)
 HDF5_LIBPATH  = -L$(HDF5PATH)/lib -L$(FUTILSDIR)
 HDF5_LIBS = -lfutils -lhdf5_fortran -lhdf5 -lz 
 INCPATHS += -I$(HDF5PATH)/lib

 LIBS += $(HDF5_LIBPATH) $(HDF5_LIBS)
 PREPROC +=  -DWITHFUTILS
 INCPATHS += -I$(FUTILSDIR)
endif


ifeq ($(SLEPC),yes)
 include  $(SLEPC_DIR)/conf/slepc_common
 INCPATHS += $(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
 LIBS += $(PETSC_LIB) $(SLEPC_LIB) 
 PREPROC += -DWITHSLEPC
endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -DWITHPERF
endif
ifeq ($(USE_PERFLIB),PERF)
 LIBS += -L$(BASEDIR)/perf -lperf_r
endif

ifeq ($(SCALAPACK),yes)
# LIBS += -lmkl_scalpack
  LIBS += -Wl,--start-group $(MKL)/libmkl_scalapack_lp64.a \
$(MKL)/libmkl_blacs_intelmpi20_lp64.a $(MKL)/libmkl_intel_lp64.a \
$(MKL)/libmkl_sequential.a $(MKL)/libmkl_core.a -Wl,--end-group \
-lpthread 
  PREPROC += -DWITHSCAL
 PREPROC += -DWITHSCAL
else
#Insert BLAS library
LIBS += -L$(MKLLIBPATH)
LIBS += -Wl,--start-group $(MKLLIBPATH)/libmkl_intel_lp64.a \
$(MKLLIBPATH)/libmkl_sequential.a $(MKLLIBPATH)/libmkl_core.a \
-Wl,--end-group -lpthread
endif

###########################################################################
### COMPILER & COMPILER FLAGS                                           ###
###########################################################################

#FORTRAN COMPILER
MPFC = mpif90

#ARCHIVE command
ARCHIVE = ar r

#FORTAN COMPILER FLAGS
MKLFLAGS = -module $(OBJDIR)

#PRECOMILER SWITCHES
PREPROC += -DLINUX -DMPI $(SVNDEF)

#LINKER (usually same as MPFC)
MPLD =$(MPFC)

#LDFLAGS
LDFLAGS =

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################
#OPENMP
ifeq ($(OPENMP),yes)
MKLFLAGS +=
LDFLAGS += 
PREPROC +=-DWITHOMP
endif

FFLAGS = $(MKLFLAGS)

ifeq ($(DEBUG),yes)
 OPT = -O0
 FFLAGS += $(OPT) -align -check -traceback -g -fpp1 -FR 
else 
 OPT = -O2
 ifeq ($(SOURCEINC),yes)
  FFLAGS  += -g -traceback $(OPT) -align  -FR 
 else
  FFLAGS  += $(OPT) -FR -align 
 endif
endif

#Specify compiler flag for conversion to double precision
#(e.g. -r8, -qautodbl, etc.)
ifeq ($(PRECISION),double)
FFLAGS += -r8
PREPROC +=-DDOUBLE_PREC
endif

###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

#------- mpirun
run:	$(EXECDIR)/$(EXEC)
	cd $(RUNDIR);\
	mpiexec -n $(N_PES) ./$(EXEC)


###########################################################################
### Machine dependent compiling rules                                   ###
###########################################################################

ifeq ($(FFTLIB),mkl)
$(OBJDIR)/par_mod.o:  	    $(OBJDIR)/mkl_dfti.o
$(OBJDIR)/fourier_mkl.o:    $(OBJDIR)/mkl_dfti.o
$(OBJDIR)/mkl_dfti.o: 	    $(MKL)/../../include/mkl_dfti.f90
	$(MPFC) $(MKLFLAGS) -c -o $@ $<
endif


