###########################################################################
### architecture dependent GENE makefile for RZG BlueGene/P             ###
### http://www.rzg.mpg.de/computing/hardware/BGP			###
###########################################################################

###########################################################################
### make sure you have the following line in your init script           ###
### module load hdf5/1.8.4p1 petsc slepc lapack scalapack               ###
###########################################################################

###########################################################################
### SWITCHES                                                            ###
###########################################################################

# FFTLIB - needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB = essl
PRECISION = double
OPENMP = no

SLEPC= yes
# USE_PERFLIB set to perf, none, hpm or scalasca
USE_PERFLIB = none
MASSLIB = no
DEBUG = no
PRODRUN = no
SCALAPACK = yes
FUTILS = yes

#memory per core
PREPROC= -WF,-D'MB_PER_CORE=450'
###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
INCPATHS = -I$(OBJDIR) -I. -I$(SRCDIR)

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
LIBS = 

#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 INCPATHS +=
 LIBS +=
endif
ifeq ($(FFTLIB),essl)
 INCPATHS +=
 LIBS += -L/opt/ibmmath/lib -lesslbg
 PREPROC += -WF,-DWITHESSL
endif
ifeq ($(FFTLIB),fftw)
 INCPATHS += -I$(FFTW_HOME)/include
 LIBS += -L$(FFTW_HOME)/lib -lfftw3
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(SLEPC),yes)
  PETSC_DIR=$(PETSC_HOME)/ppc64_sles10-mpi-xlf90-xlc-xlcxx-essl-double-cplx
  SLEPC_DIR=$(SLEPC_HOME)/ppc64_sles10-mpi-xlf90-xlc-xlcxx-essl-double-cplx
  PETSC_ARCH=
  include $(SLEPC_DIR)/conf/slepc_common
  INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
  LIBS += $(SLEPC_LIB)
  PREPROC += -WF,-DWITHSLEPC
endif

ifeq ($(SCALAPACK),yes)
  LIBS += -L$(SCALAPACK_HOME)/lib -lscalapack -lblacsF77init -lblacs -lblacsCinit -L/opt/ibmmath/lib -lesslbg -L$(LAPACK_HOME)/lib -llapack -lesslbg -Wl,--allow-multiple-definition
  PREPROC += -WF,-DWITHSCAL
endif

ifeq ($(FUTILS),yes)
 #  FUTILS and HDF5
 FUTILSDIR = $(EXTDIR)/$(MACHINE)/futils/src
 HDF5PATH = $(HDF5_HOME)
 HDF5_LIBPATH  = -L$(HDF5PATH)/lib -L$(FUTILSDIR)
 HDF5_LIBS = -lfutils -lhdf5_fortran -lhdf5 -lz 

 LIBS := $(HDF5_LIBPATH) $(HDF5_LIBS) $(LIBS)
 PREPROC +=  -WF,-DWITHFUTILS
 INCPATHS += -I$(FUTILSDIR)

endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -WF,-DWITHPERF
endif
ifeq ($(USE_PERFLIB),perf)
 LIBS += -L/usr/local/lib
endif

ifeq ($(MASSLIB),mass)
 MASS_LIB = -lmass
endif

###########################################################################
### COMPILER & COMPILER FLAGS       				  	###
###########################################################################

#FORTRAN COMPILER
ifeq ($(USE_PERFLIB),scalasca)
  MPFC = scalasca -instrument -user mpixlf90_r
else
  MPFC = mpixlf90_r
endif
CC = mpixlc_r
ARCHIVE = ar r 

#FORTAN COMPILER FLAGS
ifeq ($(PRODRUN),yes)
 OPTLEVEL = 4
 IPA= -qipa
else
 OPTLEVEL = 3
endif
ifeq ($(DEBUG),yes)
 FOPT= -qcheck -g -O0 -C
else
 COMMON_OPT= -O$(OPTLEVEL) $(IPA) -Q -qarch=450d -qtune=450\
	-qunroll 
 FOPT = -qalias=:nopteovrlp:nointptr:std
endif
COMMON_FLAGS = $(COMMON_OPT) -qmaxmem=-1 -qflag=I:I
FFLAGS =  $(FOPT) -qmoddir=$(OBJDIR) 

#PRECOMILER SWITCHES
PREPROC += -WF,-DBGP
PREPROC += -WF,$(SVNDEF)
CPREPROC += -Wp,-DBGP

#LINKER (usually same as MPFC)
MPLD =$(MPFC)

#LDFLAGS
LDFLAGS =

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################

#OPENMP
ifeq ($(OPENMP),yes)
COMMON_FLAGS += -qsmp
LDFLAGS += 
PREPROC += -WF,-DWITHOMP
endif

#Specify compiler flag for conversion to double precision
#(e.g. -r8, -qrealsize=8, etc.)
ifeq ($(PRECISION),double)
FFLAGS += -qautodbl=dbl4
PREPROC += -WF,-DDOUBLE_PREC
endif

FFLAGS += $(COMMON_FLAGS)
CFLAGS += $(COMMON_FLAGS)
###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

$(EXECDIR)/BGP_test: $(OBJDIR)/BGP_personality_check.o
	$(MPLD) -o $@ $<

#------- mpirun
run:	$(EXECDIR)/$(EXEC)
	cd $(RUNDIR);\
	OMP_NUM_THREADS=$(OMP_NUM_THREADS)\
	mpirun -np $(N_PES) -mapfile TXYZ -mode VN ./$(EXEC)

###########################################################################
### Machine dependent compiling rules                                   ###
###########################################################################

# The following compling rule is due to a compiler bug. If used -O3 
# the compiler does make an error so we need -qstrict also.
# This rule can be deleted, if the compiler has been repaired.

#$(OBJDIR)/boundary.o:$(SRCDIR)/boundary.F90
#	@echo "Special compiling rule for " $<
#	$(MPFC) $(FFLAGS) -qstrict $(INCPATHS) $(PREPROC) -c -o $@ $<

$(OBJDIR)/BGP_personality_check.o: $(SRCDIR)/BGP_personality_check.c
	@echo "Compiling " $<
	$(CC) $(CFLAGS) $(INCPATHS) -I/bgsys/drivers/ppcfloor/arch/include -c -o $@ $<
