###########################################################################
### architecture dependent GENE makefile, Cray XT4/6 version            ###
###########################################################################

###########################################################################
### include the following lines to ~/.bashrc without leading '#'        ###
###########################################################################
###choose between XT4 (normal) or X2 (vector nodes)
#export MACHINE=hector
### Load modules for compiler
#	module unload pgi PrgEnv-pgi
#	module load PrgEnv-pathscale pathscale
#	module load fftw hdf5-parallel petsc-complex/3.0.0.10
#
## on X2 nodes:
#	module add x2-env
## on phase2a (XT4) system:
#       module add xtpe-barcelona
## on phase2b (XT6) system:
#       module add xtpe-mc12
#
###########################################################################
### SWITCHES                                                            ###
###########################################################################

MPRUN = OMP_NUM_THREADS=$(OMP_NUM_THREADS)\
	aprun -n $(N_PES) -N $(NTASK) ./$(EXEC)

# FFTLIB defines the library used for Fourier transformations. 
# Needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB = fftw
PRECISION= double
OPENMP = no

SLEPC= no
SCALAPACK = yes
USE_PERFLIB = none
#USE_PERFLIB = none, hpm or perf

FUTILS=no 

DEBUG = no
OUTPUT=no
PRODRUN = no
RP5 = no
BITMODE = 64
COMPILER = cray
#COMPTYPE = pathscale

#memory per core 
#(on phase2a: officially 8GB/quad-core minus est. OS consumption)
#PREPROC= -D'MB_PER_CORE=1500'
#on phase2b:
PREPROC= -D'MB_PER_CORE=1100'

###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)


#LIBRARIES AND LIBFLAGS
#Insert BLAS library
LIBS =

#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 INCPATHS +=
 LIBS +=
endif

ifeq ($(FFTLIB),fftw)
 INCPATHS += $(FFTW_INCLUDE_OPTS)
 LIBS += $(FFTW_POST_LINK_OPTS)
#$(FFTW_LIB) -lfftw3f -lfftw3
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(SLEPC),yes)
 ifeq ($(COMPTYPE),pathscale)
   PETSC_DIR = $(CRAY_PETSC_DIR)
   #PETSC_DIR is only sometimes set in the modules for strange reason
   SLEPC_DIR = /home/e24/rzg19gen/rzg19tbg/soft/slepc-3.0.0-p7
 endif
 ifeq ($(COMPTYPE),pgi)
   SLEPC_DIR = /home/e24/rzg19gen/rzg19tbg/soft/slepc-3.1-p6
 endif
 include $(SLEPC_DIR)/conf/slepc_common
 INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
 LIBS += $(SLEPC_LIB)
 PREPROC += -DWITHSLEPC
endif


ifeq ($(USE_PERFLIB),hpm)
 HPM_DIR = 
 INCPATHS += -I$(HPM_DIR)
 LIBS += -L(HPM_DIR)/lib -lhpm_r -lperfhpm_r -lpmapi -lm
endif

#SCALAPACK is only necessary if you want to use the direct eigenvalue solver
ifeq ($(SCALAPACK),yes)
  #LIBS += -lsci
  PREPROC += -DWITHSCAL
endif

ifeq ($(FUTILS),yes)
 #  FUTILS and HDF5
 FUTILSDIR = $(EXTDIR)/$(MACHINE)/futils/src
 HDF5PATH = $(CRAY_HDF5_DIR)
 HDF5_LIBPATH  = -L$(HDF5PATH)/lib -L$(FUTILSDIR)
 HDF5_LIBS = -lfutils -lhdf5_fortran -lhdf5 -lz 

 LIBS += $(HDF5_LIBPATH) $(HDF5_LIBS)
 PREPROC +=  -DWITHFUTILS
 INCPATHS += -I$(FUTILSDIR)
endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -DWITHPERF
endif
ifeq ($(USE_PERFLIB),perf)
 LIBS += -lperf_r -lpmapi
endif

###########################################################################
### COMPILER & COMPILER FLAGS                                           ###
###########################################################################

#FORTRAN COMPILER
MPFC = ftn
ARCHIVE = ar r

#FORTAN COMPILER FLAGS


#PRECOMILER SWITCHES

#the loaded PGI compiler does not support Fortran 2003 standard 
#yet and requires external erf (error function) definition
#PREPROC += -DEXTERNAL_ERF -DF2003_NO_NEW_LINE

#LINKER (usually same as MPFC)
MPLD = $(MPFC)

#LDFLAGS
LDFLAGS = 

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################

ifeq ($(PRODRUN),yes)
 OPTLEVEL = 4
 IPA = -Mipa=fast,inline:3
else
 OPTLEVEL = 3
endif

ifeq ($(DEBUG),yes)
 OPT=	-g -O0 $(DEBUG_FFLAGS)
 ifeq ($(COMPTYPE),pgi)
   OPT+= -Mchkfpstk -Mchkptr -traceback
 endif
else
 ifeq ($(COMPTYPE),pgi)
   OPT= -fastsse
 endif
 ifeq ($(COMPTYPE),pathscale)
   OPT= -OPT:Ofast $(PATHSCALE_PRE_COMPILE_OPTS)
 endif 
 OPT+= -O$(OPTLEVEL)
endif

FFLAGS += $(OPT)
LDFLAGS += $(OPT)

ifeq ($(OUTPUT),yes)
PREPROC+= -DOUTPUT
endif


#OPENMP
ifeq ($(OPENMP),yes)
 FFLAGS += -mp
 PREPROC+= -DWITHOMP
 LDFLAGS += -mp
endif

###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

###########################################################################
### Machine dependent compiling rules                                   ###
###########################################################################
.SUFFIXES:
.SUFFIXES:
	.o .F90

