###########################################################################
### architecture dependent GENE makefile, Jaguar version                ###
###########################################################################

###########################################################################
### SWITCHES                                                            ###
###########################################################################

MPRUN = OMP_NUM_THREADS=1;\
	aprun -n $(N_PES) ./$(EXEC)

# FFTLIB defines the library used for Fourier transformations. 
# (Needed in /src/files.mk)
# set to: mkl, fftw or essl

FFTLIB = fftw
PRECISION= double
OPENMP = no

SLEPC= no
SCALAPACK = yes
USE_PERFLIB = none
#USE_PERFLIB = none, hpm or perf

DEBUG = no
OUTPUT=no
PRODRUN = no
RP5 = no
BITMODE = 64


###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
#FTN links automatically
INCPATHS = -I$(OBJDIR) -I. -I$(SRCDIR)

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
LIBS =
#FTN links automatically to BLAS

#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 INCPATHS +=
 LIBS +=
endif
ifeq ($(FFTLIB),essl)
 INCPATHS +=
 LIBS +=
 PREPROC += -DWITHESSL
endif
ifeq ($(FFTLIB),fftw)
 INCPATHS += -I/opt/fftw/3.1.1/cnos/include
 LIBS += $$FFTW_POST_LINK_OPTS
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(SLEPC),yes)
 ifeq ($(PRECISION),double)
  include $(SLEPC_DIR)/bmake/slepc_common
  INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
  LIBS +=$(PETSC_LIB) $(SLEPC_LIB)
  PREPROC += -DWITHSLEPC
 else
  SLEPC=no
 endif
endif

ifeq ($(SCALAPACK),yes)
  LIBS += /usr/common/acts/SCALAPACK/1.8.0/libscalapack.a
  PREPROC += -DWITHSCAL
endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -DWITHPERF
endif
ifeq ($(USE_PERFLIB),hpm)
 HPM_DIR = 
 INCPATHS += -I$(HPM_DIR)
 LIBS += -L(HPM_DIR)/lib -lhpm_r -lperfhpm_r -lpmapi -lm
endif
ifeq ($(USE_PERFLIB),perf)
 LIBS += -lperf_r -lpmapi
endif

###########################################################################
### COMPILER & COMPILER FLAGS                                           ###
###########################################################################

#FORTRAN COMPILER
MPFC = ftn
ARCHIVE = ar r

#FORTAN COMPILER FLAGS
FFLAGS = -module $(OBJDIR)


#PRECOMILER SWITCHES
PREPROC += -DMPI $(SVNDEF)
#erf(x) is not ftn intrinsic
PREPROC += -DEXTERNAL_ERF

#memory per core (officially, 2GB.
#However, the OS usually consumes a significant fraction.)
#Therefore:
PREPROC += -DMB_PER_CORE=1900

#LINKER (usually same as MPFC)
MPLD = $(MPFC)

#LDFLAGS
LDFLAGS = 

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################

ifeq ($(PRODRUN),yes)
 OPTLEVEL = 4
 IPA = -Mipa=fast,inline:3
else
 OPTLEVEL = 3
endif

ifeq ($(DEBUG),yes)
 OPT=	-O0 -Mbounds
else
 OPT=	-fastsse
 OPT+=	-O$(OPTLEVEL) -fast -Mnoopenmp -Minfo=opt,inline -Mneginfo=loop $(IPA)
endif

FFLAGS += $(OPT)
LDFLAGS += $(OPT)

ifeq ($(OUTPUT),yes)
PREPROC+= -DOUTPUT
endif

ifeq ($(PRECISION),double)
 FFLAGS += -r8
 PREPROC+= -DDOUBLE_PREC -D'erf=derf'
endif

#OPENMP
ifeq ($(OPENMP),yes)
 FFLAGS += 
 PREPROC+= -DWITHOMP
 LDFLAGS += 
endif

###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

###########################################################################
### Compiling rules                                                     ###
###########################################################################
.SUFFIXES:
.SUFFIXES:
	.o .F90
