###########################################################################
## GENE makefile for JUGENE                                              ##
## http://www2.fz-juelich.de/jsc/jugene	                                 ##
###########################################################################

###########################################################################
### make sure you have the following line in your ~/.bashrc script      ###
### module load lapack blacs scalapack                                  ###
###########################################################################

###########################################################################
### SWITCHES                                                            ###
###########################################################################

# FFTLIB - needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB = essl
PRECISION = double
OPENMP = no

SLEPC= yes
# USE_PERFLIB set to perf, none, hpm or scalasca
USE_PERFLIB = none
MASSLIB = no
FUTILS=no
DEBUG = no
PRODRUN = yes
SCALAPACK = yes

#memory per core
PREPROC= -WF,-D'MB_PER_CORE=450'

###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
BGP_SYS = /bgsys/drivers/ppcfloor/comm
INCPATHS = -I$(OBJDIR) -I. -I$(SRCDIR) -I$(BGP_SYS)/include

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
LIBS = -L$(BGP_SYS)/lib

#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 INCPATHS +=
 LIBS +=
endif
ifeq ($(FFTLIB),essl)
 INCPATHS +=
 LIBS += -L/opt/ibmmath/lib -lesslbg
 PREPROC += -WF,-DWITHESSL
endif
ifeq ($(FFTLIB),fftw)
 INCPATHS +=
 LIBS +=
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(FUTILS),yes)
 #  FUTILS and HDF5
 FUTILSDIR = $(EXTDIR)/$(MACHINE)/futils/src
 HDF5PATH = /bgsys/local/hdf5
 HDF5_LIBPATH  = -L$(HDF5PATH)/lib -L$(FUTILSDIR) -L/bgsys/local/libjpeg/v6b_450/lib\
                 -L/bgsys/local/zlib/lib -L/bgsys/local/szip/lib
 HDF5LIBS = -lfutils -lhdf5_fortran -lhdf5 -ljpeg -lz -lsz
 export HDF5_LIBS=$(HDF5_LIBPATH) $(HDF5LIBS)
 LIBS += $(HDF5_LIBPATH) $(HDF5LIBS)
 PREPROC +=  -WF,-DWITHFUTILS
 INCPATHS += -I$(FUTILSDIR)

endif

ifeq ($(SLEPC),yes)
  PETSC_DIR=/homea/pra047/pra04701/soft/petsc/
  SLEPC_DIR=/homea/pra047/pra04701/soft/slepc/
  include $(SLEPC_DIR)/conf/slepc_common
  INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
  #LIBS +=$(PETSC_LIB) $(SLEPC_LIB)
  LIBS += $(SLEPC_LIB)
  PREPROC += -WF,-DWITHSLEPC
endif

ifeq ($(SCALAPACK),yes)
  LIBS += -L$(SCALAPACK_LIB) -lscalapack
  LIBS += -L$(LAPACK_LIB) -llapack
  LIBS += -L$(BLACS_LIB) -lblacsF77init -lblacs
  LIBS += -L/opt/ibmmath/lib -lesslbg
  PREPROC += -WF,-DWITHSCAL
endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -WF,-DWITHPERF
endif
ifeq ($(USE_PERFLIB),perf)
 LIBS += -L/usr/local/lib
endif

ifeq ($(MASSLIB),mass)
 MASS_LIB = -lmass
endif

###########################################################################
### COMPILER & COMPILER FLAGS       				  	###
###########################################################################

#FORTRAN COMPILER
ifeq ($(USE_PERFLIB),scalasca)
  MPFC = scalasca -instrument -user mpixlf90_r
else
  MPFC = mpixlf90_r
endif
ARCHIVE = ar r 

#FORTAN COMPILER FLAGS
ifeq ($(PRODRUN),yes)
 OPTLEVEL = 4
 IPA= -qipa=level=0 -qhot
else
 OPTLEVEL = 3
endif
ifeq ($(DEBUG),yes)
 FLTTRAP= -qflttrap=overflow:enable:invalid:zerodivide:nanq
 FOPT= -O0 -C -qcheck -qsigtrap $(FLTTRAP)
else
 OPT= -O$(OPTLEVEL) $(IPA) -Q -qarch=450d -qtune=450\
	-qunroll -qalias=:nopteovrlp:nointptr:std
endif
FFLAGS = $(OPT) -qmaxmem=-1 -qmoddir=$(OBJDIR) -qflag=I:I

#PRECOMILER SWITCHES
PREPROC += -WF,-DBGP
PREPROC += -WF,$(SVNDEF)
CPREPROC += -Wp,-DBGP

#LINKER (usually same as MPFC)
MPLD =$(MPFC)

#LDFLAGS
LDFLAGS =

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################
#OPENMP
ifeq ($(OPENMP),yes)
FFLAGS += -qsmp
LDFLAGS += 
PREPROC += -WF,-DWITHOMP
endif

#Specify compiler flag for conversion to double precision
#(e.g. -r8, -qrealsize=8, etc.)
ifeq ($(PRECISION),double)
FFLAGS += -qautodbl=dbl4
PREPROC += -WF,-DDOUBLE_PREC
endif


###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)


#------- mpirun
run:	$(EXECDIR)/$(EXEC)
	cd $(RUNDIR);\
	OMP_NUM_THREADS=$(OMP_NUM_THREADS)\
	mpirun -np $(N_PES) -mapfile TXYZ -mode VN ./$(EXEC)

###########################################################################
### Machine dependent compiling rules                                   ###
###########################################################################

#On Jugene we need to use the mpimod wrapper:
MPIDEPOBJ = \
	$(OBJDIR)/BoundaryDescription.o\
	$(OBJDIR)/boundary_exchange_general.F90\
	$(OBJDIR)/boundary.o\
	$(OBJDIR)/checkpoint.o\
	$(OBJDIR)/comm.o\
	$(OBJDIR)/field_solve_xky.o\
	$(OBJDIR)/gene.o\
	$(OBJDIR)/gyro_average_xky.o\
	$(OBJDIR)/storebandedmatrix.o\
	$(OBJDIR)/storefullmatrix.o\
	$(OBJDIR)/storevector.o

$(MPIDEPOBJ): $(OBJDIR)/mpimod.o


$(OBJDIR)/boundary.o:$(SRCDIR)/boundary.F90
	@echo "Special compiling rule for " $<
	$(MPFC) $(FFLAGS) -qstrict $(INCPATHS) $(PREPROC) -c -o $@ $<


# The following compling rule is due to a compiler bug. If used -O3 
# the compiler does make an error so we need -qstrict also.
# This rule can be deleted, if the compiler has been repaired.
#$(OBJDIR)/vel_space.o: $(SRCDIR)/vel_space.F90
#	@echo "Special compiling rule for " $<
#	$(MPFC) $(FFLAGS) -qstrict $(INCPATHS) $(PREPROC) -c -o $@ $<

