#!/usr/bin/env python


def build_tree(mytree,coord,ancestors,igrid):
    """igrid the coordinates"""

    #print "Root coordinates: ",coord, " with ancestors = ",ancestors," on igrid = ", igrid
    alist = [coord]
    alist.extend(ancestors)

    if coord[0]+1 < igrid[0] and [coord[0]+1,coord[1],coord[2]] not in ancestors:
        mytree.append({'coord':[coord[0]+1,coord[1],coord[2]],'children':[],'ancestors':alist})
        
    if coord[0]-1 >= 0 and [coord[0]-1,coord[1],coord[2]] not in ancestors:
        mytree.append({'coord':[coord[0]-1,coord[1],coord[2]],'children':[],'ancestors':alist})

    if coord[1]+1 < igrid[1] and [coord[0],coord[1]+1,coord[2]] not in ancestors:
        mytree.append({'coord':[coord[0],coord[1]+1,coord[2]],'children':[],'ancestors':alist})

    if coord[1]-1 >= 0 and [coord[0],coord[1]-1,coord[2]] not in ancestors:
            mytree.append({'coord':[coord[0],coord[1]-1,coord[2]],'children':[],'ancestors':alist})

    if coord[2]+1 < igrid[2] and [coord[0],coord[1],coord[2]+1] not in ancestors:
        mytree.append({'coord':[coord[0],coord[1],coord[2]+1],'children':[],'ancestors':alist})

    if coord[2]-1 >= 0 and [coord[0],coord[1],coord[2]-1] not in ancestors:
        mytree.append({'coord':[coord[0],coord[1],coord[2]-1],'children':[],'ancestors':alist})
    
    for subtree in mytree:
        build_tree(subtree['children'],subtree['coord'],subtree['ancestors'],igrid)

def get_subpaths(mytree):
    #print "Working on node with coord = ",mytree['coord']," and ",len(mytree['children'])," children."
    local_list = []
    if len(mytree['children'])==0:
        # Leaf node, just return the coordinates
        return [[mytree['coord']]]
    else:
        templist = []
        for c in mytree['children']:
            sublist = get_subpaths(c)
            #print "sublist = ",sublist
            templist.extend(sublist)

        for sl in templist:
            sl.append(mytree['coord'])

        return templist

def get_cost_alltoall(igrid,bgrid,pgrid):
    """get_cost_alltoall computes the cost of an alltoall communication
    of a given inner grid (3D array) of building blocks of dimension bgrid 
    on a outer grid pgrid."""

    # first we build a list of all points of the outer grid
    cost = 0
    for k in range(igrid[2]):
        for j in range(igrid[1]):
            for i in range(igrid[0]):
                # starting point is i,j,k
                # loop over target points
                for tk in range(igrid[2]):
                    for tj in range(igrid[1]):
                        for ti in range(igrid[0]):
                            dx = abs(ti-i)*bgrid[0]
                            dy = abs(tj-j)*bgrid[1]
                            dz = abs(tk-k)*bgrid[2]
                            if dx>pgrid[0] or dy>pgrid[1] or dz>pgrid[2]:
                                return 0
                            else:
                                cost += min([dx,pgrid[0]-dx])\
                                    +min([dy,pgrid[1]-dy])\
                                    +min([dz,pgrid[2]-dz])
    return cost

def get_cost_nearest_neighbor(plist,bgrid,pgrid):
    """calculates the cost for nearest neighbor communication in a
    list of building blocks (bgrid) given by plist in a total network of
    pgrid processes."""
    
    cost = 0
    length = len(plist)
    for ip in xrange(length):
        p = plist[ip]
        p_succ = plist[(ip+1)%length]
        p_pred = plist[(ip-1+length)%length]
        
        # to successor
        dx = abs(p_succ[0]-p[0])*bgrid[0]
        dy = abs(p_succ[1]-p[1])*bgrid[1]
        dz = abs(p_succ[2]-p[2])*bgrid[2]
        cost += min([dx,pgrid[0]-dx])\
            +min([dy,pgrid[1]-dy])\
            +min([dz,pgrid[2]-dz])

        # to predecessor
        dx = abs(p_pred[0]-p[0])*bgrid[0]
        dy = abs(p_pred[1]-p[1])*bgrid[1]
        dz = abs(p_pred[2]-p[2])*bgrid[2]
        cost += min([dx,pgrid[0]-dx])\
            +min([dy,pgrid[1]-dy])\
            +min([dz,pgrid[2]-dz])

    return cost

def sort_for_cost(a,b):
    #print "a = ",a,", b = ", b,a['cost'],b['cost']
    return a['cost']-b['cost']

def get_all_combinations(np):
    # Assumption: np is a power of 2
    power = 0
    remainder = np
    while remainder>1:
        remainder = remainder/2
        power += 1

    # np = 2^power
    # Verteilungsmoeglichkeiten von np auf dem Prozessgitter
    #print "We have np = %u = 2^%u\n"%(np,power)
    combination = []
    for powx in range(power+1):
        for powy in range(power+1-powx):
            powz = power-powx-powy
            combination.append([powx,powy,powz]);
    return combination

def optimize_alltoall(np,bgrid,pgrid):
    """ Take the input parameter np, which is the number of processes
    in the alltoall communicator and calculate the cost for all valid 
    permutations. It is assumed that np is a power of 2."""

    print "Starting optimize_alltoall with np = ", np, ", bgrid = ",bgrid, " and pgrid = ", pgrid,"."
    combination = get_all_combinations(np)
    cost_array = []
    for combi in combination:
        #outerTorus = [1,1,1]
        npx = pow(2, combi[0])
        npy = pow(2, combi[1])
        npz = pow(2, combi[2])
        if npx*bgrid[0]>pgrid[0] or npy*bgrid[1]>pgrid[1] or npz*bgrid[2]>pgrid[2]:
            continue
        cost_array.append({'np':[npx,npy,npz],'pgrid':pgrid,'cost': get_cost_alltoall([npx,npy,npz],bgrid,pgrid), 'type':'alltoall'})
    
    cost_array.sort(sort_for_cost)
    return cost_array[0]


def optimize_nearest_neighbor(np,bgrid,pgrid):
    """ assumption: np is power of 2 
    Periodicity is also assumed, otherwise a each continuous line is optimal."""

    print "Starting optimize_nearest_neighbor with np = ", np," bgrid = ", bgrid," and pgrid = ",pgrid,"."
    combination = get_all_combinations(np)

    # Check all orders for all combinations
    cost_array = []
    for combi in combination:
        # check if combi is valid, fits into pgrid
        np_grid = [2**x for x in combi]
        if np_grid[0]*bgrid[0]<=pgrid[0] and \
                np_grid[1]*bgrid[1]<=pgrid[1] and \
                np_grid[2]*bgrid[2]<=pgrid[2]:
            # construct tree for valid combi
            mytree = {'coord':[0,0,0],'ancestors':[],'children':[]}
            build_tree(mytree['children'],[0,0,0],[],np_grid)

            paths = get_subpaths(mytree)
            for p in paths:
                # We want only complete pathes
                if len(p)==reduce(lambda x,y:x*y,np_grid):
                    #print "path = ",p
                    cost_array.append({'path':p,'np':np_grid,'cost':get_cost_nearest_neighbor(p,bgrid,pgrid),'type':'nn'})
        
    cost_array.sort(sort_for_cost)
    return cost_array[0]

def optimize_collective(np,bgrid,pgrid):
    print "Starting optimize_collective with np = ", np," bgrid = ", bgrid," and pgrid = ",pgrid,"."
    combinations = get_all_combinations(np)

    valid_combi = []
    for combi in combinations:
        #outerTorus = [1,1,1]
        npx = pow(2, combi[0])
        npy = pow(2, combi[1])
        npz = pow(2, combi[2])
        if npx*bgrid[0]>pgrid[0] or npy*bgrid[1]>pgrid[1] or npz*bgrid[2]>pgrid[2]:
            continue
        valid_combi.append({'np':[npx,npy,npz],'pgrid':pgrid,'type':'collective'})
    
    #return map(lambda x,y:x*y,valid_combi[0]['np'],bgrid)
    if len(valid_combi)==0:
        print "no valid combination found for collective communication."
        sys.exit(1)
    else:
        return valid_combi[0]

def write_innermost_block(dim,where,outfile):
    """Write some lines in the given outfile, for the inner block, which left bottom corner 
    is given in where."""
    for npz in xrange(dim[2]):
        for npy in xrange(dim[1]):
            for npx in xrange(dim[0]):
                outfile.write("%4u %4u %4u %4u\n"\
                                  %(where[0]+npx,where[1]+npy,where[2]+npz,0))
    
def dist(coord):
    """ calculates the distance to [0,0,0] in adjacent unit steps"""
    return abs(coord[0])+abs(coord[1])+abs(coord[2])

def write_block(bc_arr,where,outerblock,outfile,pernode):
    """bc_arr is the best cost array, which has an entry for each
    level beyond the actual and the actual. The entry is a dictionary with at 
    least the elements 'np'. For 'nn' communication it also must have an entry 'path'.
    
    We take the last entry, do some work and pass the remainder to a recursive call."""

    if len(bc_arr)==0:
        for core in xrange(pernode):
            outfile.write("%6u %6u %6u %6u\n"%(where[0],where[1],where[2],core))
        return
    else:
        bc = bc_arr.pop()
        np = bc['np']
        blocksize = map(lambda x,y:x/y,outerblock,np)

        if bc['type']=='nn':
            # follow path
            path = bc['path'][::-1]
            #path.reverse()
            #print "path = ",path, ", where = ", where
            for point in path:
                write_block(bc_arr,map(lambda x,y:x+y,where,map(lambda x,y:x*y,point,blocksize)),blocksize,outfile,pernode)
        else:
            # follow x,y,z ordering
            for ipz in xrange(np[2]):
                start_z = where[2]+ipz*blocksize[2]
                for ipy in xrange(np[1]):
                    start_y = where[1]+ipy*blocksize[1]
                    for ipx in xrange(np[0]):
                        start_x = where[0]+ipx*blocksize[0]
                        # now write the next inner block
                        write_block(bc_arr,[start_x,start_y,start_z],blocksize,outfile,pernode)
        # restore last element
        bc_arr.append(bc)

##########################################################
################ Start main program ######################
##########################################################

import sys
from optparse import OptionParser

parser = OptionParser("optimize_mapfile.py [options] n_procs")
parser.add_option("-t","--type", dest="type",help="array of the types of communication, use 'alltoall', 'nn', 'collective' as entries.")
parser.add_option("-p","--mgrid",dest="mgrid",help="3D array of the midplane topology")
parser.add_option("-m","--mode", dest="mode",help="Mode of BlueGene operation, SMP, DUAL or VN.")
(options,args) = parser.parse_args()

print "options = ", options,", args = ",args

if options.mode:
    BG_mode = options.mode
else:
    # default is SMP mode
    BG_mode = "SMP"

if options.mgrid:
    # parsing of the option
    mgrid = [int(x) for x in options.mgrid.split()]
else:
    # default is one midplane
    mgrid = [1,1,1]

pgrid = [8*x for x in mgrid]

if BG_mode=="VN":
    topology={'per node':4,'pgrid':pgrid}
    pernode = 4
elif BG_mode=="DUAL":
    topology={'per node':2,'pgrid':pgrid}
    pernode = 2
else:
    topology={'per node':1,'pgrid':pgrid}
    pernode = 1
n_procs = reduce(lambda x,y:x*y,topology['pgrid'])*topology['per node']


print "The array of midplanes is ",mgrid," using ",BG_mode," mode, \n",\
    "what translates to a processor grid of ",\
    pgrid," and ",topology['per node']," ranks per node => total of ",n_procs," processors."

if len(args) < 1:
    parser.error("At least, you have to input the number of processes in each direction as an array.")

# so we have some argument and some options
n_procs_arr = [int(x) for x in args]
# gene alone
#type_arr    = ['nn','alltoall','nn','collective','collective','collective']
# gene with trinity, last dimension is the number of instances of gene
type_arr    = ['alltoall','nn','collective','collective','collective','collective']

remainder = pernode
new_n_procs_arr = []
for np in n_procs_arr:
    if np<remainder:
        new_n_procs_arr.append(1)
        remainder = remainder/np
    else:
        new_n_procs_arr.append(np/remainder)
        remainder = 1

n_procs_arr=new_n_procs_arr
print "n_procs_arr = ", n_procs_arr

if len(n_procs_arr)!=len(type_arr):
    print "type array and n_procs array must have the same length."
    sys.exit(1)

print "We do the optimization for n_procs = ",n_procs_arr," with the types ", type_arr,"."


########################### start main program ################
# starting from the innermost to the outermost processes and optimize
block = [1, 1, 1]
best_cost = []
for ip in xrange(len(n_procs_arr)):
    if type_arr[ip]=='alltoall':
        best_cost.append(optimize_alltoall(n_procs_arr[ip],block,pgrid))
        #print "after alltoall, best_cost = ", best_cost
    elif type_arr[ip]=='nn':
        best_cost.append(optimize_nearest_neighbor(n_procs_arr[ip],block,pgrid))
        #print "after nearest neighbor, best_cost = ",best_cost
    elif type_arr[ip] == 'collective':
        best_cost.append(optimize_collective(n_procs_arr[ip],block,pgrid))
        #print "after collective, best_cost = ", best_cost
    else:
        print "type can only be 'alltoall', 'nn' or 'collective'!"
        sys.exit(2)
    block = map(lambda x,y:x*y,block,best_cost[-1]['np'])



# Schreibe mapfile fuer BlueGene fuer den billigsten errechneten Parametersatz

print "Writing mapfile for n_procs = ", reduce(lambda x,y:x*y,n_procs_arr), ", pgrid = ", pgrid
print "best_cost = ",best_cost
outfile = open("mapfile.map","wt")
#pgrid is processor grid which has to be filled
write_block(best_cost,[0,0,0],pgrid,outfile,pernode)
outfile.close()

