###########################################################################
### template for architecture dependent GENE makefiles                  ###
###########################################################################
#module unload mvapich
#module load mvapich2
#module load mkl
#module load petsc/3.1-complex
#module load slepc/3.1-complex
#module load scalapack
#module load fftw3_3

###########################################################################
### SWITCHES                                                            ###
###########################################################################

# FFTLIB - needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB = fftw
PRECISION= double
OPENMP = no

SLEPC= yes
SCALAPACK = yes
WITHNAG = no
USE_PERFLIB = none

# FUTILS and HDF5 are required for some geometry interfaces
FUTILS = no

#insert memory per core and uncomment the following line
PREPROC= -D'MB_PER_CORE=750'

###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
INCPATHS = -I$(OBJDIR) -I. -I$(SRCDIR) -I$(TACC_MKL_INC)

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
LIBS = -L$(TACC_MKL_LIB) -lmkl_em64t

#FFT LIBRARY
#specify at least one of the following choices
ifeq ($(FFTLIB),mkl)
 MKLFLAGS +=
 INCPATHS += -I$(TACC_MKL_INC) 
 LIBS += -Wl,-rpath,$(TACC_MKL_LIB) -L$(TACC_MKL_LIB) -lmkl_em64t
endif
ifeq ($(FFTLIB),essl)
 INCPATHS +=
 LIBS +=
 PREPROC += -DWITHESSL
endif
ifeq ($(FFTLIB),fftw)
 INCPATHS += -I$(TACC_FFTW33ALPHA1_INC)
 LIBS += -L$(TACC_FFTW33ALPHA1_LIB) -lfftw3 -lfftw3f
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(WITHNAG),yes)
 INCPATHS +=
 LIBS +=
#DEFAULT: use NAG-FORTRAN 90
 PREPROC +=-DWITHNAGF90
#Alternative: use NAG_FORTRAN-F77
#PREPROC +=-DWITHNAGF77 
endif

ifeq ($(SLEPC),yes)
# PETSC_ARCH = 
# PETSC_DIR = 
# SLEPC_DIR = 
#slepc version < 3.0: include $(SLEPC_DIR)/bmake/slepc_common
#else:
 include $(SLEPC_DIR)/conf/slepc_common
 INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
 LIBS +=$(PETSC_LIB) $(SLEPC_LIB)
 PREPROC += -DWITHSLEPC
endif

#SCALAPACK is only necessary if you want to use the direct eigenvalue solver
ifeq ($(SCALAPACK),yes)
  INCPATHS += -I$(TACC_SCALAPACK_INC) 
  LIBS += -L$(TACC_SCALAPACK_LIB) -lscalapack $(TACC_SCALAPACK_LIB)/blacs_MPI-LINUX-0.a $(TACC_SCALAPACK_LIB)/blacsF77init_MPI-LINUX-0.a 
  PREPROC += -DWITHSCAL
endif

ifeq ($(FUTILS),yes)
 #  FUTILS and HDF5
 FUTILSDIR = $(EXTDIR)/$(MACHINE)/futils/src
 HDF5PATH = 
 HDF5_LIBPATH  = -L$(HDF5PATH)/lib -L$(FUTILSDIR)
 HDF5_LIBS = -lfutils -lhdf5_fortran -lhdf5 -lz 

 LIBS += $(HDF5_LIBPATH) $(HDF5_LIBS)
 PREPROC +=  -DWITHFUTILS
 INCPATHS += -I$(FUTILSDIR)

endif


###########################################################################
### COMPILER & COMPILER FLAGS       				  	###
###########################################################################

#FORTRAN COMPILER
MPFC = mpif90

#ARCHIVE command
#Note: On IBM machines you might need to add -X$(BITMODE) to the "ar" call
ARCHIVE = ar r

#FORTAN COMPILER FLAGS
FFLAGS = 

#PRECOMILER SWITCHES
PREPROC += -DMPI $(SVNDEF)

ifeq ($(DEBUG),yes)
  OPT = -O0
  FFLAGS += -g -C
else
  OPT = -O3 -fastsse -tp barcelona-64
endif
FFLAGS += $(OPT) 

MKLFLAGS = $(OPT) -module $(OBJDIR)

#PGI compiler does not support Fortran 2003 standard yet and requires
#external erf (error function) definition
PREPROC += -DEXTERNAL_ERF -DF2003_NO_NEW_LINE

#LINKER (usually same as MPFC)
MPLD =$(MPFC)

#LDFLAGS
LDFLAGS =

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################
#OPENMP
ifeq ($(OPENMP),yes)
FFLAGS += -mp
LDFLAGS +=-mp
PREPROC +=-DWITHOMP
endif

#Specify compiler flag for conversion to double precision
#(e.g. -r8, -qrealsize=8, etc.)
ifeq ($(PRECISION),double)
FFLAGS += -r8
PREPROC += -DDOUBLE_PREC
PREPROC += -D'erf=derf'
endif

FFLAGS += $(MKLFLAGS)

###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

#------- mpirun
run:	$(RUNDIR)/$(EXEC)
	ulimit -s unlimited;\
	cd $(RUNDIR);\
	OMP_NUM_THREADS=$(OMP_NUM_THREADS)\
	ibrun ./$(EXEC)


###########################################################################
### Machine dependent compiling rules                                   ###
###########################################################################

#>>>> Try this:
ifeq ($(FFTLIB),mkl)
 $(OBJDIR)/par_mod.o:		$(OBJDIR)/mkl_dfti.o
 $(OBJDIR)/fourier_mkl.o:	$(OBJDIR)/mkl_dfti.o
 $(OBJDIR)/mkl_dfti.o: $(TACC_MKL_DIR)/include/mkl_dfti.f90
	$(MPFC) $(MKLFLAGS) -c -o $@ $<
endif

#ifeq ($(FFTLIB),mkl)
# $(OBJDIR)/fourier_mkl.o: $(OBJDIR)/mkl_dfti.mod
# $(OBJDIR)/mkl_dfti.mod: $(TACC_MKL_DIR)/include/mkl_dfti.f90
#	$(MPFC) -c $(TACC_MKL_DIR)/include/mkl_dfti.f90  
#endif
