##############################################
### architecture dependent GENE makefiles, TOK cluster version          ###
##############################################

##############################################
### SWITCHES                                                                                           ###
##############################################

#COMPILER: INTEL or LAHEY
COMPTYPE = INTEL

# FFTLIB - needed in /src/files.mk
# set to: mkl, fftw or essl
FFTLIB = fftw
PRECISION= double
OPENMP = no

SLEPC = no
# has to be switched on
SCALAPACK = yes
DEBUG = no
INCLUDE_SYMBOLS = no
COMPILER_REPORTS=no
USE_PERFLIB = none
FUTILS = yes
PRODRUN = yes
################## some libraries are not yet available for  Lahey #######
ifeq ($(COMPTYPE),LAHEY)
 FFTLIB = fftw
 SLEPC  = no
 FUTILS  = no
 F2003_MISSING = -DF2003_NO_NEW_LINE
endif

#memory per core
PREPROC = -D'MB_PER_CORE=1900'

###########################################################################
#   COMPULSORY LIBRARIES						  #
###########################################################################

MKLPATH = /afs/ipp/common/soft/intel/mkl
MKLLIBPATH = $(MKLPATH)/lib/em64t

#INCLUDE PATHS
#Add include paths for BLAS routines (can be LAPACK,ESSL,MKL,etc.)
INCPATHS = -I$(OBJDIR) -I. -I$(SRCDIR)

#LIBRARIES AND LIBFLAGS
#Insert BLAS library
#LIBS += -L$(MKLLIBPATH)
#LIBS += -Wl,--start-group $(MKLLIBPATH)/libmkl_intel_lp64.a \
#$(MKLLIBPATH)/libmkl_sequential.a $(MKLLIBPATH)/libmkl_core.a \
#-Wl,--end-group -lpthread

### FFT LIBRARY
ifeq ($(FFTLIB),mkl)
 MKLFLAGS += 
 INCPATHS +=
# LIBS += -lmkl_em64t -Bstatic -lguide -Bdynamic -lpthread
endif
ifeq ($(FFTLIB),essl)
 INCPATHS +=
 LIBS +=
 PREPROC += -DWITHESSL
endif
ifeq ($(FFTLIB),fftw)
#FFTW seems to be faster than MKL when single precision is used
#further tests needed
  INCPATHS += -I$(FFTW_HOME)/include
  LIBS += -L$(FFTW_HOME)/lib

  ifeq ($(PRECISION),double)
   LIBS += -lfftw3
  else
   LIBS += -lfftw3f
  endif
endif

###########################################################################
# ADDITIONAL LIBRARIES (set via SWITCH in header)			  #
###########################################################################

ifeq ($(SLEPC),yes)
official_install = yes

ifeq ($(official_install),yes)
#official RZG installation
  ifeq ($(PRECISION),double)
   #PETSC_ARCH = amd64_sles10-intel11-mvapich2-em64t-sdr-mkl-cplx
   PETSC_ARCH = amd64_sles11-intel11.1-impi3.2.2-mkl10.2-em64t-cplx
   PETSC_DIR = $(PETSC_HOME)
   #PETSC_DIR = /afs/ipp/common/soft/petsc/petsc-3.0.0-p2
   SLEPC_DIR = $(SLEPC_HOME)
   include $(SLEPC_DIR)/conf/slepc_common
   INCPATHS += $(SLEPC_INCLUDE) $(PETSC_FC_INCLUDES)
   LIBS += $(SLEPC_LIB)
   PREPROC += -DWITHSLEPC
  else
   SLEPC=no
  endif
else
  ifeq ($(PRECISION),double)
   PETSC_ARCH = linux-gnu-c-opt-dbl
  else
   PETSC_ARCH = linux-gnu-c-opt
  endif
   PETSC_DIR = /gpfs/ipp/tbg/petsc-3.0.0-p9
   SLEPC_DIR = /gpfs/ipp/tbg/slepc-3.0.0-p7
   include $(SLEPC_DIR)/conf/slepc_common
   INCPATHS +=$(PETSC_FC_INCLUDES) $(SLEPC_INCLUDE)
   LIBS += $(SLEPC_LIB)
   PREPROC += -DWITHSLEPC
 endif
endif

ifeq ($(SCALAPACK),yes)
#  MKLLIBPATH = $(MKLPATH)/lib/em64t
#  LIBS += -L$(MKLLIBPATH)
  LIBS += -Wl,--start-group $(MKLLIBPATH)/libmkl_scalapack_lp64.a \
$(MKLLIBPATH)/libmkl_blacs_intelmpi20_lp64.a $(MKLLIBPATH)/libmkl_intel_lp64.a \
$(MKLLIBPATH)/libmkl_sequential.a $(MKLLIBPATH)/libmkl_core.a -Wl,--end-group \
-lpthread 
  PREPROC += -DWITHSCAL
else
#Insert BLAS library
LIBS += -L$(MKLLIBPATH)
LIBS += -Wl,--start-group $(MKLLIBPATH)/libmkl_intel_lp64.a \
$(MKLLIBPATH)/libmkl_sequential.a $(MKLLIBPATH)/libmkl_core.a \
-Wl,--end-group -lpthread
endif

ifeq ($(FUTILS),yes)
 #  FUTILS and HDF5
 FUTILSDIR = $(EXTDIR)/$(MACHINE)/futils/src
 HDF5PATH = $(HDF5_HOME)
 HDF5_LIBPATH  = -L$(HDF5PATH)/lib -L$(FUTILSDIR)
 HDF5_LIBS = -lfutils -lhdf5_fortran -lhdf5 -lz 

 LIBS += $(HDF5_LIBPATH) $(HDF5_LIBS)
 PREPROC +=  -DWITHFUTILS
 INCPATHS += -I$(FUTILSDIR)

endif

ifneq ($(USE_PERFLIB),none)
 PREPROC += -DWITHPERF
endif
ifeq ($(USE_PERFLIB),perf)
 LIBS += -lperf_r
endif

###########################################################################
### COMPILER & COMPILER FLAGS                                           ###
###########################################################################

#PRECOMILER SWITCHES
PREPROC += $(F2003_MISSING) -DLINUX -DMPI $(SVNDEF)
CPREPROC += -DLINUX

ifeq ($(COMPTYPE),LAHEY)
  MPFC = mpilfc -static_mpi
else ifeq($(COMPTYPE),INTEL)
  MPFC = mpiifort
endif
ARCHIVE = ar r

ifeq ($(COMPTYPE),INTEL)
 ifeq ($(DEBUG),yes)
  OPT = -O0
  FFLAGS += -C -traceback -debug inline-debug-info -warn all
#'-check noarg_temp_created' suppresses warnings about temporary arrays during runtime 
  INCLUDE_SYMBOLS=yes
 else
  ifeq ($(PRODRUN),yes)
   OPT= -O3 -ip -xSSSE3
  else
   OPT = -O3 -xSSSE3
  endif
  OPT += -vec_report0
 endif

 ifeq ($(INCLUDE_SYMBOLS),yes)
   FFLAGS += -g
 endif

 ifeq ($(COMPILER_REPORTS),yes)
   FFLAGS += -opt-report -opt-report-phase=all \
	     -opt-report-file=gene_report.opt \
	     -vec-report3
 endif
 MKLFLAGS = -module $(OBJDIR)

 MPLD =$(MPFC)
 LDFLAGS = -lsvml
endif

ifeq ($(COMPTYPE),LAHEY)
 FFLAGS += -M $(OBJDIR)
 ifeq ($(DEBUG),yes)
  FFLAGS += --o0 -g --f95 --info --warn --swm 2004,8202,8209,3126 \
	--trace --chk
 else
  FFLAGS += --o3 --fast -sse3 --prefetch --pca
 endif
 #MKLFLAGS = --f95 --o3 --info --warn --swm 2004,8202,8209,3126 \
 #	--ntrace --pca --mod $(OBJDIR) --nversion
 INCPATHS += -I/afs/@cell/@sys/soft/mvapich/mvapich2-1.2.0-em64t-sdr/lahey-8.10/include
 PREPROC += -DLAHEY
 MPLD      = $(MPFC) #-\#\#\#
 LDFLAGS = 
endif

###########################################################################
# ADDITIONAL COMPILER FLAGS (set via SWITCH in header)			  #
###########################################################################
#OPENMP
ifeq ($(OPENMP),yes)
 ifeq ($(COMPTYPE),INTEL)
  MKLFLAGS += -openmp -fpp
  LDFLAGS += -openmp
 endif
 ifeq ($(COMPTYPE),LAHEY)
  MKLFLAGS += --openmp
  LDFLAGS += --openmp
 endif 
PREPROC +=-DWITHOMP
endif

#Specify compiler flag for conversion to double precision
#(e.g. -r8, -qautodbl, etc.)
ifeq ($(PRECISION),double)
 ifeq ($(COMPTYPE),INTEL)
  FFLAGS += -r8
 endif
 ifeq ($(COMPTYPE),LAHEY)
  FFLAGS += --dbl
 endif 
PREPROC +=-DDOUBLE_PREC
endif

FFLAGS += $(MKLFLAGS) $(OPT)

###########################################################################
### Linking                                                             ###
###########################################################################

include $(FFILES)

#------- mpirun 
run:	$(EXECDIR)/$(EXEC)
	cd $(RUNDIR);\
	LD_LIBRARY_PATH=\
	OMP_NUM_THREADS=$(OMP_NUM_THREADS)\
	MKL_SERIAL=YES\
	mpiexec -n $(N_PES) -env MV2_USE_SRQ 0 ./$(EXEC)

###########################################################################
### Machine dependent compiling rules                                   ###
###########################################################################

ifeq ($(FFTLIB),mkl)
 $(OBJDIR)/par_mod.o:		$(OBJDIR)/mkl_dfti.o
 $(OBJDIR)/fourier_mkl.o:	$(OBJDIR)/mkl_dfti.o
 $(OBJDIR)/mkl_dfti.o: $(MKLPATH)/include/mkl_dfti.f90
	$(MPFC) $(MKLFLAGS) -c -o $@ $<
endif

NOOPTLIST= 

NOOPTFULLOBJ = $(addprefix $(OBJDIR)/,$(NOOPTLIST))

$(NOOPTFULLOBJ): $(OBJDIR)/%.o:$(SRCDIR)/%.F90
	$(MPFC) $(NOOPTFLAGS) $(INCPATHS) $(PREPROC) -c -o $@ $<
