#include "redef.h"
#include "switches.h"
MODULE dealiasing_module
  USE discretization
  USE lagrange_interpolation
  use BoundaryDescriptionModule
  use boundary_exchange_x
  use communications
  use blockindex, only: lbg0
  use par_in, only: rad_bc_type
  use par_mod, only: p_has_0_mode
  IMPLICIT NONE

  PUBLIC :: interpolate_to_fine_grid_x, filter_from_fine_grid_x, filter_from_fine_grid_x_wb
  public :: interpolate_to_fine_grid_x_orig, filter_from_fine_grid_x_orig
  PUBLIC :: initialize_dealiasing, finalize_dealiasing, calculate_dealiasing, initialize_da_bounds
  PUBLIC :: li1da, li2da, li0da, lbida, ubida, lx0da, lbida2, ubida2, lx0da2, ly0da, nibda

  PRIVATE
  
  
  INTERFACE interpolate_to_fine_grid_x
     MODULE PROCEDURE da_interpolate_to_fine_grid_x
  END INTERFACE

  INTERFACE interpolate_to_fine_grid_x_orig
     MODULE PROCEDURE da_interpolate_to_fine_grid_x_orig
  END INTERFACE

  INTERFACE filter_from_fine_grid_x
     module procedure da_filter_from_fine_grid_x
  END INTERFACE

  INTERFACE filter_from_fine_grid_x_orig
     module procedure da_filter_from_fine_grid_x_orig
  END INTERFACE

  INTERFACE filter_from_fine_grid_x_wb
     module procedure da_filter_from_fine_grid_x_wb
  END INTERFACE

  TYPE(BoundaryDescription) :: lx0da2_boundary, lx0da_boundary, lx0da_boundary_block,lx0da2_boundary_block
  LOGICAL :: isInitialized_module=.FALSE.
  integer(C_INT),bind(c):: li1da, li2da, li0da, lbida, ubida, lx0da, lbida2, ubida2, lx0da2, ly0da
  integer:: nibda
  complex :: complex_dummy
  integer,parameter :: complex_kind=kind(complex_dummy)
CONTAINS
  
  subroutine initialize_da_bounds
    ! boundaries for dealiasing
    If (xy_local) then
       if ((.not.nonlinear).or.turbdeal) then
          ! no additional kx modes for dealiasing
          li0da= li0
          li1da= li1
          li2da= li1da+li0da-1
          kx_offset= 0
          ly0da= 2*nky0
       else
          ! boundaries for dealiasing with 3/2-rule
          li0da= 3*(nx0+1-evenx)/2
          li1da= li1
          li2da= li1da+li0da-1
          kx_offset=nx0/2 +2*(1-evenx)  !number of modes added in kx direction for dealiasing
          ly0da= 3*nky0
       endif
    Else
       ! Always y dealiasing, the dealising switch only switches the x real space dealiasing
       if (turbdeal) then
          ly0da= 2*nj0
       else
          ly0da= 3*nj0
       endif

       if (dealiasing) then
          ! x direction
          li0da  = 2*li0
          li1da  = li1
          li2da  = li1da+li0da-1
          kx_offset = li0
          nibda =  (lag_order+1)/2
          lbida = li1-nibda+1
          ubida = li2+nibda
          lx0da = ubida-lbida+1
          lbida2 = li1da-2*nibda+1
          ubida2 = li2da+2*nibda-1
          lx0da2 = ubida2-lbida2+1
       else
          li0da  = li0;  li1da  = li1
          li2da  = li1+li0da-1
          kx_offset = 0;    nibda =  0
          !lbida = li1;   ubida = li2
          lbida = lbi;   ubida = ubi
          lx0da = li0;   lbida2 = li1 
          ubida2 = li2;  lx0da2 = li0
       endif
       !write(*,"(3(A,I4),I4,2(A,I4),I4,A,I4)") "li0=",li0,", li0da=",li0da,", li1da:li2da=",li1da,li2da,&
       !     &", nibda = ",nibda,", lbida2:ubida2=",lbida2,ubida2,", lx0da2=",lx0da2
    Endif
  end subroutine initialize_da_bounds

  SUBROUTINE initialize_dealiasing

    IF (.NOT.isInitialized_module) THEN
       CALL initialize_lag_coef(li1,li2,&
            li1da,li2da,lbida,ubida,&
            lbida2,ubida2,lag_order)

       CALL initialize_type(lx0da_boundary,  1,lx0da,  nibda-1,   nibda,     lj0)
       CALL initialize_type(lx0da_boundary_block,  1,lx0da,  nibda-1,   nibda,     lj0*lbg0)
       CALL initialize_type(lx0da2_boundary, 1,lx0da2, 2*nibda-1, 2*nibda-1, lj0)
       CALL initialize_type(lx0da2_boundary_block, 1,lx0da2, 2*nibda-1, 2*nibda-1, lj0*lbg0)
       
       IF (n_procs_x > 1) THEN
          CALL set_mpi_type(lx0da_boundary)
          CALL set_mpi_type(lx0da_boundary_block)
          CALL set_mpi_type(lx0da2_boundary)
          CALL set_mpi_type(lx0da2_boundary_block)
       ENDIF
       isInitialized_module = .TRUE.
    END IF

  END SUBROUTINE initialize_dealiasing

  SUBROUTINE finalize_dealiasing
    CALL finalize_lag_coef
    call finalize_type(lx0da_boundary)
    call finalize_type(lx0da_boundary_block)
    CALL finalize_type(lx0da2_boundary)
    CALL finalize_type(lx0da2_boundary_block)
    isInitialized_module = .FALSE.
  END SUBROUTINE finalize_dealiasing

  SUBROUTINE da_interpolate_to_fine_grid_x(fin,fout)
    !COMPLEX, DIMENSION(lbida:ubida,lj1:lj2), INTENT(IN) :: fin
    COMPLEX, DIMENSION(li1:li2,lj1:lj2,1:lbg0), INTENT(IN) :: fin
    COMPLEX, DIMENSION(0:li0da-1,lj1:lj2,1:lbg0), INTENT(OUT) :: fout

    ! Local variables
    COMPLEX, dimension(lbida:ubida,lj1:lj2,1:lbg0)::  temp1

    ! copy the inner points, but now we have larger boundaries
    temp1(li1:li2,lj1:lj2,1:lbg0) = fin(li1:li2,lj1:lj2,1:lbg0)

    ! fill the new boundaries by exchange
    CALL exchange_x(lx0da_boundary_block,temp1)

    ! interpolate to finer grid
    CALL lag_interp(temp1,fout,lj0*lbg0)

  END SUBROUTINE da_interpolate_to_fine_grid_x

  SUBROUTINE da_interpolate_to_fine_grid_x_orig(fin,fout)
    COMPLEX, DIMENSION(li1:li2,lj1:lj2), INTENT(IN) :: fin
    COMPLEX, DIMENSION(0:li0da-1,lj1:lj2), INTENT(OUT) :: fout
!    INTEGER :: i

    ! Local variables
    COMPLEX::  temp1(lbida:ubida,lj1:lj2)

    ! copy the inner points, but now we have larger boundaries
    temp1(li1:li2,lj1:lj2) = fin(li1:li2,lj1:lj2)

    ! fill the new boundaries by exchange
    CALL exchange_x(lx0da_boundary,temp1)
    !If ((rad_bc_type.eq.2).and.(p_has_0_mode).and.(my_pex.eq.0)) then
    !   Do i=1,nibda-1
    !      temp1(li1-i,lj1) = temp1(li1+i,lj1)
    !   Enddo
    !Endif

    ! interpolate to finer grid
    CALL lag_interp(temp1,fout,lj0)

  END SUBROUTINE da_interpolate_to_fine_grid_x_orig


  SUBROUTINE da_filter_from_fine_grid_x(fin,fout)
    COMPLEX,intent(INOUT) :: fin(lbida2:ubida2,lj1:lj2,1:lbg0)
    COMPLEX,INTENT(OUT) :: fout(li1:li2,lj1:lj2,1:lbg0)
!    INTEGER :: i,klmn

    !do klmn=1,lbg0
    !   write(*,"(2I3,2(A,I2),A,ES17.10)") mype,klmn," : vor exchange,  fin(",li1da,":",li2da,") = ",&
    !        &real(sum(conjg(fin(li1da:li2da,lj1:lj2,klmn))*fin(li1da:li2da,lj1:lj2,klmn)))
    !end do
    CALL exchange_x(lx0da2_boundary_block,fin)
    !do klmn=1,lbg0
    !   write(*,"(2I3,A,ES17.10)") mype,klmn," : nach exchange, fin = ",&
    !        &real(sum(conjg(fin(li1da:li2da,lj1:lj2,klmn))*fin(li1da:li2da,lj1:lj2,klmn)))
    !end do
    CALL lag_filtering(fin,fout,lj0*lbg0,0)

    !do klmn=1,lbg0
    !   write(*,"(2I3,2(A,I2),A,ES17.10)") mype,klmn," : lag_filtering,  fout(",li1,":",li2,") = ",&
    !        &real(sum(conjg(fout(li1:li2,lj1:lj2,klmn))*fout(li1:li2,lj1:lj2,klmn)))
    !end do

  END SUBROUTINE da_filter_from_fine_grid_x

#ifdef WITH_C_NONLIN
  ! both input parameters are given as type(c_ptr). Then
  ! they need the value attribute and can only be intent(IN).
  ! The values of cin and cout as pointers are not changed in the
  ! routine, only the values in the memory where they point to.
  SUBROUTINE da_interpolate_to_fine_grid_x_from_c(cin,cout) bind(c)
    !COMPLEX, DIMENSION(lbida:ubida,lj1:lj2), INTENT(IN) :: fin
    type(C_PTR), value, INTENT(IN) :: cin
    type(C_PTR), value, INTENT(IN) :: cout

    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:,:), pointer :: fin, fout

    call C_F_POINTER(cin,fin,(/li0,lj0,lbg0/))
    call C_F_POINTER(cout,fout,(/li0da,lj0,lbg0/))

    call interpolate_to_fine_grid_x(fin,fout)
  END SUBROUTINE da_interpolate_to_fine_grid_x_from_c

  SUBROUTINE da_filter_from_fine_grid_x_from_c(cin,cout) bind(c)
    type(C_PTR), value, intent(IN) :: cin
    type(C_PTR), value, INTENT(IN) :: cout

    complex(C_DOUBLE_COMPLEX), dimension(:,:,:), pointer :: fin, fout

    call C_F_POINTER(cin,fin,(/lx0da2,lj0,lbg0/))
    call C_F_POINTER(cout,fout,(/li0,lj0,lbg0/))

    call filter_from_fine_grid_x(fin,fout)
  END SUBROUTINE da_filter_from_fine_grid_x_from_c
#endif

  SUBROUTINE da_filter_from_fine_grid_x_orig(fin,fout)
    COMPLEX,intent(INOUT) :: fin(lbida2:ubida2,lj1:lj2)
    COMPLEX,INTENT(OUT) :: fout(li1:li2,lj1:lj2)
    INTEGER :: i

    CALL exchange_x(lx0da2_boundary,fin)
    If ((rad_bc_type.eq.2).and.(p_has_0_mode).and.(my_pex.eq.0)) then
       Do i=1,2*nibda-1
          fin(li1-i,lj1) = fin(li1+i,lj1)
       Enddo
    Endif
    CALL lag_filtering(fin,fout,lj0,0)

  END SUBROUTINE da_filter_from_fine_grid_x_orig

  SUBROUTINE da_filter_from_fine_grid_x_wb(fin,fout)
    COMPLEX,intent(INOUT) :: fin(lbida2:ubida2,lj1:lj2)
    COMPLEX,INTENT(OUT) :: fout(li1:li2,lj1:lj2)
    INTEGER :: i

    CALL exchange_x(lx0da2_boundary,fin)
    If ((rad_bc_type.eq.2).and.(p_has_0_mode).and.(my_pex.eq.0)) then
       Do i=1,2*nibda-1
          fin(li1-i,lj1) = fin(li1+i,lj1)
       Enddo
    Endif
    CALL lag_filtering(fin,fout,lj0,0) !nib)

  END SUBROUTINE da_filter_from_fine_grid_x_wb

  SUBROUTINE calculate_dealiasing(u,phase_fac)
    COMPLEX, DIMENSION(li1:li2,lj1:lj2),intent(INOUT) :: u
    COMPLEX, DIMENSION(li1da:li2da,lj1:lj2), intent(IN) :: phase_fac

    ! Local variables
    COMPLEX:: temp3(0:li0da-1,lj1:lj2), temp2(lbida2:ubida2,lj1:lj2)
    !COMPLEX :: temp1(lbida:ubida,lj1:lj2)

    !temp1(li1:li2,lj1:lj2) = u(li1:li2,lj1:lj2)
    CALL interpolate_to_fine_grid_x_orig(u, temp3)
    temp2(li1da:li2da,lj1:lj2) = phase_fac*temp3

    CALL filter_from_fine_grid_x_wb(temp2, u)
  END SUBROUTINE calculate_dealiasing
END MODULE dealiasing_module
