#include "redef.h"
!>Module containing subroutines to reset the distribution function
!!and all profile information throughout the whole GENE code
Module reset_mod
  Use par_mod
  use par_other
!  Use parameters_IO, only: write_parameters
  Use spectype_mod
  Use vel_space
  use prefactors
  use geometry, only: geom, rhostar,minor_r, write_geometry
  use fieldsolver_df
  use calc_rhs
  use diagnostics
  use diagnostics_df
  use coordinates, only: calc_center_value
  use profiles_mod, only: write_tempdens_profiles
  use f0_term, only: f0_contr
  use aux_fields, only: initialize_calc_aux_fields, &
       &finalize_calc_aux_fields
  use lagrange_interpolation, only: lag3deriv

  Implicit None

  !The basic reset interface
  PUBLIC :: reset_g1_prof

  PRIVATE
  
  integer :: reset_count = 0

Contains

  Subroutine reset_g1_prof(g_1,new_profs)
    complex,dimension(li1:li2,lj1:lj2,lk1:lk2,ll1:ll2,lm1:lm2,ln1:ln2),intent(inout):: g_1
    real,dimension(0:nx0-1,0:n_spec-1,1:4),intent(inout) :: new_profs
    real,dimension(0:nx0-1,0:n_spec-1,1:4) :: smooth_new_profs
    integer :: j,k,l,m,n,dummy_int
    Real, Dimension(pi1:pi2) :: fm_new
    Logical :: dummy1, dummy2,dummy3
!    Real :: T_center, n_center
    integer:: interval
    real, dimension(pi1gl:pi2gl):: dens_electrons, omn_electrons

    !for now: reset equilibrium densities for all species to the 
    !electron density profile (times spec(n)%dens)
    do n=0,n_spec-1
       if (spec(n)%charge==-1) then
          dens_electrons=new_profs(:,n,1)
          omn_electrons=new_profs(:,n,3)
       endif
    enddo

    do n=0,n_spec-1
       if (n_spec.gt.1) then
          new_profs(:,n,1)=dens_electrons
          new_profs(:,n,3)=omn_electrons
       endif
       if (smooth_reset) then
          interval=max(nint(5./deli),9)
          call smooth_profiles(new_profs(:,n,:), smooth_new_profs(:,n,:),interval)
       else
          smooth_new_profs(:,n,:) = new_profs(:,n,:)
       endif
       !normalize profile to center value
       !call calc_center_value(smooth_new_profs(:,n,1),n_center)
       spec(n)%dens_prof=smooth_new_profs(:,n,1) !/n_center
       !call calc_center_value(smooth_new_profs(:,n,2),T_center)
       spec(n)%temp_prof=smooth_new_profs(:,n,2) !/T_center
       spec(n)%omn_prof=smooth_new_profs(:,n,3)
       spec(n)%omt_prof=smooth_new_profs(:,n,4)
    enddo

    !for now: g_1^new = g_1^old - Delta f_0 = (f_1^old-(f_0^old-f_0^new)) + q/c Aparbar^old vpar/T0 f_0^old
    !however, we might need to fix the second term as well in the future
    do n=ln1,ln2
       do m=lm1,lm2
          do l=ll1,ll2
             do k=lk1,lk2
                fm_new = spec(n)%dens_prof(pi1:pi2) *&
                     (pi*spec(n)%temp_prof(pi1:pi2))**(-1.5) *&
                     & exp(-(mu(m)*geom%Bfield(pi1:pi2,pj1,k)+vp(l)**2)/&
                     spec(n)%temp_prof(pi1:pi2))
!                do j=lj1,lj2
                if (p_has_0_mode) then
                   j=lj1
                   g_1(:,j,k,l,m,n) = g_1(:,j,k,l,m,n)+(fm(pi1:pi2,pj1,k,l,m,n)&
                        &-fm_new(pi1:pi2))/(rhostar*minor_r)
                endif
                fm(pi1:pi2,pj1,k,l,m,n)=fm_new(pi1:pi2)
!                enddo
             enddo
          enddo
       enddo
    enddo

    call finalize_all_diags
    call finalize_CalFullRhs
    call finalize_calc_aux_fields

    if (mype==0) print*, 'old file_extension= ',file_extension
    read (file_extension(2:),'(I19)') dummy_int
    dummy_int = dummy_int+1
    write(file_extension,"(A,I4.4)") '_',dummy_int
    if (mype==0) print*, 'new file_extension= ', file_extension
    itime = 0
    
    call set_prefactors
    call initialize_calc_aux_fields
    call initialize_CalFullRhs
    call initialize_all_diags

    !adapt pressure gradient? (would require additional logs whether
    !pressure gradient originated from MHD file or from gradients)

    if (mype==0) then
       print*, 'Resetted g_1 and all n and T dep. profiles'
       do n=0,n_spec-1
          Call write_tempdens_profiles(spec(n),.false.)
       enddo
    endif

    !write parameters & geometry
!    If (mype == 0) Call write_parameters !Introduces circular reference  :-(
    if(write_geom) call write_geometry

    !write current time step
    call exec_all_diags(0,time,dummy1,dummy2,dummy3)

  End Subroutine reset_g1_prof

  subroutine smooth_profiles(rawprof, smoothprof, lim)
    integer, intent(inout) :: lim ! Number of points used for filtering 2*lim+1
    real,dimension(0:nx0-1,1:4), intent(in) :: rawprof
    real,dimension(0:nx0-1,1:4), intent(inout) :: smoothprof
    real,dimension(-lim:nx0-1+lim) :: exp_rawprof !used for the boundary points
    real,dimension(-lim:lim) :: srt_tmp, weight ! used for selection algorithm, weights for filters
    real,dimension(0:nx0-1):: temparr
    integer :: minIndex
    real :: minValue, val_newcenter, val_oldcenter
    integer :: i,o,p,q
    
    lim=ceiling(real((lim-1))/2)

    select case (smooth_reset_type)
    !smooth profiles by taking a running median of the gradients
    case('median')
       !smooth the gradient profiles
       do q=3,4
          exp_rawprof(0:nx0-1) = rawprof(:,q)
          do i=1,lim !mirror boundary points
             exp_rawprof(-i) = rawprof(i,q)
             exp_rawprof(nx0-1+i) = rawprof(nx0-1-i,q)
          enddo

          ! These loops replace the value of a profile point with the median of the
          ! 2*lim+1 points around it (including the point itself)
          do i=0,nx0-1
             srt_tmp = exp_rawprof(i-lim:i+lim)
             do o=-lim,0
                minIndex=o
                minValue=srt_tmp(o)
                do p=o+1,lim
                   if (srt_tmp(p) < minValue) then
                      minIndex = p
                      minValue = srt_tmp(p)
                   endif
                enddo
                srt_tmp(minIndex) = srt_tmp(o)
                srt_tmp(o) = minValue
             enddo
             smoothprof(i,q) = srt_tmp(0)
          enddo
       enddo

      ! integrate the logarithmic gradients (trapezoidal rule) to generate n(x) and T(x)
       do q=1,2
          smoothprof(:,q) = 0.0
          do i=0,nx0-2
             do o=0,i
                smoothprof(i,q)= smoothprof(i,q) + (smoothprof(o+1,q+2)+smoothprof(o,q+2))/2/nx0*lx*rhostar*minor_r
             enddo
          enddo
          !outward boundary point can be set to second last point (the boundary has
          !enough distortions already)
          smoothprof(nx0-1,q) = smoothprof(nx0-2,q)
          smoothprof(:,q) = exp(-smoothprof(:,q))
          call calc_center_value(rawprof(:,q),val_oldcenter)
          call calc_center_value(smoothprof(:,q),val_newcenter)
          smoothprof(:,q) = smoothprof(:,q)*val_oldcenter/val_newcenter
       enddo
    !smooth profiles by taking a running average of the profiles
    case('average')
       !boundary conditions: zero gradient
       do q=1,2
          exp_rawprof(0:nx0-1)=rawprof(:,q)
          do i=1,lim
             exp_rawprof(-i)=rawprof(i,q)
             exp_rawprof(nx0-1+i)=rawprof(nx0-1-i,q)
          enddo
          do i=0,nx0-1
             smoothprof(i,q)=sum(exp_rawprof(i-lim:i+lim))/(2*lim+1)
          enddo
          temparr = -log(smoothprof(:,q))
          call lag3deriv(temparr, xval, nx0,smoothprof(:,q+2), xval, nx0)
          smoothprof(:,q+2) = smoothprof(:,q+2)/(rhostar*minor_r)
       enddo
    !smooth profiles with a gaussian filter
    case('gauss')
       do q=1,2
          exp_rawprof(0:nx0-1)=rawprof(:,q)
          do i=1,lim
             exp_rawprof(-i)=rawprof(i,q)
             exp_rawprof(nx0-1+i)=rawprof(nx0-1-i,q)
          enddo
          do o=-lim,lim
             weight(o) = exp(-(real(o)/real(lim))**2/0.3)
          enddo
          weight(-lim:lim) = weight(-lim:lim)/sum(weight(-lim:lim))
          !if(mype == 0) print*, weight(-lim:lim)
          smoothprof(:,q) = 0.0
          do i=0,nx0-1
             do o=-lim,lim
                smoothprof(i,q)= smoothprof(i,q) + (exp_rawprof(i+o))*weight(o)
             enddo
          enddo
          temparr = -log(smoothprof(:,q))
          call lag3deriv(temparr, xval, nx0,smoothprof(:,q+2), xval, nx0)
          smoothprof(:,q+2) = smoothprof(:,q+2)/(rhostar*minor_r)
       enddo
    case default
       stop 'smooth_reset_type specified incorrectly!'
    end select

  end subroutine smooth_profiles
     
End Module reset_mod
