#!/usr/bin/env python
### Python script for plotting of profile data, using the classes ProfileData and PlotNeodata
import sys
import os
import copy
import argparse
#from matplotlib.ticker import MaxNLocator, FormatStrFormatter

import ProfileData
import PlotNeodata as Pnd
# This set controls what is plotted. Without the aprropriate element here, the plotting routines return nothing.
plotthis = [
    'plotT',
    'plotchi',
    'plotQ',
    'plotch',
    'plotlocQ',
#    'plotP',
    'plotj',
    'plotG',
#    'plotturb',
    'plotnu',
    'plotfsa',
    ]
plotthis = set(plotthis)


parser = argparse.ArgumentParser()
parser.add_argument("time", help="Time window for averaging in the form start:end. " 
        "Use f for the first and l for the last time in the profile file.")
parser.add_argument("fileextension",nargs='+', help= "List of file extensions to be processed.")
parser.add_argument("--turb", "-t", action='store_true', help="Include turbulent transport in plots")
parser.add_argument("--contin", "-c", action='store_true', help="Try to combine the given files into one averaged profile"
                    "This requires matching parameter sets and corresponding times at the end and beginning of the profiles")
parser.add_argument("--makepdf", "-p", action='store_true', help="Generate pdf files of the plots")
args = parser.parse_args()

fileextension = args.fileextension

if args.turb:
    plotthis.add('plotturb')

if args.time.split(':')[0] == 'f':
    starttime = -1
elif args.time.split(':')[0] == 'l':
    starttime = -2 
else:
    starttime = int(float(args.time.split(':')[0]))
if args.time.split(':')[1] == 'f':
    endtime = -1
elif args.time.split(':')[1] == 'l':
    endtime = -2 
else:
    endtime = int(float(args.time.split(':')[1]))

inputdata = []
#plotarr = []

for fext in fileextension:
    prdat = ProfileData.ProfileData(fext,starttime,endtime,plotthis)
    prdat.readprofiles()
    inputdata.append(prdat)
if args.contin:
    for i in range(1,len(inputdata)):
        glue = inputdata[0].gluecontindata(inputdata[i])
        if glue: 
            del(inputdata[i])

#inputdata[0].plotset.remove('plotch')

plot = Pnd.PlotNeodata(inputdata)
plot.plot_T(args.makepdf)
plot.plot_Q(args.makepdf)
plot.plot_chi(args.makepdf)
plot.plot_jbs(args.makepdf)
plot.plot_nustar(args.makepdf)
plot.plot_fsa(args.makepdf)

plot.show()
