import sys
import os
import math
import copy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.font_manager import fontManager, FontProperties
#from matplotlib.ticker import MaxNLocator, FormatStrFormatter
import matplotlib as mpl
mpl.rcParams['lines.linewidth'] = 1.5
mpl.rc('xtick', labelsize=18)
mpl.rc('ytick', labelsize=18)
from ParIO import Parameters
import ProfileData
import read_mom_field.fieldlib as flib
from read_mom_field.averages import *
from read_mom_field.common import *
from read_mom_field.geom import *


class PlotNeodata(object):
    """ Plot the data from list of profdata objects"""
    def __init__(self, profdata):
        self.leg1=[]
        self.leg2=[]
        self.leg3=[]
        self.leg4=[]
        self.leg5=[]
        self.leg6=[]
        self.leg7=[]
        self.prdlist = profdata
        for prd in self.prdlist:
            if (prd.isDataPresent == False):
                print "No profile data present in object ", prd.fext
                prd.plotset.clear()
        self.font = FontProperties(size=16)
        self.xyfs = 20
        self.xmin = min( [prd.xs[1] + prd.lbuffer for prd in self.prdlist]) 
        self.xmax = max( [prd.xs[-1] - prd.ubuffer for prd in self.prdlist])
   
    def _get_erad(self,prd):
        ##TODO: This is redundant and ugly...
        par=Parameters()
        par.Read_Pars('parameters%s' %(prd.fext))
        pars=par.pardict
        phifile = flib.fieldfile("field"+prd.fext,pars)
        phifile.get_minmaxtime()
        times = []
        delta_t = int(pars['istep_field'])*float(pars['dt_max'])
        times.extend([time for time in phifile.tfld if time>prd.starttime-delta_t and time<prd.endtime+delta_t])
        common=common_data(pars)
        common.tlen=len(times)
        common.times=np.array(times,dtype=np.float64)
        common.fileext=prd.fext
        common.rundir=""
        geom=geometry(common,pars)
        avg=averages(common,geom)
        phi_ky0 = np.zeros((len(times),prd.nx0))
        for time in times:
            phifile.set_time(time)
            phi_ky0[times.index(time),:] = avg.z(phifile.phi())[0,:]
        phi_ky0_av = avg.t(phi_ky0)
        erad = np.zeros(phi_ky0_av.shape,np.float)
        erad[0:-1] = np.diff(phi_ky0_av)/np.diff(prd.xs)
        erad[-1] = (phi_ky0_av[-1] - phi_ky0_av[-2])/(prd.xs[-1] - prd.xs[-2])
        erad = -erad*prd.rhostar
        return erad

    def _calc_fb_param(self,prd, n, q, epsilon):
        erad = self._get_erad(prd)
        fb_param = -(prd.jbs[:,n] /(q/epsilon*np.sqrt(1-epsilon**2)* prd.ns[:,n]* prd.Ts[:,n]) - (prd.omns[:,n] + prd.omts[:,n] + erad/prd.Ts[:,n]))/prd.omts[:,n]         
        return fb_param

    def plot_T(self, poutput):
        fig1 = plt.figure(1)
        ax1 = fig1.add_subplot(111)
        for prd in self.prdlist:
            if 'plotT' not in prd.plotset:
                continue
            ax1.plot(prd.xs,prd.Ts[:,0],'c')
            ax1.plot(prd.xs,prd.ns[:,0],'k')
            ax1.plot(prd.xs,prd.omts[:,0]*prd.major_R,'r')
            ax1.plot(prd.xs,prd.omns[:,0]*prd.major_R,'b')
            ax1.plot(prd.xs,prd.T0s[:,0],'c--')
            ax1.plot(prd.xs,prd.omt0s[:,0]*prd.major_R, 'r--')
            ax1.plot(prd.xs,prd.n0s[:,0],'k--')
            ax1.plot(prd.xs,prd.omn0s[:,0]*prd.major_R,'b--')
            self.leg1.append(r'$T$, %s')
            #self.leg1.append(r'$T_0$')
            self.leg1.append(r'$n$')
            #self.leg1.append(r'$n_0$')
            self.leg1.append(r'$R/L_T$')
            #self.leg1.append(r'$R/L_{T0}$')
            self.leg1.append(r'$R/L_n$')
            #self.leg1.append(r'$R/L_{n0}$')
        ax1.legend(self.leg1,loc=1,prop=self.font)
        ax1.set_xlabel(r'$x/a$',fontsize=self.xyfs)
        ax1.set_xlim(self.xmin,self.xmax)
        ax1.set_ylabel(r'$T, n$ / $T_{ref}$, $n_{ref}$',fontsize=self.xyfs)
        if (poutput == 1):
            fig1.frameon = False
            fig1.savefig('profiles_Ions%s.pdf'%(self.prdlist[0].fext))


    def plot_Q(self,poutput):
        fig2 = plt.figure(2)
        ax2 = fig2.add_subplot(111)
        labely =""
        for prd in self.prdlist:
            if 'plotQ' not in prd.plotset and 'plotG' not in prd.plotset and 'plotP' not in prd.plotset:
                continue
            for n in range(prd.nspec):
                if 'plotQ' in prd.plotset:
                    ax2.plot(prd.xs,prd.Qs[:,n],'b')
                    self.leg2.append(r"$Q_{nc}$ %s" %(prd.specname[n]))
                    if 'plotturb' in prd.plotset:
                        ax2.plot(prd.xs,prd.Qturb[:,n],'m') 
                        self.leg2.append(r"$Q_{turb}$ %s"%(prd.specname[n]))
                    if 'plotch' in prd.plotset:
                        ax2.plot(prd.xs,prd.Qcharr[:,n],'r')
                        self.leg2.append(r"Chang-Hinton")
                    if 'plotlocQ' in prd.plotset:
                        ax2.plot(prd.locxpos,prd.locqval,'+')
                        self.leg2.append(r"$Q_{nc}$ local")
                    if labely.find('$Q/Q_{gB}$') == -1: labely += '$Q/Q_{gB}$'
                if 'plotG' in prd.plotset:
                    ax2.plot(prd.xs,prd.Gammas[:,n], 'g')
                    self.leg2.append(r"$\Gamma_{nc}$ %s" %(prd.specname[n]))
                    if 'plotturb' in prd.plotset:
                        ax2.plot(prd.xs,prd.Gammaturb[:,n]) 
                        self.leg2.append(r"$\Gamma_{turb} $%s"%(prd.specname[n]))
                    if labely.find('$\Gamma/\Gamma_{gB}$') == -1: labely += '  $\Gamma/\Gamma_{gB}$'
                if 'plotP' in prd.plotset:
                    ax2.plot(prd.xs,prd.Pis[:,n])
                    self.leg2.append(r"$\Pi_{nc}$ %s" %(prd.specname[n]))
                    if 'plotturb' in prd.plotset:
                        ax2.plot(self.prd.xs,self.prd.Piturb[:,n],'c') 
                        self.leg2.append(r"$\Pi_{turb} $%s"%(prd.specname[n]))
                    if labely.find('$\Pi/\Pi_{gB}$') == -1: labely += '  $\Pi/\Pi_{gB}$'
        ax2.legend(self.leg2,loc=1,prop=self.font)
        ax2.set_xlabel(r'$x/a$',fontsize=self.xyfs)
        ax2.set_xlim(self.xmin,self.xmax)
        ax2.set_ylabel(labely,fontsize=self.xyfs)
        if (poutput==1):
            fig2.frameon = False
            fig2.savefig("radfluxes%s.pdf"%(self.prdlist[0].fext))

    def plot_chi(self,poutput):
        fig3 = plt.figure(3)
        ax3 = fig3.add_subplot(111)
        for prd in self.prdlist:
            if 'plotchi' not in prd.plotset:
                continue
            for n in range(prd.nspec):
                #rln_x0= prd.omn0s[prd.x0ind,n]
                ax3.plot(prd.xs,prd.Qs[:,n]/prd.Ts[:,n]/prd.ns[:,n]**2/prd.omts[:,n]/prd.omns[:,n])
                self.leg3.append(r"$\chi_{nc}$  %s" %(prd.specname[n]))
                # if (plotlocQ == 1):
                #    plt.plot(locxpos,locchi,'+')
                #    leg3.append(r"Q nc local")
                if 'plotturb' in prd.plotset:
                    ax3.plot(prd.xs,prd.Qturb[:,n]/prd.Ts[:,n]/prd.ns[:,n]**2/prd.omts[:,n]/prd.omns[:,n]) 
                    self.leg3.append(r"$\chi_{turb} $ %s"%(prd.specname[n]))
                if 'plotch' in prd.plotset:
                    ax3.plot(prd.xs,prd.Qcharr[:,n]/prd.Ts[:,n]/prd.ns[:,n]**2/prd.omts[:,n]/prd.omns[:,n],'r')
                #    self.leg3.append(r"$\chi_{nc} $ %s, $\Phi = 0$"%(self.prd.specname[n]))
                    self.leg3.append(r"Chang-Hinton")
        ax3.legend(self.leg3,loc=1,prop=self.font)
        ax3.set_xlim(self.xmin,self.xmax)
        #todo: not nice
        #ax3.set_ylim(0,2.0)
        ax3.set_xlabel(r'$x/a$',fontsize=self.xyfs)
        ax3.set_ylabel(r'$\chi [c_s(x_0) \rho_s^{2}(x_0)/L_n]$',fontsize=self.xyfs)
        if (poutput == 1):
            fig3.frameon = False
            fig3.savefig("raddiffusivity%s.pdf"%(self.prdlist[0].fext))
            
    def plot_jbs(self,poutput):
        fig4 = plt.figure(4)
        ax4_1 = fig4.add_subplot(121)
        ax4_2 = fig4.add_subplot(122)
        for prd in self.prdlist:
            if 'plotj' not in prd.plotset:
                continue
            epsilon = np.array(prd.xs*prd.minor_r/prd.major_R)
            q = np.array(map(prd.q_fun, prd.xs))
            for n in range(prd.nspec):
                ax4_1.plot(prd.xs,prd.jbs[:,n])
                self.leg4.append(r"$j_{b}$ %s"%(prd.specname[n]))
                #ax4.plot(self.prd.xs,np.array(map(self.prd.q_fun,self.prd.xs))/epsilon*self.prd.omns[:,n]*self.prd.ns[:,n])
                #ax4_1.plot(prd.xs, np.array(q*q/epsilon*np.sqrt(1-epsilon**2)*prd.omns[:,n]*prd.ns[:,n]**2 *prd.Ts[:,n] *( 1.46*np.sqrt(epsilon) - 1.35 * np.sqrt(prd.nustar[:,n])/epsilon**0.25)))
                #self.leg4.append(r"Prediction")
                if 'plotlocQ' in prd.plotset:
                     ax4_1.plot(prd.locxpos,np.array(prd.locjbval),'+')
                     self.leg4.append(r"$j_{bs}$ local")
                fb_param = self._calc_fb_param(prd,n,q, epsilon)
                ax4_2.plot(prd.xs,fb_param)
                ax4_1.plot(prd.xs, np.array(q/epsilon*np.sqrt(1-epsilon**2)*prd.ns[:,n]*prd.Ts[:,n]*(prd.omns[:,n] +  prd.omts[:,n] - fb_param*prd.omts[:,n]))) #still lacks erad!!!
        ax4_1.legend(self.leg4,loc=1,prop=self.font)
        ax4_1.set_xlabel(r'$x/a$',fontsize=self.xyfs)
        ax4_2.set_xlabel(r'$x/a$',fontsize=self.xyfs)
        ax4_1.set_xlim(self.xmin,self.xmax)
        ax4_2.set_xlim(self.xmin,self.xmax)
        ax4_1.set_ylabel(r'$j_B$',fontsize=self.xyfs)
        if poutput == 1:
            fig4.frameon = False
            fig4.savefig("j_botstrap%s.pdf"%(self.prdlist[0].fext))

    def plot_nustar(self,poutput):
        fig5 = plt.figure(5)
        ax5 = fig5.add_subplot(111)
        mspec = 1 #TODO: include species mass in prd!!!!
        for prd in self.prdlist:
            if 'plotnu' not in prd.plotset:
                continue
            eps =  (prd.xs*prd.minor_r/prd.major_R)
            for n in range(prd.nspec): 
                ax5.plot(prd.xs, prd.nustar[:,n]* eps**1.5)
                self.leg5.append(r"$\nu* \cdot \epsilon^{3/2}$ ")
            ax5.plot(prd.xs, eps**1.5)
            self.leg5.append(r"$\epsilon^{3/2}$")
            q = np.array(map(prd.q_fun, prd.xs))
            colltime = (prd.nustar[:,n]/q/prd.major_R*np.sqrt(2*prd.Ts[:,n]/mspec)*eps**1.5)**(-1)
            #This is tau_ii as defined in Helander/Sigmar, note tau_i = sqrt(2) * tau_ii
            ax5.text(0.1,0.5,"Highest collision time: %f"%np.max(colltime))
            ax5.text(0.1,0.3,"Lowest collision time: %f"%np.min(colltime))
        ax5.set_yscale('log')
        ax5.legend(self.leg5,loc=1,prop=self.font)
        ax5.set_xlabel(r'$x/a$',fontsize=self.xyfs)
            #plt.ylabel(r'$\nu_* \cdot \epsilon^{3/2}$',fontsize=18)
        ax5.set_xlim(self.xmin,self.xmax)
        if poutput == 1:
            fig5.frameon = False
            fig5.savefig("nustar%s.pdf"%(self.prdlist[0].fext))

    def plot_fsa(self,poutput):
        fig6 = plt.figure(6)
        ax6 = fig6.add_subplot(111)
        for prd in self.prdlist:
            if 'plotfsa' not in prd.plotset:
                continue
            ax6.plot(prd.xs,prd.intfsamom[:,2])
            self.leg6.append("$dE/dt_{tot}$")
            ax6.plot(prd.xs,prd.intfsamom[:,4])
            self.leg6.append("$d\chi/dxy$")
            ax6.plot(prd.xs,prd.intfsamom[:,5])
            self.leg6.append("$f_1 sources$")
            ax6.plot(prd.xs,prd.intfsamom[:,6])
            self.leg6.append("$f_1 buffer$")
            ax6.plot(prd.xs,prd.intfsamom[:,7])
            self.leg6.append("$f_0 term$")
            ax6.plot(prd.xs,-prd.intfsamom[:,8])
            self.leg6.append("$-v_D f_1$")
            ax6.legend(self.leg6,loc=1,prop=self.font)
        ax6.set_xlabel(r'$x/a$',fontsize=self.xyfs)
        ax6.set_xlim(self.xmin,self.xmax)
        ax6.set_ylabel(r'$\int \frac{dE}{dt} dx$',fontsize=self.xyfs)
        if poutput == 1:
            fig6.frameon = False
            fig6.savefig("fsamoments%s.pdf"%(self.prdlist[0].fext))

    def show(self):
        plt.show()
