#!/usr/bin/env python
import sys
import numpy as np
import pylab as pl
from matplotlib.pyplot import *
from matplotlib.font_manager import fontManager, FontProperties
from ParIO import Parameters
import os
import re

#input arguments
noplot=0
showdt=0
infiles=[]
fextout=[]
fext=[]
nfiles=0
count=0
my_xylim = 0
mintime=0
maxtime=1000000
input_error = 0
iarg=0

def showhelp():
   print "usage:"
   print "nlevplot.py eigenvalues_nl file:"
   print "options: -np : no plot"
   print "         -maxtime <number>  : set maxtime"
   print "         -h                 : show help"
   print "         -showdt            : show dt_nl plot (no eigenvalue plot)"
   exit(0)

while iarg < (len(sys.argv)-1):
   iarg+=1 #skip first arg==skript name
   arg=sys.argv[iarg]
   valid=re.match(r'eigenvalues_nl(.*)',arg)
   if valid:
      infiles.append(str(arg))
      ext=str(valid.group(1))
      fext.append(ext)
      if re.match(r'[.]dat$',ext):
         ext='_dat'
      fextout.append(ext)
      nfiles+=1
   if str(arg)=='-np':
      noplot=1
   if str(arg)=='-h' or str(arg)=='--h' or str(arg)=='-help' or str(arg)=='--help':
      showhelp()
   if str(arg)=='-showdt':
      showdt=1
      noplot=1
   if str(arg)=='-maxtime':
      iarg+=1  #read following arg (should be number)
      maxtime = float(sys.argv[iarg])
      if (maxtime < 0):
         print "maxtime ill defined"
         showhelp()

if nfiles==0:
   print "no iniput files"
   showhelp()

for filename in infiles:
   if (not os.path.isfile(filename)):
      print "%s is not a file: exit\n" %(filename)
      showhelp()


def file_len(fname):
   with open(fname) as f:
      for i, l in enumerate(f):
         pass
   return i + 1

def parse_nlevfile(fname):
   with open(fname) as f:
      nblocks = 0
      max_ev_ind = 0
      ev_ind = 0
      i = 0
      for i, l in enumerate(f):
         ishead=re.search(r'#\S*linear',l)
         if ishead:
            max_ev_ind = max(max_ev_ind,ev_ind)
            nblocks+=1
            ev_ind = 0
         iseigenvalue=re.match(r'(?![#])\s*(\S+)\s*(\S+)\s*$',l)
         if iseigenvalue:
            ev_ind+=1
   return [i + 1,nblocks,max_ev_ind]

def RKpoly(x,y,kin):
   #RK stability polynomial for up to six stages
   ## stability boundary: poly==0
   nstages=len(kin)
   k=np.zeros(6)
   for i in range(nstages):
      k[i]=kin[i]
   k1=k[0]
   k2=k[1]
   k3=k[2]
   k4=k[3]
   k5=k[4]
   k6=k[5]
   #poly = 1+ 2*x +x**2+y**2
   #poly = (x-1)**2+(y)**2
   #poly= 1. + 2*k1*x + k1**2*x**2 + 2*k2*x**2 + 2*k1*k2*x**3 + k2**2*x**4 + \
   #k1**2*y**2 - 2*k2*y**2 + 2*k1*k2*x*y**2 + 2*k2**2*x**2*y**2 + k2**2*y**4

   poly = \
   1 + 2*k1*x + k1**2*x**2 + 2*k2*x**2 + 2*k1*k2*x**3 + 2*k3*x**3 + \
   k2**2*x**4 + 2*k1*k3*x**4 + 2*k4*x**4 + 2*k2*k3*x**5 + \
   2*k1*k4*x**5 + 2*k5*x**5 + k3**2*x**6 + 2*k2*k4*x**6 + 2*k1*k5*x**6 + \
   2*k6*x**6 + 2*k3*k4*x**7 + 2*k2*k5*x**7 + \
   2*k1*k6*x**7 + k4**2*x**8 + 2*k3*k5*x**8 + 2*k2*k6*x**8 + 2*k4*k5*x**9 + \
   2*k3*k6*x**9 + k5**2*x**10 + 2*k4*k6*x**10 + \
   2*k5*k6*x**11 + k6**2*x**12 + k1**2*y**2 - 2*k2*y**2 + 2*k1*k2*x*y**2 - \
   6*k3*x*y**2 + 2*k2**2*x**2*y**2 - \
   12*k4*x**2*y**2 + 4*k2*k3*x**3*y**2 - 4*k1*k4*x**3*y**2 - \
   20*k5*x**3*y**2 + 3*k3**2*x**4*y**2 + 2*k2*k4*x**4*y**2 - \
   10*k1*k5*x**4*y**2 - 30*k6*x**4*y**2 + 6*k3*k4*x**5*y**2 - \
   2*k2*k5*x**5*y**2 - 18*k1*k6*x**5*y**2 + 4*k4**2*x**6*y**2 + \
   4*k3*k5*x**6*y**2 - 8*k2*k6*x**6*y**2 + 8*k4*k5*x**7*y**2 + \
   5*k5**2*x**8*y**2 + 6*k4*k6*x**8*y**2 + 10*k5*k6*x**9*y**2 + \
   6*k6**2*x**10*y**2 + k2**2*y**4 - 2*k1*k3*y**4 + 2*k4*y**4 + \
   2*k2*k3*x*y**4 - 6*k1*k4*x*y**4 + 10*k5*x*y**4 + \
   3*k3**2*x**2*y**4 - 2*k2*k4*x**2*y**4 - 10*k1*k5*x**2*y**4 + \
   30*k6*x**2*y**4 + 6*k3*k4*x**3*y**4 - 10*k2*k5*x**3*y**4 - \
   10*k1*k6*x**3*y**4 + 6*k4**2*x**4*y**4 - 20*k2*k6*x**4*y**4 + \
   12*k4*k5*x**5*y**4 - 12*k3*k6*x**5*y**4 + \
   10*k5**2*x**6*y**4 + 4*k4*k6*x**6*y**4 + 20*k5*k6*x**7*y**4 + \
   15*k6**2*x**8*y**4 + k3**2*y**6 - 2*k2*k4*y**6 + \
   2*k1*k5*y**6 - 2*k6*y**6 + 2*k3*k4*x*y**6 - 6*k2*k5*x*y**6 + \
   10*k1*k6*x*y**6 + 4*k4**2*x**2*y**6 - 4*k3*k5*x**2*y**6 - \
   8*k2*k6*x**2*y**6 + 8*k4*k5*x**3*y**6 - 16*k3*k6*x**3*y**6 + \
   10*k5**2*x**4*y**6 - 4*k4*k6*x**4*y**6 + 20*k5*k6*x**5*y**6 + \
   20*k6**2*x**6*y**6 + k4**2*y**8 - 2*k3*k5*y**8 + 2*k2*k6*y**8 + \
   2*k4*k5*x*y**8 - 6*k3*k6*x*y**8 + 5*k5**2*x**2*y**8 - \
   6*k4*k6*x**2*y**8 + 10*k5*k6*x**3*y**8 + 15*k6**2*x**4*y**8 + k5**2*y**10 - \
   2*k4*k6*y**10 + 2*k5*k6*x*y**10 + 6*k6**2*x**2*y**10 + k6**2*y**12
   return poly


#output figures:
if noplot!=1:
   fig1 = pl.figure()
   label1=[]
   leg1=[]
   plot11 = fig1.add_subplot(1,1,1)
   ax1 = fig1.gca()
   subplots_adjust(left=0.15 , bottom=0.15, right=0.95, top=0.95, wspace=0.2, hspace=0.2)
if (showdt==1):
   fig2 = pl.figure()
   label2=[]
   leg2=[]
   plot21 = fig2.add_subplot(1,1,1)
   ax2 = fig2.gca()
   subplots_adjust(left=0.15 , bottom=0.15, right=0.95, top=0.95, wspace=0.2, hspace=0.2)

par=Parameters()

for n in range(nfiles):
   filename=infiles[n]
   par.Read_Pars('parameters%s' %(fext[n]))
   n_nlev=float(par.pardict['n_nlev'])
   [filelen,nblocks,max_ev_ind]=parse_nlevfile(filename)

   #print "filelen=%i ,nblocks=%i ,max_ev_ind=%i\n" %(filelen,nblocks,max_ev_ind)
   evs=np.zeros((nblocks,max_ev_ind,2))+1000
   dt=np.empty((nblocks))
   dt_nl=np.empty((nblocks))
   time=np.empty((nblocks))
   maxreal = [0.0,0.0]
   maximag = 0.0

   file=open(filename)
   lines=file.readlines()
   iblock = -1
   ev_ind = 0
   
   if nblocks==0:
     print "skipping file %s (no (%i) data blocks)" %(filename,nblocks)
     break
   else:
     print "reading %i data blocks..." %(nblocks)
   for i in range(filelen):

      #print ev_ind
      islinear=re.search(r'#linear(.*)dt_max\s*=\s*(\S*)\s*\)',lines[i])
      if islinear:
         iblock+=1
         dt_max = float(islinear.group(2))
         print "dt_max = %s\n" %(str(dt_max))
         time[iblock]=0.0
         dt_nl[iblock]=dt_max
         dt[iblock] = dt_max
         ev_ind = 0

      isnonlinear=re.search(r'#nonlinear.*time\s*(\S+).*dt_nl\s*=\s*(\S+)\s*,.*dt\s*=\s*(\S+)\)\s*$',lines[i])
      if (isnonlinear):
         iblock+=1
         time[iblock]=float(isnonlinear.group(1))
         print "reading time[%i] = %f\n" %(iblock,time[iblock])
         dt_nl[iblock]=float(isnonlinear.group(2))
         dt[iblock]=float(isnonlinear.group(3))
         print "dt_nl = %s\n" %(str(dt_nl[iblock]))
         ev_ind = 0
      
      iseigenvalue=re.match(r'(?![#])\s*(\S+)\s+(\S+)\s*$',lines[i])
      if (iseigenvalue):
         real = float(iseigenvalue.group(1))*dt_max
         imag = float(iseigenvalue.group(2))*dt_max
         if real<0.0:
            maxreal[0] = np.max([maxreal[0],np.abs(real)])
         else:
            maxreal[1] = np.max([maxreal[1],np.abs(real)])
         maximag = np.max([maximag,np.abs(imag)])
         evs[iblock,ev_ind,0]=real
         evs[iblock,ev_ind,1]=imag
         print iblock,ev_ind,evs[iblock,ev_ind,:]
         ev_ind +=1


   mindtnl=np.min(dt_nl)
   maxreal=[3.3,0.25] #looks nice for RK4
   maximag=1.2*maximag

   #maximag=maximag/dt_max*mindtnl
   #for i in range(len(maxreal)):
   #   maxreal[i]=maxreal[i]/dt_max*mindtnl

   res=50
   x=np.linspace(-maxreal[0],maxreal[1],num=res)
   y=np.linspace(-maximag,maximag,num=res)
   #x=np.linspace(-3,0.5,num=res)
   #y=np.linspace(-4,4,num=res)
   Z = np.zeros((res,res))
   Znl = np.zeros((res,res))
   RKtaylor=[1,1./2,1./6,1./24]  #RK4
   #RKtaylor=[1,1./7.8,0.0]      #RKC2
   for i in range(res):
      for j in range(res):
         #WARNING: we need the transposed Z
         Z[j,i]=float(RKpoly(x[i],y[j],RKtaylor))
         Znl[j,i]=float(RKpoly(x[i]/dt_max*mindtnl,y[j]/dt_max*mindtnl,RKtaylor))
         #print "x=%f y= %f z= %f" %(x[i],y[j],Z[i,j])


   for iblock in range(1,nblocks):
      if (time[iblock] >= mintime and time[iblock] <=maxtime):
         print "plotting time = %f\n" %(time[iblock])
         lw = 3
         sz = 10
         fsz = 20
         lsz = 16
         style=['m^','bo','rd','bo','ks','bo']
         if noplot!=1:
            plot11.set_xlabel(r'Re($\lambda\, \Delta t)$',fontsize=fsz)
            plot11.set_ylabel(r'Im($\lambda\, \Delta t)$',fontsize=fsz)
  
            plot11.set_ylim([-maximag,maximag])
            plot11.set_xlim([-maxreal[0],maxreal[1]])
  
            plot11.xaxis.set_tick_params ( which='major', labelsize=lsz ) 
            plot11.yaxis.set_tick_params ( which='major', labelsize=lsz ) 
            
            plot11.plot(evs[0,:,0],evs[0,:,1],'go',linewidth=lw)
            plot11.plot(evs[iblock,:,0],evs[iblock,:,1],'rx',linewidth=lw)
            plot11.contour(x,y,Z,levels=[1])
            plot11.contour(x,y,Znl,levels=[1],cmap=cm.gray)
  
            savefig('nlev%s_%.4i.png'%(fextout[n],iblock))   
            plot11.clear()
   print "min dt_nl = %f" %(mindtnl)

   if (showdt==1):
      font= FontProperties(size=14);
      fsz = 16
      if (nblocks>1):
         if time[1]-time[0] > 0.1*(time[-1]-time[1]):
            print "set tmin to time(1)"
            tmin=time[1]
         else:
            tmin=min(time)

      plot21.plot(time,dt_nl,time,dt,time,np.ones(len(time))*dt_max)
      plot21.set_xlabel(r'time',fontsize=fsz)
      plot21.set_xlim([tmin,max(time)])
      plot21.set_ylim([0,1.1*dt_max])
      leg1=['dt_nlev','dt_sim','dt_lin']
      legend(leg1,loc=4,prop=font)
      savefig('nlev_dt%s_%.4i.eps'%(fextout[n],iblock))   
      show()
   

